#include <stdlib.h>
#include <string.h>
#include "anamod.h"
#include "petscmat.h"

#define CATEGORYCOOKIE 983429
#define CHECKVALIDCATEGORY(x) ANAMODCHECKVALID(x,CATEGORYCOOKIE,"category")

struct categoryobject_ {
  int cookie;
  char *name;
  int enabled; /* 1: enable, 0: never compute, -1: skip from loops */
  int maxcomponents,ncomponents; 
  PetscErrorCode (*optionfunction)(char*);
  componentobject *components; const char **componentnames;
  AnalysisDataType *types;
} ;

extern int ncategories;
int maxcategories;
extern categoryobject *categoryobjects;
const char **categorynames; /* pointers to the names in the actual objects */

#undef __FUNCT__
#define __FUNCT__ "AllocCategoryObjects"
PetscErrorCode AllocCategoryObjects()
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  maxcategories = MXC; ncategories = 0;
  ierr = PetscMalloc
    (maxcategories*sizeof(categoryobject),&categoryobjects); CHKERRQ(ierr);
  ierr = PetscMalloc
    (maxcategories*sizeof(char*),&categorynames); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "FreeCategoryObjects"
PetscErrorCode FreeCategoryObjects()
{
  int icat; PetscErrorCode ierr;
  PetscFunctionBegin;
  for (icat=0; icat<ncategories; icat++) {
    ierr = DestroyCategoryObject(categoryobjects[icat]); CHKERRQ(ierr);
  }
  ierr = PetscFree(categoryobjects); CHKERRQ(ierr);
  ierr = PetscFree(categorynames); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CreateCategoryObject"
/*! Create a category object with a given name.
  See also DestroyCategoryObject(). 
*/
PetscErrorCode CreateCategoryObject(const char *cat,categoryobject *obj)
{
  categoryobject nnew; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = PetscMalloc(sizeof(struct categoryobject_),&nnew); CHKERRQ(ierr);
  ierr = PetscStrallocpy(cat,(char**)&(nnew->name)); CHKERRQ(ierr);
  nnew->cookie = CATEGORYCOOKIE; nnew->enabled = CATCMP_ENABLE;
  nnew->maxcomponents = MXC; nnew->ncomponents = 0;
  ierr = PetscMalloc
    (nnew->maxcomponents*sizeof(componentobject),&(nnew->components)); CHKERRQ(ierr);
  ierr = PetscMalloc
    (nnew->maxcomponents*sizeof(char*),&(nnew->componentnames)); CHKERRQ(ierr);
  ierr = PetscMalloc
     (nnew->maxcomponents*sizeof(AnalysisDataType),&(nnew->types)); CHKERRQ(ierr);
  *obj = nnew;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DestroyCategoryObject"
/*! Deallocate a category object. See also CreateCategoryObject(). */
PetscErrorCode DestroyCategoryObject(categoryobject obj)
{
  int icmp; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDCATEGORY(obj);
  for (icmp=0; icmp<obj->ncomponents; icmp++) {
    ierr = DestroyComponentObject(obj->components[icmp]); CHKERRQ(ierr);
  }
  ierr = PetscFree(obj->components); CHKERRQ(ierr);
  ierr = PetscFree(obj->componentnames); CHKERRQ(ierr);
  ierr = PetscFree(obj->types); CHKERRQ(ierr);
  ierr = PetscFree(obj->name); CHKERRQ(ierr);
  ierr = PetscFree(obj); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "GetCategoryIndex"
/*! Test for existence of a category, and return its index
  if present. Both output parameters can be null.
*/
static PetscErrorCode GetCategoryIndex(const char* cat,int *icat,PetscBool *flag)
{
  int i; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  flg = PETSC_FALSE;
  for (i=0; i<ncategories; i++) {
    ierr = PetscStrcmp(cat,categoryobjects[i]->name,&flg); CHKERRQ(ierr);
    if (flg) {if (icat) *icat = i; goto found;}
  }
 found:
  if (flag) *flag = flg;
  if (!flag && !flg)
    SETERRQ(MPI_COMM_WORLD,1,"Category not found but no flag provided to report this");
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "GetOrCreateCategory"
/*! Return a named category, creating it if necessary */
PetscErrorCode GetOrCreateCategory(const char *cat,categoryobject *catg)
{
  int icat; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetCategoryIndex(cat,&icat,&flg); CHKERRQ(ierr);
  if (!flg) {
    icat = ncategories++;
    if (icat>=maxcategories)
      SETERRQ(MPI_COMM_WORLD,1,"Category reallocation not implemented");
    ierr = CreateCategoryObject(cat,&categoryobjects[icat]); CHKERRQ(ierr);
    categorynames[icat] = categoryobjects[icat]->name;
  }
  *catg = categoryobjects[icat];
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "GetCategory"
/*! Return a named category */
PetscErrorCode GetCategory(const char *cat,categoryobject *catg,PetscBool *f)
{
  int icat; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetCategoryIndex(cat,&icat,&flg); CHKERRQ(ierr);
  if (f) *f = flg;
  if (flg && catg) *catg = categoryobjects[icat];
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "GetCategories"
PetscErrorCode GetCategories(int *ncat,const char ***cats)
{
  PetscFunctionBegin;
  if (ncat) *ncat = ncategories;
  if (cats) *cats = categorynames;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CategoryGetComponentIndex"
/*! Test for presence of a component in a category, and return its index
  if present. The index parameter can be null.
*/
PetscErrorCode CategoryGetComponentIndex
(categoryobject catg,const char *cmp,int *icmp,PetscBool *flag)
{
  int i; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDCATEGORY(catg);
  flg = PETSC_FALSE;
  for (i=0; i<catg->ncomponents; i++) {
    ierr = PetscStrcmp(cmp,catg->componentnames[i],&flg); CHKERRQ(ierr);
    if (flg) {if (icmp) *icmp = i; goto found;}
  }
 found:
  if (flag) *flag = flg;
  if (!flg && !flag)
    SETERRQ(MPI_COMM_WORLD,1,"Component not found but no flag to return this info");
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CategoryGetOrCreateComponent"
PetscErrorCode CategoryGetOrCreateComponent
(categoryobject catg,const char *cmp,componentobject *cmpt)
{
  int icmp; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDCATEGORY(catg);
  ierr = CategoryGetComponentIndex(catg,cmp,&icmp,&flg); CHKERRQ(ierr);
  if (!flg) {
    icmp = catg->ncomponents++;
    if (icmp>=catg->maxcomponents)
      SETERRQ(MPI_COMM_WORLD,1,"Component reallocation not implemented");
    ierr = CreateComponentObject(cmp,&(catg->components[icmp])); CHKERRQ(ierr);
    ierr = ComponentGetName
      (catg->components[icmp],&(catg->componentnames[icmp])); CHKERRQ(ierr);
  }
  *cmpt = catg->components[icmp];
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CategoryGetComponent"
/*! Retrieve a component.

  See also CategoryGetOrCreateComponent(), CategoryGetComponentIndex().
*/
PetscErrorCode CategoryGetComponent
(categoryobject catg,const char *cmp,componentobject *cmpt,PetscBool *success)
{
  int icmp; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDCATEGORY(catg);
  ierr = CategoryGetComponentIndex(catg,cmp,&icmp,&flg); CHKERRQ(ierr);
  if (success) *success = flg;
  if (flg) {
    if (cmpt) *cmpt = catg->components[icmp];
  } else if (!success)
    SETERRQ(MPI_COMM_WORLD,1,"Component not found but no flag given to report this");
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CategoryComponentSetModule"
PetscErrorCode CategoryComponentSetModule
(const char *cat,const char *cmp,AnalysisDataType type,int id,
 PetscErrorCode(*f)(AnaModNumericalProblem,AnalysisItem*,int*,PetscBool*))
{
  categoryobject catg; componentobject cmpt; int icmp; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetOrCreateCategory(cat,&catg); CHKERRQ(ierr);
  ierr = CategoryGetOrCreateComponent(catg,cmp,&cmpt); CHKERRQ(ierr);
  ierr = ComponentSetModule(cmpt,type,id,f); CHKERRQ(ierr);
  ierr = CategoryGetComponentIndex(catg,cmp,&icmp,PETSC_NULL); CHKERRQ(ierr);
  catg->types[icmp] = type;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CategoryGetModules"
/*! Query the modules in a specified category.

  The category name has to exist. The routine will call the Petsc error
  handler if the name is invalid.

  Parameters:
  - \c cat : the category that is being queried
  - \c ms (optional) : the names of the modules in the category
  - \c t (optional) : the corresponding list of datatypes
  - \c id (optional) : the list of module IDs (see RetrieveQuantityByID())
  - \c n (optional) : the number of modules in the category

  See also GetCategories() and HasComputeCategory().
*/
PetscErrorCode CategoryGetModules
(const char *cat,const char ***ms,AnalysisDataType **t,int **id,int *n)
{
  categoryobject catg; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetCategory(cat,&catg,&flg); CHKERRQ(ierr);
  if (!flg) SETERRQ1(MPI_COMM_WORLD,1,"Unknown category <%s>",cat);
  if (ms) *ms = catg->componentnames;
  if (t) *t = catg->types;
  if (n) *n = catg->ncomponents;
  //      if (id) *id = ids[i];
  PetscFunctionReturn(0);
}

/*
 * Stuff for category reporting
 */
static int categoryreadout=-1;

#undef __FUNCT__
#define __FUNCT__ "GetFirstCategory"
PetscErrorCode GetFirstCategory(const char **rname,PetscBool *rfound)
{
  PetscBool found = PETSC_FALSE;
  PetscFunctionBegin;
  categoryreadout = 0;
  while (categoryreadout<ncategories 
	 && categoryobjects[categoryreadout]->enabled!=CATCMP_ENABLE)
    categoryreadout++;
  if (categoryreadout<ncategories) {
    found = PETSC_TRUE;
    if (rname) *rname = categoryobjects[categoryreadout]->name;
  }
  *rfound = found;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "GetNextCategory"
PetscErrorCode GetNextCategory(const char **rname,PetscBool *rfound)
{
  PetscBool found = PETSC_FALSE;
  PetscFunctionBegin;
  categoryreadout++;
  while (categoryreadout<ncategories 
	 && categoryobjects[categoryreadout]->enabled!=CATCMP_ENABLE)
    categoryreadout++;
  if (categoryreadout<ncategories) {
    found = PETSC_TRUE;
    if (rname) *rname = categoryobjects[categoryreadout]->name;
  }
  *rfound = found;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CategoryEnableByName"
/*! Mark a category as enabled/disabled. Values (CatCmpEnableMode):
  - 0 : enable
  - 1 : skip from loops
  - 2 : skip altogether
*/
PetscErrorCode CategoryEnableByName(const char *cat,int mode)
{
  int icat;
  PetscFunctionBegin;
  for (icat=0; icat<ncategories; icat++) {
    if (!strcmp(cat,categoryobjects[icat]->name)) {
      categoryobjects[icat]->enabled = mode;
    }
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeclareCategoryOptionFunction"
/*! This function allows the module developer to give the user
  commandline options for control of a module.

  See DeclareCategoryOptionFunction(),GetCategoryOptionFunction(),
  AnaModOptionsHandling() and section \ref optionsfile.
*/
PetscErrorCode DeclareCategoryOptionFunction
  (const char *cat,PetscErrorCode(*f)(char*))
{
  PetscBool flg; categoryobject catg; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetCategory(cat,&catg,&flg); CHKERRQ(ierr);
  if (flg) {
    catg->optionfunction = f;
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CategoryGetOptionFunction"
/*! This function is called in AnaModOptionsHandling().
  There is probably no reason for the user ever to call it.

  See DeclareCategoryOptionFunction(),GetCategoryOptionFunction(),
  AnaModOptionsHandling() and section \ref optionsfile.
*/
PetscErrorCode CategoryGetOptionFunction
(const char *cat,PetscErrorCode(**f)(char*))
{
  PetscBool flg; categoryobject catg; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetCategory(cat,&catg,&flg); CHKERRQ(ierr);
  if (flg) {
    *f = catg->optionfunction;
  } else *f = NULL;
  PetscFunctionReturn(0);
}
 
