#include <stdlib.h>
#include "string.h"
#include "anamod.h"
#include "petsc.h"

/*! \page featureset Feature sets
  
  AnaMod has a mechanism for querying single category/component pairs,
  but often a specific set of features is needed quickly, and/or multiple
  times. For this, the `feature set' mechanism exists.

  A feature set is a set of category/component pairs. It does not contain
  actual values. A feature set can be instantiated to a `feature values' 
  object, which contains the actual values that the feature set takes
  on a given problem.

  This is the workflow:
  - a FeatureSet object is created, once, with NewFeatureSet().
  - category/component pairs are added to it with AddToFeatureSet().
  - a FeatureValues object is created with NewFeatureValues().
  - a call to InstantiateFeatureSet() fills in the feature set on a 
  specified numerical problem.
  - user code can retrieve values with GetFeatureValue()

  Here is a piece of example code:
\code
  {
    FeatureSet symmetry; int sidx,aidx;
    ierr = NewFeatureSet(&symmetry); CHKERRQ(ierr);
    ierr = AddToFeatureSet
      (symmetry,"simple","symmetry-snorm",&sidx); CHKERRQ(ierr);
    ierr = AddToFeatureSet
      (symmetry,"simple","symmetry-anorm",&aidx); CHKERRQ(ierr);
    ierr = TransformObjectSetSuitabilityFunction
      (cur,(void*)symmetry,&onlyforsymmetricproblem); CHKERRQ(ierr);
  }

PetscErrorCode onlyforsymmetricproblem
(NumericalProblem problem,void *ctx,SuitabilityValue *v)
{
  FeatureSet features = (FeatureSet)ctx; FeatureValues values;
  AnalysisItem sn,an; PetscBool f1,f2; PetscErrorCode ierr;
  
  PetscFunctionBegin;
  ierr = NewFeatureValues(&values); CHKERRQ(ierr);
  ierr = InstantiateFeatureSet((void*)problem,features,values); CHKERRQ(ierr);
  ierr = GetFeatureValue(values,sidx,&sn,&f1); CHKERRQ(ierr);
  ierr = GetFeatureValue(values,aidx,&an,&f2); CHKERRQ(ierr);
  ierr = DeleteFeatureValues(values); CHKERRQ(ierr);
  if (f1 && f2 && an.r>1.e-12*sn.r) {
    printf("problem too unsymmetric\n");
  }
\endcode

 */

#define NALLOC 25
#define FSETCOOKIE 9876
#define CHECKVALIDFSET(i) {ANAMODCHECKVALID(i,FSETCOOKIE,"feature set");}
struct FeatureSet_ {
  int cookie;
  AnalysisDataType *types; IntArray IDs; StringArray features;
};
#define FVALCOOKIE 9877
#define CHECKVALIDFVAL(i) {ANAMODCHECKVALID(i,FVALCOOKIE,"feature values");}
struct FeatureValues_ {
  int cookie;
  AnalysisItemArray values; AnalysisDataTypeArray types;
};

#undef __FUNCT__ 
#define __FUNCT__ "NewFeatureSet"
/*! Allocate a featureset object. See \ref featureset */
PetscErrorCode NewFeatureSet(FeatureSet *set)
{
  FeatureSet fnew; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscMalloc(sizeof(struct FeatureSet_),&fnew); CHKERRQ(ierr);
  ierr = PetscMemzero(fnew,sizeof(struct FeatureSet_)); CHKERRQ(ierr);
  fnew->cookie = FSETCOOKIE;
  ierr = CreateIntArray("IDs",50,&(fnew->IDs)); CHKERRQ(ierr);
  ierr = CreateStringArray("features",50,&(fnew->features)); CHKERRQ(ierr);
  *set = fnew;
  PetscFunctionReturn(0);
}

#undef __FUNCT__ 
#define __FUNCT__ "DeleteFeatureSet"
/*! Delete a featureset object. See \ref featureset */
PetscErrorCode DeleteFeatureSet(FeatureSet set)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDFSET(set);
  ierr = DeleteIntArray(set->IDs); CHKERRQ(ierr);
  ierr = DeleteStringArray(set->features); CHKERRQ(ierr);
  ierr = PetscFree(set); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__ 
#define __FUNCT__ "AddToFeatureSet"
/*! Add a requested feature to a featureset object. See \ref featureset

  Arguments:
  - \c set : the featureset
  - \c cat,cmp : the category and component name. It is an error to 
  supply unknown names
  - \c idx : the index under which the feature is known in this featureset.
  This index can be supplied to GetFeatureValue().
  This parameter can be null.

*/
PetscErrorCode AddToFeatureSet
(FeatureSet set,const char *cat,const char *cmp,int *idx)
{
  PetscErrorCode ierr;
  char *name; int l1,l2;
  PetscFunctionBegin;
  CHECKVALIDFSET(set);

  l1 = strlen(cat); l2 = strlen(cmp);
  ierr = PetscMalloc((l1+l2+1)*sizeof(char),&name); CHKERRQ(ierr);
  sprintf(name,"%s/%s",cat,cmp);
  ierr = StringArrayAdd(set->features,name,idx); CHKERRQ(ierr);
  ierr = PetscFree(name); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__ 
#define __FUNCT__ "NewFeatureValues"
/*! Allocate a featurevalues object. See \ref featureset */
PetscErrorCode NewFeatureValues(FeatureValues *values)
{
  FeatureValues fnew; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscMalloc(sizeof(struct FeatureValues_),&fnew); CHKERRQ(ierr);
  ierr = PetscMemzero(fnew,sizeof(struct FeatureValues_)); CHKERRQ(ierr);
  fnew->cookie = FVALCOOKIE;
  ierr = CreateAnalysisItemArray("values",50,&fnew->values); CHKERRQ(ierr);
  ierr = CreateAnalysisDataTypeArray("values",50,&fnew->types); CHKERRQ(ierr);
  *values = fnew;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeleteFeatureValues"
/*! Free a featurevalues object. See \ref featureset */
PetscErrorCode DeleteFeatureValues(FeatureValues values)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDFVAL(values);
  ierr = DeleteAnalysisItemArray(values->values); CHKERRQ(ierr);
  ierr = DeleteAnalysisDataTypeArray(values->types); CHKERRQ(ierr);
  ierr = PetscFree(values); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__ 
#define __FUNCT__ "InstantiateFeatureSet"
/*! Fill in a featurevalues object. See \ref featureset */
PetscErrorCode InstantiateFeatureSet
(AnaModNumericalProblem prob,FeatureSet set,FeatureValues values)
{
  PetscErrorCode (*retriever)
    (void*,char*,char*,AnalysisItem*,AnalysisDataType*,PetscBool*);
  int nval,ival; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDFSET(set);
  CHECKVALIDFVAL(values);
  ierr = AnaModGetRetrievalFunction(&retriever,&flg); CHKERRQ(ierr);
  if (!flg) PetscFunctionReturn(0);
  ierr = StringArrayGetFill(set->features,&nval); CHKERRQ(ierr);
  for (ival=0; ival<nval; ival++) {
    char *feature,*cat,*cmp; int l;
    AnalysisItem res; AnalysisDataType t;
    ierr = StringArrayGetAt(set->features,ival,&feature); CHKERRQ(ierr);
    for (l=0; l<strlen(feature); l++) {
      if (feature[l]=='/') {
	feature[l] = 0; cat = feature; cmp = feature+l+1;
	ierr = (*retriever)(prob,cat,cmp,&res,&t,&flg); CHKERRQ(ierr);
	if (flg) {
	  ierr = AnalysisItemArraySetAt(values->values,ival,res); CHKERRQ(ierr);
	  ierr = AnalysisDataTypeArraySetAt(values->types,ival,t); CHKERRQ(ierr);
	}
	feature[l] = '/'; break;
      }
    }
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "GetFeatureValue"
/*! Extract a value from a featurevalues object. See \ref featureset.

  Arguments:
  - \c values : the FeatureValues object.
  - \c index : the index, as returned by AddToFeatureSet().
  - \c val (output) : the value; this argument can be null.
  - \c f (output) : indicates whether the return value was indeed
  preset in the featurevalues object; this argument can be null.

*/
PetscErrorCode GetFeatureValue
(FeatureValues values,int index,AnalysisItem *val,PetscBool *f)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDFVAL(values);
  ierr = AnalysisItemArrayTryGetAt(values->values,index,val,f); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

static PetscErrorCode(*retriever)(void*,char*,char*,AnalysisItem*,AnalysisDataType*,PetscBool*) = NULL;
#undef __FUNCT__ 
#define __FUNCT__ "AnaModSetRetrievalFunction"
PetscErrorCode AnaModSetRetrievalFunction
(PetscErrorCode(*fun)(void*,char*,char*,AnalysisItem*,AnalysisDataType*,PetscBool*))
{
  PetscFunctionBegin;
  retriever = fun;
  PetscFunctionReturn(0);
}

#undef __FUNCT__ 
#define __FUNCT__ "AnaModGetRetrievalFunction"
PetscErrorCode AnaModGetRetrievalFunction
(PetscErrorCode(**fun)(void*,char*,char*,AnalysisItem*,AnalysisDataType*,PetscBool*),PetscBool *flg)
{
  PetscBool has;
  PetscFunctionBegin;
  has = (PetscBool)(retriever!=NULL);
  if (flg) *flg = has;
  if (has &&fun) *fun = retriever;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModCheckValidFeatureSet"
PetscErrorCode AnaModCheckValidFeatureSet(void *f)
{
  PetscFunctionBegin;
  CHECKVALIDFSET((FeatureSet)f);
  PetscFunctionReturn(0);
}
