/*! \defgroup categories Files with defined categories */
/*! \defgroup functions User-accessible functions */
/*! \defgroup implementation Functions only of use to the implementation */

/*! \file module_functions.c \ingroup functions \anchor modules
  \brief User functions for accessing the analysis modules

  Analysis modules need to be defined with RegisterModule(), after which
  they can be invoked with ComputeQuantity(). See also HasQuantity().
  The functions AnaModRegisterStandardModules() installs all standard 
  available modules.

  There are utility functions for querying the existence of modules:
  GetCategories(), CategoryGetModules(),
  HasComputeCategory(), HasComputeModule().

  Utility functions: QuantityAsString(), GetDataType(), GetDataID(),
  GetCategoryIndex(), GetModuleIndex().
 */

/*! \mainpage AnaMod : an Analysis Modules library
  \section Introduction

  This is a library of modules that can compute various properties of
  a matrix. The code heavily uses the Petsc library, and the matrix has to
  be stored in Petsc format.

  Modules are divided into several categories, each defined in its own file.
  Typical categories are for structural information, normlike properties, 
  spectral estimate. A number of modules have been supplied, but because of
  the modular setup, it is easy to add new modules.
 
  This library was written for use in the Salsa project
  (http://icl.cs.utk.edu/salsa/), but can be
  used independently of it. If the Salsa NMD library 
  (http://icl.cs.utk.edu/salsa/software/index.html) is present (see the
  installation instructions), its data types are used.

  \subpage installation

  \subpage usage

  \subpage list

  \subpage customization 

  \subpage getall

  \ref tracing

  \author Victor Eijkhout
  \version 1.91
  \date unreleased

  Change log:
  - 2.0 : internals revamped, bug fixes in simple:diagonal-dominance
  - 1.91 : spectrum modules with ";re" and ";im" have been renamed
    to "-re" and "-im" for MySQL sake.
  - 1.9 : fixed memory leak in dummy row computation
  - 1.8.a : added the stats.c file
  - 1.8 : much code cleanup; unit tests to ensure proper operation
  - 1.7 : handling of arrays is now completely changed. most routines
    have acquired an explicit array length parameter
  - 1.6 : handling of the location of slepc and such. Make sure to check
    the Make.inc.example file
  - 1.5 :
    - module-specific options
    - \c -anamod_force now has underscore for consistency
    - several spectrum imaginary parts were miscomputed
    - added SLEPC computation of spectrum
  - 1.4 : 
    - improved trace mode; incompatible change in the prototype
    of the trace function
    - removed a11 module
  - 1.3 :
    - trace mode added (see \ref tracing)
    - renamed header files: module_functions.h -> anamod.h ,
      module_types.h -> anamodtypes.h , nmdmodules.h -> anamodsalsamodules.h ,
      module_utils.h -> anamodutils.h
  - 1.2 : 
    - added runtime options: forced computation of single-processor
    modules in parallel context, of expensive modules
    - bug fix of aux:a11.
    - compatibility with gcc4 (ignoring 'const' used to be a warning,
    it's now an error)
  - 1.1 : correct handling of parallelism
*/

/*! \page installation Installation and compiling programs

  \section Installation

  See the README file. The main thing is that you need to have Petsc installed.

  \section compilation Compilation of user programs

  The easiest way to compile a program that uses the Salsa Analysis Modules
  is to have these includes in your makefile:
\verbatim
include $(PETSC_DIR)/bmake/common/variables
include $(SALSA_MODULES_DIR)/Make.inc
\endverbatim

  The following flags for your C compiler are
  - line 1: Petsc option
  - line 2: Analysis Modules options
  - line 3: (optionally) options for using the NMD library; HAVE_NMD_DEFINE 
    has to be "-DHAVE_NMD" if the library is present
\verbatim
CFLAGS = \
    $(PETSC_INCLUDE) $(COPTFLAGS) \
    -I$(SALSA_MODULES_DIR) \
    $(HAVE_NMD_DEFINE) -I$(LIBNMD_INCLUDE_DIR) -I$(LIBXMLSC_INCLUDE_DIR)
\endverbatim

  Your program needs to include the following header files:
\verbatim
#include "anamod.h"
#include "anamodsalsamodules.h"
\endverbatim

  The following link line brings together all the needed libraries
  (add libnmd if required)
\verbatim
yourprog : 
	$(CLINKER) -o yourprog yourprog.o \
	    -L$(SALSA_MODULES_LIB_DIR) -lsalsamodules -lothermodules \
	    $(PETSC_LIB)
\endverbatim
*/

/*! \page usage Use of the analysis modules

  \section init Library Initialization

  There are calls AnaModInitialize() and AnaModFinalize() which do 
  global allocation. These are necessary.

  \section registration Category registration

  Every category needs to be registered before use with a call
  
  Register\<Category\>Modules();

  where the available categores are \ref icmk, \ref iprs, \ref jpl, 
  \ref normal, \ref simple,
  \ref spectrum, \ref structure, \ref variance.
  The functions AnaModRegisterStandardModules() installs all standard 
  available modules.

  \section compute Quantity computation

  After registring a category, its elements can be computed as

  PetscErrorCode ComputeQuantity(char *cat,char *cmp,Mat A,AnalysisItem *res,PetscBool *success);

  with
  -# the name of the category
  -# the name of the element
  -# the matrix
  -# the result
  -# a success indicator

  For the main user functions, see \ref modules "the module functions file"

  Commandline options are discussed in section \ref options.

  \section types Datatypes

  The \c AnalysisDataType type is an integer. To get a printable name
  of the datatype, use AnaModGetTypeName() or AnaModGetTypeMySQLName().
 */

/*! \page list List of the available modules

  \subpage icmk

  \subpage iprs

  \subpage jpl

  \subpage normal

  \subpage simple

  \subpage spectrum

  \subpage structure

  \subpage variance
*/

/*! \page customization Customization

  While this library comes with a number of categories supplied, you can 
  write your own category, taking any of the given ones as example.

  Each analysis module has to have the following prototype:
\verbatim
static PetscErrorCode MyModule(Mat A,AnalysisItem *rv,PetscBool *flg)
\endverbatim
  where
  - \c A is the input matrix
  - \c rv is the return value
  - \c flg is true if the quantity was succesfully computed, false otherwise

  The module return code should be 0 for success, anything else for
  (catastrophic) failure. A simple failure (or refusal) to compute should
  be indicated through the \c flg variable.

  The modules (including the return type of their computed quantities)
  are declared to the system by calling routine
\verbatim
PetscErrorCode RegisterMyModules()
{
  PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = RegisterModule
    ("mycategory","mymodule",THERESULTTYPE,&MyModule); CHKERRQ(ierr);
  ....
\endverbatim
  or you can make the invidual calls to RegisterModule(). 
  See \ref types and \ref arraytypes.
*/

/*! \page arraytypes Array type handling

  The actual types are listed in \ref types ; here are some semantic issues
  pertaining to array types.

  The uniform calling structure of the modules allows only one parameter
  to be returned. This is a problem only for array types: we want both the 
  array data and the length of the array. Hence we adopt the convention that
  every array is longer by one element than necessary, and the zero element
  contains the length of the array proper. That is, to store 5 elements, 
  6 elements are allocated, and the zero element contains the number '5'.

  If, in this documentation, we refer to location array[i], we will
  take this to mean location `i' after the size element. Zero-based
  indexing is used.
*/

/*! \page attached Attached quantities in Petsc
  
  Petsc has a mechanism where quantities can be attached to objects. 
  For instance, integers can be attached and queried with
  PetscObjectComposedDataSetInt() and PetscObjectComposedDataGetInt().
  Such data is attached together with the current state of the object,
  which means that it will be properly invalidated if the object is 
  altered.

  This composition mechanism is used in this analysis modules package.
  There are many quantities (such as the norms of the symmetric and 
  anti-symmetric part of the matrix) that can be computed together at
  practically the same cost as either one alone. Thus, when one is requested
  by calling ComputeQuantity(), the other one will be compute as well,
  and attached to the matrix object. A call to ComputeQuantity()
  for this second quantity will then just return the attached number,
  and not be computed separately.

  The routine HasQuantity() can query which quantities are already attached.

  As a result of the use of attached quantities, benchmarking the AnaMod
  modules is a bit tricky. You can forced computation of the requested 
  quantities by calling PetscObjectIncreaseState() on the matrix object
  after each call to ComputeQuantity().
 */
#include <stdlib.h>
#include <string.h>
#include "anamod.h"
#include "anamodsalsamodules.h"

static struct{int id; const char *name,*mysqlname;} anamodtypenames[5];
static int nAnaModTypeNames;
static int AnaModIsInitialized = 0;

int ncategories=0;
categoryobject *categoryobjects = NULL;

#undef __FUNCT__
#define __FUNCT__ "AnaModInitialize"
/*! Initializatin for AnaMod. See also AnaModFinalize() */
PetscErrorCode AnaModInitialize()
{
  PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = AllocCategoryObjects(); CHKERRQ(ierr);

  anamodtypenames[0].id = NMDString;
  anamodtypenames[0].name="string"; 
  anamodtypenames[0].mysqlname="VARCHAR(256)";

  anamodtypenames[1].id = NMDInt;
  anamodtypenames[1].name="int"; 
  anamodtypenames[1].mysqlname="INTEGER";

  anamodtypenames[2].id = NMDReal;
  anamodtypenames[2].name="real"; 
  anamodtypenames[2].mysqlname="DOUBLE";

  anamodtypenames[3].id = NMDIntarray;
  anamodtypenames[3].name="intarray"; 
  anamodtypenames[3].mysqlname="VARCHAR(1024)";

  anamodtypenames[4].id = NMDRealarray;
  anamodtypenames[4].name="realarray";
  anamodtypenames[4].mysqlname="VARCHAR(1024)";

  AnaModIsInitialized = 1; nAnaModTypeNames = 5;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModFinalize"
/*! Finalization for AnaMod. See also AnaModInitialize() */
PetscErrorCode AnaModFinalize()
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = FreeCategoryObjects(); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModGetTypeName"
PetscErrorCode AnaModGetTypeName(int id,const char **name)
{
  int i;
  PetscFunctionBegin;
  for (i=0; i<nAnaModTypeNames; i++) {
    if (id==anamodtypenames[i].id) {
      *name = (char*)anamodtypenames[i].name; goto exit;
    }
  }
  SETERRQ1(MPI_COMM_WORLD,1,"Unknown AnaMod datatype %d",id);
 exit:
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "AnaModGetTypeMySQLName"
PetscErrorCode AnaModGetTypeMySQLName(int id,const char **name)
{
  int i;
  PetscFunctionBegin;
  if (AnaModIsInitialized<1) 
      SETERRQ(MPI_COMM_WORLD,1,"AnaMod has not been not formally initialized.\nInsert a call to AnaModInitialize() in your code");
  for (i=0; i<nAnaModTypeNames; i++) {
    if (id==anamodtypenames[i].id) {
      *name = (char*)anamodtypenames[i].mysqlname; goto exit;
    }
  }
  SETERRQ1(MPI_COMM_WORLD,1,"Unknown AnaMod datatype %d",id);
 exit:
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "RegisterModule"
/*!
  Register a new computational module

  This adds a computational routine (the \c f argument) into the modules
  database under the given category (\c cat) and module (\c cmp) label.
  If the category does not exist yet, it is created.
  
  The available types are defined in \ref types "anamodtypes.h"

  If the routine is NULL, only the category and component are created.

  Routine prototype:
  - problem (input)
  - item (output)
  - array length (output)
  - success (output)

  See also HasComputeModule(), ComputeQuantity().
*/
PetscErrorCode RegisterModule
(const char *cat,const char *cmp,AnalysisDataType type,
 PetscErrorCode(*f)(AnaModNumericalProblem,AnalysisItem*,int*,PetscBool*))
{
  int id; PetscErrorCode ierr;
  PetscFunctionBegin;

  if (!AnaModIsInitialized)
      SETERRQ(MPI_COMM_WORLD,1,"AnaMod not initialized: insert call to AnaModInitialize()\n");
  /*
   * Check for illegal names
   */
  if (strchr(cat,':'))   SETERRQ(MPI_COMM_WORLD,1,"Illegal colon in category name");
  if (strchr(cmp,':'))   SETERRQ(MPI_COMM_WORLD,1,"Illegal colon in component name");

#if defined(ANAMODDEBUG)
  printf("Defining module <%s:%s>\n",cat,cmp);
#endif
  ierr = PetscObjectComposedDataRegister(&id); CHKERRQ(ierr);
  ierr = CategoryComponentSetModule(cat,cmp,type,id,f); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeRegisterCategory"
/*! Deallocate the storage for a particular category of analysis modules.
 No longer needed. */
PetscErrorCode DeRegisterCategory(const char *cat)
{
/*   categoryobject catg; */
/*   PetscBool flg; PetscErrorCode ierr; */
  PetscFunctionBegin;

  printf("Please remove use of legacy function DeRegisterCategory\n");
/*   ierr = GetCategory(cat,&catg,&flg); CHKERRQ(ierr); */
/*   if (!flg) SETERRQ1(MPI_COMM_WORLD,1,"Invalid category <%s>",cat); */
/*   ierr = DestroyCategoryObject(catg); CHKERRQ(ierr); */
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeregisterModules"
/*! This function is no longer needed */
PetscErrorCode DeregisterModules()
{
/*   int icat,ncat; const char **cat; */
/*   PetscErrorCode ierr; */
  PetscFunctionBegin;
  printf("Please remove use of legacy function DeRegisterModules\n");
/*   ierr = GetCategories(&ncat,&cat); CHKERRQ(ierr); */
/*   for (icat=0; icat<ncat; icat++) { */
/*     ierr = DeRegisterCategory(cat[icat]); CHKERRQ(ierr); */
/*   } */
  PetscFunctionReturn(0);
}

/*! \page getall Bulk computation
  After the modules have been installed, the typical user will only
  use ComputeQuantity(). However, for purposes of reporting and such
  it may be necessary to query systematicaly the available modules.
  This can be done with GetCategories() and CategoryGetModules().

  Related: HasComputeCategory(), HasComputeModule().

  Auxiliary routines: GetCategoryIndex(), GetModuleIndex().
*/
#undef __FUNCT__
#define __FUNCT__ "HasComputeCategory"
/*! Query whether a specified category has been declared. */
PetscErrorCode HasComputeCategory(const char *cat,PetscBool *f)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetCategory(cat,PETSC_NULL,f); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "HasComputeModule"
/*!
  Query whether a specified module exists inside a specified category. 
  The category need not itself have been declared.
 */
PetscErrorCode HasComputeModule(const char *cat,const char *cmp,PetscBool *f)
{
  categoryobject catg; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetCategory(cat,&catg,f); CHKERRQ(ierr);
  if (*f) {
    ierr = CategoryGetComponent(catg,cmp,PETSC_NULL,f); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#define MODE_RETRIEVE 1
#define MODE_COMPUTE 0
#undef __FUNCT__
#define __FUNCT__ "ComputeOrRetrieveQuantity"
static PetscErrorCode ComputeOrRetrieveQuantity
(AnaModNumericalProblem prob,const char *cat,const char *cmp,
 AnalysisItem *res,int *rreslen,PetscBool *success,int mode)
{
  categoryobject catg; componentobject cmpt; PetscBool flg,trace;
  int reslen; PetscErrorCode ierr;
  
  PetscFunctionBegin;
  ierr = AnaModHasTrace(&trace); CHKERRQ(ierr);

  /*
   * Get the module and compute
   */
  ierr = GetCategory(cat,&catg,&flg); CHKERRQ(ierr);
  if (!flg) SETERRQ1(MPI_COMM_WORLD,1,"Could not find category <%s>",cat);
  ierr = CategoryGetComponent(catg,cmp,&cmpt,&flg); CHKERRQ(ierr);
  if (!flg) SETERRQ1(MPI_COMM_WORLD,1,"Could not find component <%s>",cmp);
  if (mode==MODE_COMPUTE) {
    ierr = ComponentCompute(cmpt,prob,res,&reslen,&flg); CHKERRQ(ierr);
  } else if (mode==MODE_RETRIEVE) {
    ierr = ComponentRetrieve(cmpt,prob,res,&reslen,&flg); CHKERRQ(ierr);
  } else   SETERRQ(MPI_COMM_WORLD,1,"Invalid mode");

  if (rreslen) *rreslen = reslen;
  if (success) *success = flg;

  /*
   * Tracing
   */
  if (trace) {
    if (flg) {
      char *string; PetscBool trace_arrays; AnalysisDataType t;

      ierr = ComponentGetType(cmpt,&t); CHKERRQ(ierr);
      ierr = AnaModTraceArrays(&trace_arrays); CHKERRQ(ierr);
      if (t<ANALYSISINTARRAY || trace_arrays) {
	ierr = QuantityAsString(res,t,(const char**)&string); CHKERRQ(ierr);
	ierr = AnaModTraceMessage
	  ("Anamod computed <%s:%s> = <%s>\n",cat,cmp,string); CHKERRQ(ierr);
	ierr = PetscFree(string); CHKERRQ(ierr);
      }
    } else {
      ierr = AnaModTraceMessage
	("Anamod failed to compute <%s:%s>\n",cat,cmp); CHKERRQ(ierr);
    }
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "ComputeQuantity"
/*!
  Compute a computational module from a certain category. 

  Argument:
  -# the name of the category (see GetCategories())
  -# the name of the module (see CategoryGetModules())
  -# the matrix
  -# a pointer to the result. This is given as \c "(AnalysisItem*)&res" ;
    see \ref types for the definition of the AnalysisItem data type
  -# the length of the result if the result is an array. This
    argument can be NULL.
  -# a success indicator. Failure can have obvious causes,
    such as breakdown of an internal routine, but the routine can also
    refuse to compute a quantity if doing so would be too expensive (see
    an example in the \ref normal category).

  A call to this routine need not
  involve actual computation: the requested quantity can already be attached
  to the matrix object (see \ref attached "attached quantities" for details).
  This mechanism is used in all the standard modules that come with the 
  AnaMod package.

  The workings of this function can be traced by specifying a trace function;
  see \ref tracing.
*/
PetscErrorCode ComputeQuantity
(AnaModNumericalProblem prob,const char *cat,const char *cmp,
 AnalysisItem *res,int *rreslen,PetscBool *success)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  if (!AnaModIsInitialized)
      SETERRQ(MPI_COMM_WORLD,1,"AnaMod not initialized: insert call to AnaModInitialize()\n");
  ierr = ComputeOrRetrieveQuantity
    (prob,cat,cmp,res,rreslen,success,MODE_COMPUTE); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "RetrieveQuantity"
/*! Retrieve an attached quantity. 
  Note that this does not report the length of arrays; you have to 
  know under which name this is stored.
*/
PetscErrorCode RetrieveQuantity
(AnaModNumericalProblem prob,const char *cat,const char *cmp,
 AnalysisItem *res,int *rreslen,PetscBool *success)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = ComputeOrRetrieveQuantity
    (prob,cat,cmp,res,rreslen,success,MODE_RETRIEVE); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
 
#undef __FUNCT__
#define __FUNCT__ "GetDataID"
 PetscErrorCode GetDataID(const char *cat,const char *cmp,int *id,PetscBool *f)
 {
   categoryobject catg; componentobject cmpt; PetscBool flg;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetCategory(cat,&catg,&flg); CHKERRQ(ierr);
  if (!flg) {
    if (!f)   SETERRQ(MPI_COMM_WORLD,1,"Could not find cat/cmp, but no flag to report this");
    *f = PETSC_FALSE; PetscFunctionReturn(0);
  }
  ierr = CategoryGetComponent(catg,cmp,&cmpt,&flg); CHKERRQ(ierr);
  if (!flg) {
    if (!f)   SETERRQ(MPI_COMM_WORLD,1,"Could not find cat/cmp, but no flag to report this");
    *f = PETSC_FALSE; PetscFunctionReturn(0);
  }
  ierr = ComponentGetId(cmpt,id); CHKERRQ(ierr);
  if (f) *f = PETSC_TRUE;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "GetDataType"
PetscErrorCode GetDataType
   (const char *cat,const char *cmp,AnalysisDataType *t,PetscBool *f)
{
  categoryobject catg; componentobject cmpt; PetscBool flg;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetCategory(cat,&catg,&flg); CHKERRQ(ierr);
  if (!flg) {
    if (!f)   SETERRQ(MPI_COMM_WORLD,1,"Could not find cat/cmp, but no flag to report this");
    *f = PETSC_FALSE; PetscFunctionReturn(0);
  }
  ierr = CategoryGetComponent(catg,cmp,&cmpt,&flg); CHKERRQ(ierr);
  if (!flg) {
    if (!f)   SETERRQ(MPI_COMM_WORLD,1,"Could not find cat/cmp, but no flag to report this");
    *f = PETSC_FALSE; PetscFunctionReturn(0);
  }
  ierr = ComponentGetType(cmpt,t); CHKERRQ(ierr);
  if (f) *f = PETSC_TRUE;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "HasQuantity"
/*!
  Check if a certain quantity is precomputed, meaning that it can 
  be retrieved (with a call to ComputeQuantity()) at no computational cost.

  The category and module names have to exist.
  Use HasComputeModule() to test whether
  a category and module is known to the system.

  See \ref attached "the page on attached quantities" for an explanation
  of the mechanism behind this routine.
*/
PetscErrorCode HasQuantity
(AnaModNumericalProblem prob,const char *cat,const char *cmp,PetscBool *f)
{
  categoryobject catg; componentobject cmpt;
  AnalysisDataType t; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetCategory(cat,&catg,f); CHKERRQ(ierr);
  if (!*f) PetscFunctionReturn(0);
  ierr = CategoryGetComponent(catg,cmp,&cmpt,f); CHKERRQ(ierr);
  if (!*f) PetscFunctionReturn(0);
  ierr = ComponentGetId(cmpt,&id); CHKERRQ(ierr);
  ierr = ComponentGetType(cmpt,&t); CHKERRQ(ierr);
  ierr = HasQuantityByID(prob,id,t,f); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "HasQuantityByID"
/*! Auxiliary routine with lookup by ID, which is much faster 
  than by string indexing.
*/
PetscErrorCode HasQuantityByID
(AnaModNumericalProblem prob,int id,AnalysisDataType type,PetscBool *f)
{
  Mat A = (Mat)prob; PetscReal rv,*rsv; int iv,*isv;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  switch (type) {
  case ANALYSISINTEGER :
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,iv,*f); CHKERRQ(ierr);
    break;
  case ANALYSISDOUBLE :
    ierr = PetscObjectComposedDataGetReal
      ((PetscObject)A,id,rv,*f); CHKERRQ(ierr);
    break;
  case ANALYSISINTARRAY :
    ierr = PetscObjectComposedDataGetIntstar
      ((PetscObject)A,id,isv,*f); CHKERRQ(ierr);
    break;
  case ANALYSISDBLARRAY :
    ierr = PetscObjectComposedDataGetRealstar
      ((PetscObject)A,id,rsv,*f); CHKERRQ(ierr);
    break;
  default : SETERRQ1(MPI_COMM_WORLD,1,"Unknown data type %d",type);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "RetrieveQuantityByID"
/*! See also HasQuantityByID() */
PetscErrorCode RetrieveQuantityByID
(AnaModNumericalProblem prob,int id,AnalysisDataType type,
 AnalysisItem *result,PetscBool *f)
{
  Mat A = (Mat)prob; PetscErrorCode ierr;
  PetscFunctionBegin;

  switch (type) {
  case ANALYSISINTEGER :
    ierr = PetscObjectComposedDataGetInt
      ((PetscObject)A,id,result->i,*f); CHKERRQ(ierr);
    break;
  case ANALYSISDOUBLE :
    ierr = PetscObjectComposedDataGetReal
      ((PetscObject)A,id,result->r,*f); CHKERRQ(ierr);
    break;
  case ANALYSISINTARRAY :
    ierr = PetscObjectComposedDataGetIntstar
      ((PetscObject)A,id,result->ii,*f); CHKERRQ(ierr);
    break;
  case ANALYSISDBLARRAY :
    ierr = PetscObjectComposedDataGetRealstar
      ((PetscObject)A,id,result->rr,*f); CHKERRQ(ierr);
    break;
  default : SETERRQ1(MPI_COMM_WORLD,1,"Unknown data type %d",type);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "QuantityAsString"
/*!
  Generate a character string for a given quantity.
*/
PetscErrorCode QuantityAsString(AnalysisItem *q,AnalysisDataType t,char **s)
{
  PetscErrorCode ierr; size_t l;

  PetscFunctionBegin;
  switch (t) {
  case ANALYSISINTEGER : 
    ierr = PetscMalloc(12*sizeof(char),s); CHKERRQ(ierr);
    ierr = PetscMemzero((void*)*s,12*sizeof(char)); CHKERRQ(ierr);
    sprintf((char*)*s,"%d",q->i);
    break;
  case ANALYSISDOUBLE : 
    ierr = PetscMalloc(22*sizeof(char),s); CHKERRQ(ierr);
    ierr = PetscMemzero((void*)*s,22*sizeof(char)); CHKERRQ(ierr);
    sprintf((char*)*s,"%e",q->r);
    break;
  case ANALYSISSTRING :
    ierr = PetscMalloc((strlen(q->c)+1)*sizeof(char),s); CHKERRQ(ierr);
    ierr = PetscMemzero((void*)*s,(strlen(q->c)+1)*sizeof(char)); CHKERRQ(ierr);
    sprintf((char*)*s,"%s",q->c);
    break;
  case ANALYSISINTARRAY :
    {
      int i,n,*iar;
      iar = q->ii;
      n = q->len;
      ierr = PetscMalloc((10*n+1)*sizeof(char),s); CHKERRQ(ierr);
      ierr = PetscMemzero((void*)*s,(10*n+1)*sizeof(char)); CHKERRQ(ierr);
      for (i=0; i<=n; i++) {
	ierr = PetscStrlen(*s,&l); CHKERRQ(ierr);
	sprintf((char*)*s+l,"%d,",iar[i]);
      }
    }
    break;
  case ANALYSISDBLARRAY :
    {
      int i,n; PetscReal *rar; 
      rar = q->rr;
      n = q->len;
      ierr = PetscMalloc((15*n+1)*sizeof(char),s); CHKERRQ(ierr);
      ierr = PetscMemzero((void*)*s,(15*n+1)*sizeof(char)); CHKERRQ(ierr);
      for (i=0; i<=n; i++) {
	ierr = PetscStrlen(*s,&l); CHKERRQ(ierr);
	sprintf((char*)*s+l,"%e,",rar[i]);
      }
    }
    break;
  default : SETERRQ1(MPI_COMM_WORLD,1,"Cannot string quantity type %d",t);
  }
  PetscFunctionReturn(0);
}

