/*! \file normal.c \ingroup categories 
  \brief Various estimates of the departure from normality

  \section normal Estimates for the departure from normality

  The departure from normality is a very hard to calculate quantity.
  This file provides several estimates.

  \section Usage

  Activate this module with:

  PetscErrorCode RegisterNormalityModules();

  Compute these elements with

  ComputeQuantity("normal",\<element\>,A,&res,&reslen,&flg);

  The auxiliary quantity of the Frobenius norm of the commutator 
  can be very expensive to calculate. If the bandwidth of the matrix is more
  then a specified times (default: 4) the square root of the matrix size, 
  this routine returns with failure. Set this tolerance with

  PetscErrorCode CommutatorNormFAllowSqrtTimes(int n);

  If this quantity is unavailable, so are the Lee96 bounds.

  Available elements are:
  - "trace-asquared" : an auxiliary quantity
  - "commutator-normF" : the Frobenius norm of the commutator \f$AA^t-A^tA\f$
  - "ruhe75-bound" : the bound from Ruhe[1975]
  - "lee95-bound" : the bound from Lee[1995]
  - "lee96-ubound" : the upper bound from Lee[1996]
  - "lee96-lbound" : the lower bound from Lee[1996]

  \section References

\verbatim
@article{ElPa:nonnormality,
author = {L. Elsner and M.H.C. Paardekoper},
title = {On Measures of Non-normality for Matrices},
journal = LAA,
volume = {307/308},
year = {1979},
pages = {107--124}
}

@article{Lee:upper-bound,
author = {Steven L. Lee},
title = {A Practical Upper Bound for Departure from Normality},
journal = SIMAT,
volume = {16},
issue = {2},
year = {1995},
pages = {462--468},
abstract = {Upper bound expressed in terms of Hermitian and Anti-Hermitian
    part; hence computable in $O(n^2)$ ops, as opposed to bounds by
    Henrici~\cite{Henrici:nonnormal-bounds}, which are based on $A^HA$ and
    take $O(n^3)$ ops.}
}

@article{Lee:available-bound,
author = {Steven L. Lee},
title = {Best Available Bounds for Departure from Normality},
journal = SIMAT,
volume = {17},
issue = {4},
year = {1996},
pages = {984--991}
\endverbatim
}

*/
#include <stdlib.h>
#include "anamod.h"
#include "anamodsalsamodules.h"
#include "anamatrix.h"
#include "petscerror.h"
#include "petscksp.h"
#include "petscmat.h"
#include <complex.h>

#undef __FUNCT__
#define __FUNCT__ "SparseVecProd"
static PetscErrorCode SparseVecProd
(int na,const int *ia,const PetscScalar *va,
 int nb,const int *ib,const PetscScalar *vb,
 PetscScalar *result)
{
  int pa = 0, pb = 0; PetscScalar v = 0;
  PetscFunctionBegin;

  while (pa<na && pb<nb) {
    if (ia[pa]==ib[pb]) {
      v += va[pa]*vb[pb]; pa++; pb++;
    } else if (ia[pa]<ib[pb]) pa++;
    else pb++;
  }
  *result = v;

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "MatCenter"
static PetscErrorCode MatCenter(Mat A)
{
  MPI_Comm comm; AnalysisItem q; PetscBool f;
  Vec D; PetscScalar trace,*d; int N,n,i; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
  ierr = MatGetLocalSize(A,&n,PETSC_NULL); CHKERRQ(ierr);
  ierr = MatGetSize(A,&N,PETSC_NULL); CHKERRQ(ierr);

  ierr = MatrixComputeQuantity
    (A,"simple","trace",&q,PETSC_NULL,&f); CHKERRQ(ierr);
  if (!f) 
    SETERRQ(MPI_COMM_WORLD,1,"Failed to compute trace");
  trace = q.r;

  ierr = VecCreateMPI(comm,n,PETSC_DECIDE,&D); CHKERRQ(ierr);
  ierr = MatGetDiagonal(A,D); CHKERRQ(ierr);
  ierr = VecGetArray(D,&d); CHKERRQ(ierr);
  PetscScalar ctrace = trace;
  ctrace /= N;
  for (i=0; i<n; i++) d[i] -= ctrace;
  ierr = VecRestoreArray(D,&d); CHKERRQ(ierr);
  ierr = MatDiagonalSet(A,D,INSERT_VALUES); CHKERRQ(ierr);
  ierr = VecDestroy(&D); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

/*
 * this routine needs to be rewritten for the complex case:
 * AA needs to be centred with only the Re of the trace,
 * A needs to be centred by the Im of the trace
 */
#undef __FUNCT__
#define __FUNCT__ "Lee95bounds"
static PetscErrorCode Lee95bounds(Mat A)
{
  Mat AA; AnalysisItem q;
  PetscReal mnorm,nnorm,bound; int id; 
  PetscBool f,has; PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = MatDuplicate(A,MAT_COPY_VALUES,&AA); CHKERRQ(ierr);
  ierr = MatCenter(AA); CHKERRQ(ierr);

  ierr = MatrixComputeQuantity
    (AA,"simple","symmetry-fsnorm",&q,PETSC_NULL,&f); CHKERRQ(ierr);
  if (!f) goto exit;
  mnorm = q.r;
  ierr = MatrixComputeQuantity
    (A,"simple","symmetry-fanorm",&q,PETSC_NULL,&f); CHKERRQ(ierr);
  if (!f) goto exit;
  nnorm = q.r;
  bound = sqrt(2) * PetscMin(mnorm,nnorm);

  ierr = GetDataID("normal","lee95-bound",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetReal
    ((PetscObject)A,id,bound); CHKERRQ(ierr);

 exit:
  ierr = MatDestroy(&AA); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DepartureLee95"
static PetscErrorCode DepartureLee95
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  PetscReal v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("normal","lee95-bound",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetReal
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (lv) *lv = 1;
  if (*flg) rv->r = v;
  else {
    ierr = Lee95bounds(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetReal
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->r = v;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "compute_tracea2_seq"
static PetscErrorCode compute_tracea2_seq(Mat A,PetscReal *trace)
{
  MPI_Comm comm;
  Mat B; int first,last,row,nacols,nbcols; const int *acols,*bcols;
  const PetscScalar *avals,*bvals; PetscReal vsum,v;
  PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
  ierr = MatGetOwnershipRange(A,&first,&last); CHKERRQ(ierr);
  ierr = MatTranspose(A,MAT_INITIAL_MATRIX,&B); CHKERRQ(ierr);
  vsum = 0;
  for (row=first; row<last; row++) {
    ierr = MatGetRow(A,row,&nacols,&acols,&avals); CHKERRQ(ierr);
    ierr = MatGetRow(B,row,&nbcols,&bcols,&bvals); CHKERRQ(ierr);
    ierr = SparseVecProd
      (nacols,acols,avals,nbcols,bcols,bvals,(PetscScalar*)&v); CHKERRQ(ierr);
    vsum += v;
    ierr = MatRestoreRow(A,row,&nacols,&acols,&avals); CHKERRQ(ierr);
    ierr = MatRestoreRow(B,row,&nbcols,&bcols,&bvals); CHKERRQ(ierr);
  }
  ierr = MatDestroy(&B); CHKERRQ(ierr);
  *trace = vsum;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "compute_tracea2"
static PetscErrorCode compute_tracea2(Mat A,PetscBool *done)
{
  MPI_Comm comm; Mat Awork; PetscReal trace; int id;
  PetscBool has,alloc_work,do_work,any_work; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
  ierr = AnaModGetSequentialMatrix
    (A,&Awork,&alloc_work,&do_work,&any_work); CHKERRQ(ierr);
  if (!any_work) {
    *done = PETSC_FALSE; PetscFunctionReturn(0);
  }
  if (do_work) {
    ierr = compute_tracea2_seq(Awork,&trace); CHKERRQ(ierr);
  }
  if (alloc_work) {
    ierr = MatDestroy(&Awork); CHKERRQ(ierr);
  }
  ierr = MPI_Bcast((void*)&trace,1,MPIU_REAL,0,comm);
  ierr = GetDataID("normal","trace-asquared",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetReal
    ((PetscObject)A,id,trace); CHKERRQ(ierr);
  *done = PETSC_TRUE;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TraceA2"
static PetscErrorCode TraceA2
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  PetscReal v=0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("normal","trace-asquared",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetReal
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (lv) *lv = 1;
  if (*flg) rv->r = v;
  else {
    ierr = compute_tracea2(A,flg); CHKERRQ(ierr);
    if (flg) {
      ierr = PetscObjectComposedDataGetReal
	((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
      if (*flg) rv->r = v;
    }
  }
  
  PetscFunctionReturn(0);
}

static int sqrt_times = 50;
/* This sets a bound on the bandwidth of a matrix; matrices wider than
   this bound will not have their Commutator norm computed.

   The bandwidth is specified as the number of multiples of the square root
   of the matrix size.
*/
PetscErrorCode CommutatorNormFAllowSqrtTimes(int n)
{
  sqrt_times = n;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "MatCommutatorNormF_seq"
/*!
  Compute the Frobenius norm of the commutator
  \latexonly $AA^t-A^t$ \endlatexonly
  of a sequential matrix.

  This routine is accessed through MatCommutatorNormF().
*/
static PetscErrorCode MatCommutatorNormF_seq
(Mat A,PetscScalar *result,PetscBool *done)
{
  Mat B,E,F;
  PetscScalar vsum; PetscBool force,flg;
  int afirst,alast,bfirst,blast,p,q,N;
  PetscErrorCode ierr;

  PetscFunctionBegin;

  ierr = MatGetSize(A,&N,PETSC_NULL); CHKERRQ(ierr);

  ierr = AnaModHasForcedExpensiveComputation(&force); CHKERRQ(ierr);
  ierr = MatrixComputeQuantity
    (A,"structure","left-bandwidth",(AnalysisItem*)&p,PETSC_NULL,&flg); CHKERRQ(ierr);
  if (!flg) p = N;
  ierr = MatrixComputeQuantity
    (A,"structure","right-bandwidth",(AnalysisItem*)&q,PETSC_NULL,&flg); CHKERRQ(ierr);
  if (!flg) q = N;
  if (p+q>sqrt_times*sqrt(N) && !force) {
    /*printf("too expensive\n");*/
    *done = PETSC_FALSE;
    PetscFunctionReturn(0);
  }

  /* set up matrix and transpose */
  ierr = MatDuplicate(A,MAT_COPY_VALUES,&B); CHKERRQ(ierr);
  afirst = 0; alast = N; bfirst = 0; blast = N;
  ierr = MatTranspose(A,MAT_INITIAL_MATRIX,&E); CHKERRQ(ierr);
  ierr = MatDuplicate(E,MAT_COPY_VALUES,&F); CHKERRQ(ierr);
#define min(a,b) (a<b?a:b)
#define max(a,b) (a>b?a:b)
  {
    const PetscReal *avals,*bvals,*evals,*fvals; PetscScalar v1,v2;
    int a,b, nacols,nbcols, necols,nfcols;
    const int *acols,*bcols,*ecols,*fcols;
    vsum = 0;
    for (a=afirst; a<alast; a++) {
      int bmin=max(bfirst,a-(p+q)),bmax=min(blast,a+(p+q));
      ierr = MatGetRow(A,a,&nacols,&acols,(const PetscScalar**)&avals); CHKERRQ(ierr);
      ierr = MatGetRow(E,a,&necols,&ecols,(const PetscScalar**)&evals); CHKERRQ(ierr);
      for (b=bmin; b<bmax; b++) {
	ierr = MatGetRow(B,b,&nbcols,&bcols,(const PetscScalar**)&bvals); CHKERRQ(ierr);
	ierr = MatGetRow(F,b,&nfcols,&fcols,(const PetscScalar**)&fvals); CHKERRQ(ierr);
	ierr = SparseVecProd
	  (nacols,acols,(const PetscScalar*)avals,nbcols,bcols,(const PetscScalar*)bvals,&v1); CHKERRQ(ierr);
	ierr = SparseVecProd
	  (necols,ecols,(const PetscScalar*)evals,nfcols,fcols,(const PetscScalar*)fvals,&v2); CHKERRQ(ierr);
	v1 -= v2;
	v1 = PetscAbsScalar(v1);
	vsum += v1*v1;
	ierr = MatRestoreRow(B,b,&nbcols,&bcols,(const PetscScalar**)&bvals); CHKERRQ(ierr);
	ierr = MatRestoreRow(F,b,&nfcols,&fcols,(const PetscScalar**)&fvals); CHKERRQ(ierr);
      }
      ierr = MatRestoreRow(A,a,&nacols,&acols,(const PetscScalar**)&avals); CHKERRQ(ierr);
      ierr = MatRestoreRow(E,a,&necols,&ecols,(const PetscScalar**)&evals); CHKERRQ(ierr);
    }
  }
  ierr = MatDestroy(&B); CHKERRQ(ierr);
  ierr = MatDestroy(&E); CHKERRQ(ierr);
  ierr = MatDestroy(&F); CHKERRQ(ierr);
  *result = sqrt(vsum); *done = PETSC_TRUE;

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "MatCommutatorNormF"
/*!
  Compute the Frobenius norm of the commutator
  \latexonly $AA^t-A^t$ \endlatexonly

  This computation can only be done in single-processor mode.
  If the commandline
  option (see \ref options) \c "-anamod-force sequential" is specified,
  the matrix is gathered on processor zero to compute this quantity.

  This is an expensive operation \latexonly ($O(N^3)$) \endlatexonly
  which can only be optimized for banded matrices. The maximum allowed 
  bandwidth is set through CommutatorNormFAllowSqrtTimes().

  See MatCommutatorNormF_seq() for the actual computation.
*/
static PetscErrorCode MatCommutatorNormF(Mat A,PetscBool *done)
{
  MPI_Comm comm; Mat Awork; PetscScalar result;
  PetscBool alloc_work,do_work,any_work,has;
  int id; PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
  ierr = AnaModGetSequentialMatrix
    (A,&Awork,&alloc_work,&do_work,&any_work); CHKERRQ(ierr);
  if (!any_work) {
    *done = PETSC_FALSE; PetscFunctionReturn(0);
  }
  if (do_work) {
    ierr = MatCommutatorNormF_seq(Awork,&result,done); CHKERRQ(ierr);
    /*printf("result %d:%e\n",*done,result);*/
  }
  if (alloc_work) {
    ierr = MatDestroy(&Awork); CHKERRQ(ierr);
  }

  ierr = MPI_Bcast((void*)done,1,MPI_INT,0,comm);
  if (*done) {
    ierr = MPI_Bcast((void*)&result,1,MPIU_REAL,0,comm);
  } else PetscFunctionReturn(0);

  ierr = GetDataID("normal","commutator-normF",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetReal
    ((PetscObject)A,id,result.real()); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "Commutator"
static PetscErrorCode Commutator
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  PetscReal v = 0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("normal","commutator-normF",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetReal
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (lv) *lv = 1;
  if (*flg) rv->r = v;
  else {
    ierr = MatCommutatorNormF(A,flg); CHKERRQ(ierr);
    if (*flg) {
      ierr = PetscObjectComposedDataGetReal
	((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
      if (*flg) rv->r = v;
    }
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "Lee96bounds"
static PetscErrorCode Lee96bounds(Mat A)
{
  Mat AA; PetscReal upper,lower;
  PetscScalar tracea2; PetscReal cnorm,fnorm; int id,ql; 
  PetscBool f,has; PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = MatDuplicate(A,MAT_COPY_VALUES,&AA); CHKERRQ(ierr);
  ierr = MatCenter(AA); CHKERRQ(ierr);
  ierr = TraceA2(AA,(AnalysisItem*)&tracea2,&ql,&f); CHKERRQ(ierr);
  if (!f) {
    /*printf(">>warning: did not compute trace of A squared\n");*/
    goto exit;
  }
  ierr = MatNorm(AA,NORM_FROBENIUS,&fnorm); CHKERRQ(ierr);

  upper = fnorm*fnorm-PetscAbsScalar(tracea2);
  ierr = Commutator(A,(AnalysisItem*)&cnorm,&ql,&f); CHKERRQ(ierr);
  if (!f) {
    /*printf(">>warning: did not compute commutator Frobenius norm\n");*/
    goto exit;
  }
  fnorm *= fnorm;
  lower = sqrt(fnorm - sqrt(fnorm*fnorm-cnorm*cnorm/2));

  ierr = GetDataID("normal","lee96-ubound",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetReal
    ((PetscObject)A,id,upper); CHKERRQ(ierr);
  ierr = GetDataID("normal","lee96-lbound",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataSetReal
    ((PetscObject)A,id,lower); CHKERRQ(ierr);

 exit:
  ierr = MatDestroy(&AA); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DepartureLee96U"
static PetscErrorCode DepartureLee96U
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  PetscReal v=0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("normal","lee96-ubound",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetReal
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (lv) *lv = 1;
  if (*flg) rv->r = v;
  else {
    ierr = Lee96bounds(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetReal
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->r = v;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DepartureLee96L"
static PetscErrorCode DepartureLee96L
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  PetscReal v=0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("normal","lee96-lbound",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetReal
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (lv) *lv = 1;
  if (*flg) rv->r = v;
  else {
    ierr = Lee96bounds(A); CHKERRQ(ierr);
    ierr = PetscObjectComposedDataGetReal
      ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
    if (*flg) rv->r = v;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DepartureRuhe75"
static PetscErrorCode DepartureRuhe75
(AnaModNumericalProblem prob,AnalysisItem *rv,int *lv,PetscBool *flg)
{
  Mat A = (Mat)prob;
  PetscReal v=0; PetscBool has; int id; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = GetDataID("normal","ruhe75-bound",&id,&has); CHKERRQ(ierr);
  HASTOEXIST(has);
  ierr = PetscObjectComposedDataGetReal
    ((PetscObject)A,id,v,*flg); CHKERRQ(ierr);
  if (lv) *lv = 1;
  if (*flg) rv->r = v;
  else {
    AnalysisItem q; PetscReal s,er,ec,dep=0; int ql;

    ierr = ComputeQuantity
      (prob,"spectrum","sigma-min",(AnalysisItem*)&s,&ql,flg); CHKERRQ(ierr);
    if (!*flg) PetscFunctionReturn(0);
    ierr = ComputeQuantity
      (prob,"spectrum","lambda-min-by-magnitude-re",&q,&ql,flg); CHKERRQ(ierr);
    if (!*flg) PetscFunctionReturn(0);
    er = q.r;
    ierr = ComputeQuantity
      (prob,"spectrum","lambda-min-by-magnitude-im",&q,&ql,flg); CHKERRQ(ierr);
    if (!*flg) PetscFunctionReturn(0);
    ec = q.r;
    dep = PetscAbsScalar(s-sqrt(er*er+ec*ec));

    ierr = ComputeQuantity
      (prob,"spectrum","sigma-max",&q,&ql,flg); CHKERRQ(ierr);
    if (!*flg) PetscFunctionReturn(0);
    s = q.r;
    ierr = ComputeQuantity
      (prob,"spectrum","lambda-max-by-magnitude-re",&q,&ql,flg); CHKERRQ(ierr);
    if (!*flg) PetscFunctionReturn(0);
    er = q.r;
    ierr = ComputeQuantity
      (prob,"spectrum","lambda-max-by-magnitude-im",&q,&ql,flg); CHKERRQ(ierr);
    if (!*flg) PetscFunctionReturn(0);
    ec = q.r;
    dep = PetscMax(dep,PetscAbsScalar(s-sqrt(er*er+ec*ec)));

    ierr = PetscObjectComposedDataSetReal
      ((PetscObject)A,id,dep); CHKERRQ(ierr);
    rv->r = dep;
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "RegisterNormalityModules"
PetscErrorCode RegisterNormalityModules(void)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = RegisterModule
    ("normal","trace-asquared",ANALYSISDOUBLE,&TraceA2); CHKERRQ(ierr);
  ierr = RegisterModule
    ("normal","commutator-normF",ANALYSISDOUBLE,&Commutator); CHKERRQ(ierr);

  ierr = RegisterModule
    ("normal","ruhe75-bound",ANALYSISDOUBLE,&DepartureRuhe75); CHKERRQ(ierr);

  ierr = RegisterModule
    ("normal","lee95-bound",ANALYSISDOUBLE,&DepartureLee95); CHKERRQ(ierr);

  ierr = RegisterModule
    ("normal","lee96-ubound",ANALYSISDOUBLE,&DepartureLee96U); CHKERRQ(ierr);
  ierr = RegisterModule
    ("normal","lee96-lbound",ANALYSISDOUBLE,&DepartureLee96L); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}
