/*! \mainpage NMD: the Numerical MetaData library

  The NMD library serves to store two-level data structures of
  numerical metadata. While this library was intended to be used with
  AnaMod in the Salsa project, it can be used on its own. Unlike other
  libraries in the Salsa project, it does not rely on Petsc.

  \section general General notes

  All routines in NMD return an error code of type \c NMDErrorCode. This
  is zero for success and anything else for failure. Return codes can be
  caught with \c NMD_ERR_RETURN(ierr) and generate with 
  \c NMD_ERR_REPORT(msg).

  \subpage objectfunctions

  \subpage categoryfunctions

  \subpage componentfunctions

  \subpage values

  \subpage viewing

  \subpage installation

  \author Victor Eijkhout
  \version 2.5
  \date unreleased

  \section Change history
  2.5 changes
  - added routines: NMDUnsetValue(),NMDCategoryGetComponents(),
  NMDTypeGetMySQLName()
  - NMDReportObject() now takes delimiter arguments. This is useful
  for generating MySQL strings and such.

  2.4 changes
  - minor

  2.3 changes

  - NMDGetValue now returns failure (instead of aborting) 
  if cat/cmp do not exist
  - removed occurrences of Petsc's CHKERRQ
  - introduced NMDErrorCode
  - new files nmdcat.c nmdcmp.c
  - NMDGetValue and GetArrayValue now use NMDTruth
  - NMDSetValue and NMDSetArrayValue are now analogous: use ampersand 
  for all types of data
  - NMDTrue and NMDFalse instead of 1 and 0
  - unit tests added

  2.3 bug fixes

  - NMDGetValue missing case of string value added
  - NMDTryGetCategory (and various other Get routines) were able to 
  find non-existing category names. Fixed.
  - lots of memory leaks plugged

  2.2
  - Completely revamped array handling; watch out for prototype changes
  - CFLAGS is now NMD_CFLAGS
 */

/*! \page installation Installing NMD

  Installing NMD takes the following steps:
  - edit the Make.inc file for:
    - compiler options, and settings for your \c ar and \c ranlib program
    - add \c "-DNMD_HAVE_PETSC" to the compile line if you are using NMD
    with Petsc (see NMDReportObject())
    - \c NMD_LIB_DIR is the location where the library will be installed
  - do \c "make install" to generate the binaries

 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "nmd_impl.h"
#include "nmd.h"

const char *typenames[] =
  {"invalid","int","real","string","intarray","realarray"};
const char *mysqltypenames[] =
  {"VARCHAR(256)","INTEGER","DOUBLE","VARCHAR(1024)",
   "VARCHAR(1024)","VARCHAR(1024)"};
static const int nnmdtypenames = 6;

#define CATCHUNK 10

/*! \page objectfunctions Metadata object manipulation

  Top level functions for manipulation metadata objects.

  See NMDCreateObject(), NMDDestroyObject(), NMDBuildObjectStructure(),
  NMDCloneObject(), NMDCloneObjectStructure(), NMDViewObject()
 */

#undef __FUNCT__
#define __FUNCT__ "NMDCreateObject"
/*!
  This routine create an NMD_metadata object, and allocates enough 
  space in it for 10 categories of 20 elements each. Currently this can not
  be reallocated. In the future we want to be a bit more flexible.
*/
NMDErrorCode NMDCreateObject(NMD_metadata *obj)
{
  NMD_metadata nnew;

  /* allocate the metadata object */
  NMD_MALLOC(nnew,1,struct NMD_metadata_,"object");
  nnew->cookie = NMDCOOKIE;
  *obj = nnew;

  /* give it a basic array of categories */
  nnew->alloc = CATCHUNK; nnew->ncat = 0;
  NMD_MALLOC(nnew->cats,nnew->alloc,NMD_metadata_category,"categories");

  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDDestroyObject"
/*!
  Deallocate all the data in a metadata object.
*/
NMDErrorCode NMDDestroyObject(NMD_metadata obj)
{
  int i,j; NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(obj);
  for (i=0; i<obj->ncat; i++) {
    NMD_metadata_category cat = obj->cats[i];
    for (j=0; j<cat->ncmp; j++) {
      NMD_metadata_item cmp = cat->cmps[j];
      ierr = NMDComponentDestroy(cmp); NMD_ERR_RETURN(ierr);
    }
    NMD_FREE(cat->cmps);
    NMD_FREE(cat->name);
    NMD_FREE(cat);
    CHKMEMQ;
  }
  NMD_FREE(obj->cats);
  NMD_FREE(obj);

  return 0;
}

#if 0
#undef __FUNCT__
#define __FUNCT__ "NMDBuildObjectStructure"
NMDErrorCode NMDBuildObjectStructure(NMD_metadata nmd)
{
  char **categories; int ncat,icat; NMDErrorCode ierr;

  ierr = NMDGetCategories(nmd,&ncat,&categories); NMD_ERR_RETURN(ierr);
  for (icat=0; icat<ncat; icat++) {   
    NMDDataType *t; char **modules; int nmod,imod;
    ierr = NMDObjectGetOrCreateCategory(nmd,categories[icat]); NMD_ERR_RETURN(ierr);
    ierr = CategoryGetModules
      (categories[icat],&modules,&t,PETSC_NULL,&nmod); NMD_ERR_RETURN(ierr);
    for (imod=0; imod<nmod; imod++) {
      ierr = NMDCreateComponent
	(nmd,categories[icat],modules[imod],t[imod]); NMD_ERR_RETURN(ierr);
    }
  }
  return 0;
}
#endif

#undef __FUNCT__
#define __FUNCT__ "NMDCloneObjectStructure"
/*!
  This routine creates an NMD_metadata object, and fills it in with 
  the categories and components of a template object. Data is not copied; 
  for that, see NMDCloneObject() and NMDCopyCategory().
*/
NMDErrorCode NMDCloneObjectStructure(NMD_metadata old,NMD_metadata *ret)
{
  NMD_metadata nnew;
  int icat,icmp; NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(old);
  ierr = NMDCreateObject(&nnew); NMD_ERR_RETURN(ierr);
  CHECKHASNMDCOOKIE(nnew);
  for (icat=0; icat<old->ncat; icat++) {
    NMD_metadata_category oldcat = old->cats[icat];
    CHECKHASNMDCOOKIE(oldcat);
    for (icmp=0; icmp<oldcat->ncmp; icmp++) {
      NMD_metadata_item oldcmp = oldcat->cmps[icmp];
      CHECKHASNMDCOOKIE(oldcmp);
      ierr = NMDObjectEnsureCategoryComponent
	(nnew,oldcat->name,oldcmp->name,oldcmp->t,NULL); NMD_ERR_RETURN(ierr);
    }
  }
  CHKMEMQ;
  *ret = nnew;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDCloneObject"
/*!
  Given an already created NMD_metadata object, fill it with 
  the data of a template object. See also NMDCloneObjectStructure().
*/
NMDErrorCode NMDCloneObject(NMD_metadata old,NMD_metadata nnew)
{
  int icat; NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(old);
  CHECKHASNMDCOOKIE(nnew);
  for (icat=0; icat<old->ncat; icat++) {
    /* loop over input categories by index; output categories
       are retrieved by label, or newly created */
    NMD_metadata_category oldcat = old->cats[icat],newcat; int icmp;
    ierr = NMDObjectGetOrCreateCategory
      (nnew,oldcat->name,&newcat); NMD_ERR_RETURN(ierr);
    CHECKHASNMDCOOKIE(oldcat);
    for (icmp=0; icmp<oldcat->ncmp; icmp++) {
      NMD_metadata_item oldcmp = oldcat->cmps[icmp],newcmp;
      ierr = NMDCategoryGetOrCreateComponent
	(newcat,oldcmp->name,oldcmp->t,&newcmp); NMD_ERR_RETURN(ierr);
      ierr = NMDCopyItemValues(oldcmp,newcmp); NMD_ERR_RETURN(ierr);
    }
  }
  CHKMEMQ;
  return 0;
}

/*! \page viewing Viewing objects
  NMDViewObject() gives informal output; NMDReportObject() can be
  used for database records and such.
 */
#undef __FUNCT__
#define __FUNCT__ "NMDViewObject"
/*! Print out an NMD object.

  Currently only int, real, string fields are displayed,
  others are displayed as \c "***".
 */
NMDErrorCode NMDViewObject(NMD_metadata obj)
{
  int icat,icmp;
  CHECKHASNMDCOOKIE(obj);
  printf("========\n");
  for (icat=0; icat<obj->ncat; icat++) {
    NMD_metadata_category cat = obj->cats[icat];
    CHECKHASNMDCOOKIE(cat);
    printf("Category: <%s>\n",cat->name);
    for (icmp=0; icmp<cat->ncmp; icmp++) {
      NMD_metadata_item cmp = cat->cmps[icmp];
      CHECKHASNMDCOOKIE(cmp);
      printf("  <%s> : ",cmp->name);
      if (!cmp->set) {
	printf("(null)\n");
      } else {
	switch (cmp->t) {
	case NMDInt : printf("<%d>\n",cmp->i); break;
	case NMDReal : printf("<%e>\n",cmp->r); break;
	case NMDString : printf("<<%s>>\n",cmp->cc); break;
	default : printf("****\n"); break;
	}
      }
    }
  }
  printf("========\n\n");
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDReportObject"
/*! Generate a delimited representation of a metadata object.

  The returned strings are allocated in this routine and it is the user's
  responsibility to free them with NMD_FREE().

  Arguments:
  - obj : the metadata object
  - ar : boolean to indicate whether arrays need to be written out in full.
  If this is false, only the first and last couple of elements are given.
  - rkey : a string containing the names of the metadata items
  - rval : the metadata items
  - delim : delimiter character used in \c rkey and \c rval
  - itemdelim1 : an optional opening quote, used for both keys and values.
  (A NULL value will cause no delimiter to be printed, rather than a null
  character.)
  For instance, use the backquote when generating MySQL strings.
  - itemdelim2 : an optional closing quote

 */
NMDErrorCode NMDReportObject
(NMD_metadata obj,NMDTruth arrays,const char **rkey,const char **rval,
 const char delim,const char itemdelim1,const char itemdelim2)
{
  int icat,icmp,write,keylen=0,vallen=0; char *key=NULL,*val=NULL, *scratch;
  CHECKHASNMDCOOKIE(obj);

  /*
   * allocate and fill in the key and value strings
   */
#define CHKLEN 500
  NMD_MALLOC(scratch,CHKLEN,char,"scratch pad");
#define CHKSPACEFOR(ar,arlen,fmt,str) {int nr; memset(scratch,0,CHKLEN); sprintf(scratch,fmt,str); nr=strlen(scratch); if (write==0) {arlen = arlen+nr+2; } else {int l = strlen(ar); if (l+nr<arlen) {sprintf(ar+l,fmt,str);} else SETERRQ(MPI_COMM_WORLD,1,"sprintf would overflow allocated buffer");}}
  write = 0;
 writeloop:
  if (write==0) {
    keylen=vallen=0;
  } else {
    if (rkey) NMD_MALLOC(key,keylen,char,"names array");
    if (rval) NMD_MALLOC(val,vallen,char,"values array");
  }
  for (icat=0; icat<obj->ncat; icat++) {
    NMD_metadata_category cat = obj->cats[icat];
    CHECKHASNMDCOOKIE(cat);
    for (icmp=0; icmp<cat->ncmp; icmp++) {
      NMD_metadata_item cmp = cat->cmps[icmp];
      CHECKHASNMDCOOKIE(cmp);
      if (rkey) {
	if (itemdelim1) CHKSPACEFOR(key,keylen,"%c",itemdelim1);
	CHKSPACEFOR(key,keylen,"%s:",cat->name);
	CHKSPACEFOR(key,keylen,"%s",cmp->name);
	if (itemdelim2) CHKSPACEFOR(key,keylen,"%c",itemdelim2);
	CHKSPACEFOR(key,keylen,"%c",delim);
      }
      if (rval) {
	if (itemdelim1) CHKSPACEFOR(val,vallen,"%c",itemdelim1);
	if (!cmp->set) goto delimiter;
	switch (cmp->t) {
	case NMDInt : {
	  CHKSPACEFOR(val,vallen,"%8d",cmp->i); break;}
	case NMDReal : {
	  CHKSPACEFOR(val,vallen,"%15.8e",cmp->r); break;}
	case NMDString :
	  if (!cmp->cc) NMD_ERR_REPORT("Null string");
	  if (strlen(cmp->cc)>0) {
	    if (strlen(cmp->cc)<=CHKLEN) {
	      CHKSPACEFOR(val,vallen,"%s",cmp->cc);
	    } else {CHKSPACEFOR(val,vallen,"%s","stringtoolong");}
	  } else {
	    CHKSPACEFOR(val,vallen,"%s","null"); 
	  }
	  break;
	case NMDIntarray : {
	  struct NMD_intarray_struct *arraystruct =
	    (struct NMD_intarray_struct *)(cmp->ii);
	  int n = arraystruct->length,i;
	  if (arrays || n<=4) {
	    for (i=0; i<n; i++) {
	      if (i<n-1) {
		CHKSPACEFOR(val,vallen,"%8d,",arraystruct->data[i]);
	      } else {
		CHKSPACEFOR(val,vallen,"%8d",arraystruct->data[i]);
	      }
	    }
	  } else {
	    CHKSPACEFOR(val,vallen,"%8d,",arraystruct->data[0]);
	    CHKSPACEFOR(val,vallen,"%8d,",arraystruct->data[1]);
	    CHKSPACEFOR(val,vallen,"%8d,",arraystruct->data[n-2]);
	    CHKSPACEFOR(val,vallen,"%8d",arraystruct->data[n-1]);
	  }
	} ; break;
	case NMDRealarray : {
	  struct NMD_realarray_struct *arraystruct =
	    (struct NMD_realarray_struct *)(cmp->rr);
	  int n = arraystruct->length,i;
	  if (arrays || n<=4) {
	    for (i=0; i<n; i++) {
	      if (i<n-1) {
		CHKSPACEFOR(val,vallen,"%11.4e,",arraystruct->data[i]);
	      } else {
		CHKSPACEFOR(val,vallen,"%11.4e",arraystruct->data[i]);
	      }
	    }
	  } else {
	    CHKSPACEFOR(val,vallen,"%11.4e,",arraystruct->data[0]);
	    CHKSPACEFOR(val,vallen,"%11.4e,",arraystruct->data[1]);
	    CHKSPACEFOR(val,vallen,"%11.4e,",arraystruct->data[n-2]);
	    CHKSPACEFOR(val,vallen,"%11.4e",arraystruct->data[n-1]);
	  }
	} ; break;
	default :
	  CHKSPACEFOR(val,vallen,"%s","==="); break;
	}
      delimiter:
	if (itemdelim2) CHKSPACEFOR(val,vallen,"%c",itemdelim2);
	CHKSPACEFOR(val,vallen,"%c",delim);
      }
    }
  }
  write += 1;
  if (write<=1) goto writeloop;
  NMD_FREE(scratch);
  CHKMEMQ;
  /* *rkey = key; *rval = val; */
  if (key) {
    int l = strlen(key); 
    if (l>keylen)
      NMD_ERR_REPORT("key string overrun");
    *rkey = key;
  }
  if (val) {
    int l = strlen(val); 
    if (l>vallen)
      NMD_ERR_REPORTii("val string overrun; alloc/actual",vallen,l);
    *rval = val;
  }

  return 0;
}

/*! \page values Value handling

  NMD handles scalar and array values slightly differently.
  For scalar values use NMDSetValue() and NMDGetValue();
  for array values use NMDSetArrayValue(), NMDGetArrayValue(),
  NMDCopyArrayValue(). The array routines take an extra parameter 
  denoting the length of the array.

  Note that scalar values have to be passed by reference:
\code
  int i,*ii;
  NMDSetValue(nmd,....,&i);        // use an ampersand here
  NMDSetValue(nmd,....,ii,length); // no ampersand needed here!
  NMDGetValue(nmd,....,&i);
  NMDGetValue(nmd,....,&ii,&length);
\endcode

  Furthermore, see NMDCopyItemValues(), NMDGetDataType(),
  NMDIsArrayType().
 */

#undef __FUNCT__
#define __FUNCT__ "NMDSetValue"
/*! Set a metadata value, indexed by category and component name.

  The value has to be passed by reference

  String values are copied. (Reason: literal strings are treated
  differently from allocated, and Petsc has its own way of doing strings.)

  This call can be used to create categories and components; there is
  no checking of slight misspellings.

  See also \ref values.
*/
NMDErrorCode NMDSetValue(NMD_metadata obj,const char *cat,const char *cmp,NMDDataType t,void *v)
{
  NMD_metadata_category ctg; NMD_metadata_item cpt; NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectGetOrCreateCategory(obj,(char*)cat,&ctg); NMD_ERR_RETURN(ierr);
  ierr = NMDCategoryGetOrCreateComponent(ctg,(char*)cmp,t,&cpt); NMD_ERR_RETURN(ierr);
  ierr = NMDComponentSetValue(cpt,t,v); NMD_ERR_RETURN(ierr);
  CHKMEMQ;

  return 0;
}

NMDErrorCode NMDUnsetValue(NMD_metadata obj,const char *cat,const char *cmp)
{
  NMD_metadata_category ctg; NMD_metadata_item cpt; 
  NMDTruth flg; NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectTryGetCategory(obj,cat,&ctg,&flg); NMD_ERR_RETURN(ierr);
  if (!flg) return 0;
  ierr = NMDCategoryTryGetComponent(ctg,cmp,&cpt,&flg); NMD_ERR_RETURN(ierr);
  if (!flg) return 0;
  ierr = NMDComponentUnsetValue(cpt); NMD_ERR_RETURN(ierr);
  CHKMEMQ;

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDSetArrayValue"
/*! Set a metadata value, if it is an array type.

  The arrays are not copied, so the user is responsible for freeing the array. 
  Use NMDCopyArrayValue() to have the array copied; NMD will then free the 
  array when the metadata object is freed.

  This call can be used to create categories and components; there is
  no checking of slight misspellings.

  See also \ref values.
*/
NMDErrorCode NMDSetArrayValue
(NMD_metadata obj,const char *cat,const char *cmp,NMDDataType t,void *v,int l)
{
  NMD_metadata_category ctg; NMD_metadata_item cpt; NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectGetOrCreateCategory(obj,(char*)cat,&ctg); NMD_ERR_RETURN(ierr);
  ierr = NMDCategoryGetOrCreateComponent(ctg,(char*)cmp,t,&cpt); NMD_ERR_RETURN(ierr);
  ierr = NMDComponentSetArrayValue(cpt,t,v,l); NMD_ERR_RETURN(ierr);
  CHKMEMQ;

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDCopyArrayValue"
/*! Set a metadata array value; the user array is
  copied.

  This call can be used to create categories and components; there is
  no checking of slight misspellings.

  See also \ref values.
*/
NMDErrorCode NMDCopyArrayValue
(NMD_metadata obj,const char *cat,const char *cmp,NMDDataType t,void *v,int l)
{
  NMD_metadata_category ctg; NMD_metadata_item cpt; NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectGetOrCreateCategory(obj,(char*)cat,&ctg); NMD_ERR_RETURN(ierr);
  ierr = NMDCategoryGetOrCreateComponent(ctg,(char*)cmp,t,&cpt); NMD_ERR_RETURN(ierr);

  CHECKHASNMDCOOKIE(cpt);
  cpt->set = NMDTrue;
  switch (cpt->t) {
  case NMDInt :
  case NMDReal :
  case NMDString :
    ierr = NMDSetValue(obj,cat,cmp,t,v); NMD_ERR_RETURN(ierr);
    break;
  case NMDIntarray : {
    struct NMD_intarray_struct* iistruct; int *ii,i;
    NMD_MALLOC(iistruct,1,struct NMD_intarray_struct,"int array struct");
    iistruct->unique = 1; iistruct->length = l; 
    NMD_MALLOC(ii,l,sizeof(int),"copied int array");
    for (i=0; i<l; i++) ii[i] = (*(int**)v)[i];
    iistruct->data = ii;
    cpt->ii = iistruct;
    break;}
  case NMDRealarray : {
    struct NMD_realarray_struct* rrstruct; NMDRealtype *rr; int i;
    NMD_MALLOC(rrstruct,1,struct NMD_realarray_struct,"real array struct");
    rrstruct->unique = 1; rrstruct->length = l;
    NMD_MALLOC(rr,l,NMDRealtype,"copied real array");
    for (i=0; i<l; i++) rr[i] = (*(NMDRealtype**)v)[i];
    rrstruct->data = rr;
    cpt->rr = rrstruct;
    break;}
  default : NMD_ERR_REPORTi("Unknown type",(int)t); break;
  }
  CHKMEMQ;

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDGetValue"
/*! Retrieve a stored scalar value.
   If no value has been stored under the specified category and component,
   a zero flag is returned. The flag parameter can be null.

   Null pointers can be passed for the datatype or value, for instance to
   test only for the existence of a set value.

  See also \ref values.
*/
NMDErrorCode NMDGetValue
(NMD_metadata obj,const char *cat,const char *cmp,NMDDataType *t,void *v,NMDTruth *f)
{
  NMD_metadata_category ctg; NMD_metadata_item cpt; 
  NMDTruth ff; NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectTryGetCategory(obj,cat,&ctg,&ff); NMD_ERR_RETURN(ierr);
  if (!ff) {
    if (f) {
      *f = NMDFalse; return 0;
    } else NMD_ERR_REPORT
	     ("Could not find category, but no flag given to report this");
  }
  CHECKHASNMDCOOKIE(ctg);
  ierr = NMDCategoryTryGetComponent(ctg,cmp,&cpt,&ff); NMD_ERR_RETURN(ierr);
  if (!ff) {
    if (f) {
      *f = NMDFalse; return 0;
    } else NMD_ERR_REPORT
	     ("Could not find component, but no flag given to report this");
  }
  CHECKHASNMDCOOKIE(cpt);
  if (!cpt->set) {
    if (f) {
      *f = NMDFalse; return 0;
    } else NMD_ERR_REPORT
	     ("Value not set, but no flag given to report this");
  }
  if (t) *t = cpt->t;
  switch (cpt->t) {
  case NMDInt :
    if (v) *(int*)v = cpt->i;
    if (f) *f = NMDTrue; break;
  case NMDReal :
    if (v) *(double*)v = cpt->r;
    if (f) *f = NMDTrue; break;
  case NMDString :
    if (v) *(char**)v = cpt->cc;
    if (f) *f = NMDTrue; break;
  case NMDIntarray :
  case NMDRealarray :
    NMD_ERR_REPORT("Please use NMDGetArrayValue"); break;
  default :
    if (f) *f = NMDFalse;
  }
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDGetArrayValue"
/*! Retrieve a stored value.
   If no value has been stored under the specified category and component,
   a zero flag is returned. The flag parameter can be null.

   Null pointers can be passed for the datatype or value, for instance to
   test only for the existence of a set value.

  See also \ref values.
*/
NMDErrorCode NMDGetArrayValue
(NMD_metadata obj,const char *cat,const char *cmp,
 NMDDataType *t,void *v,int *len,NMDTruth *f)
{
  NMD_metadata_category ctg; NMD_metadata_item cpt; NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectGetCategory(obj,cat,&ctg); NMD_ERR_RETURN(ierr);
  CHECKHASNMDCOOKIE(ctg);
  ierr = NMDCategoryGetComponent(ctg,cmp,&cpt); NMD_ERR_RETURN(ierr);
  CHECKHASNMDCOOKIE(cpt);
  if (!cpt->set) {
    if (f) {
      *f = NMDFalse; return 0;
    } else NMD_ERR_REPORT
	     ("Value not set, but no flag given to report this");
  }
  if (t) *t = cpt->t;
  switch (cpt->t) {
  case NMDInt :
  case NMDReal :
  case NMDString :
    ierr = NMDGetValue(obj,cat,cmp,NULL,v,f); NMD_ERR_RETURN(ierr);
    break;
  case NMDIntarray :
    if (v || len) {
      struct NMD_intarray_struct *as = cpt->ii;
      if (v) *(int**)v = as->data;
      if (len) *len = as->length;
    }
    if (f) *f = NMDTrue; break;
  case NMDRealarray :
    if (v || len) {
      struct NMD_realarray_struct *as = cpt->rr;
      if (v) *(NMDRealtype**)v = as->data;
      if (len) *len = as->length;
    }
    if (f) *f = NMDTrue; break;
  default :
    if (f) *f = NMDFalse;
  }
  CHKMEMQ;
  return 0;
}

#if 0
#undef __FUNCT__
#define __FUNCT__ "NMDGetValueByIndex"
/*! Retrieve a stored value by index. As in NMDGetValue(), all output
  parameters can receive null pointers.
*/
NMDErrorCode NMDGetValueByIndex
(NMD_metadata obj,int icat,int icmp,NMDDataType *t,void *v,int *f)
{
  NMD_metadata_category cat;
  NMD_metadata_item cmp;
  NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(obj);
  if (icat<0 || icat>=obj->ncat)
    NMD_ERR_REPORTi("category index out of range",icat);
  cat = obj->cats[icat]; CHECKHASNMDCOOKIE(cat);
  if (icmp<0 || icmp>=obj->cats[icat]->ncmp)
    NMD_ERR_REPORTii("component index out of range",icat,icmp);

  cmp = cat->cmps[icmp]; CHECKHASNMDCOOKIE(cmp);
  if (t) *t = cmp->t;
  switch (cmp->t) {
  case NMDInt :
    if (v) *(int*)v = cmp->i;
    if (f) *f = NMDTrue; break;
  case NMDReal :
    if (v) *(double*)v = cmp->r;
    if (f) *f = NMDTrue; break;
  case NMDIntarray :
    if (v) *(int**)v = cmp->ii;
    if (f) *f = NMDTrue; break;
  case NMDRealarray :
    if (v) *(double**)v = cmp->rr;
    if (f) *f = NMDTrue; break;
  default :
    if (f) *f = NMDFalse;
  }
  return 0;
}
#endif

#undef __FUNCT__
#define __FUNCT__ "NMDGetDataType"
NMDErrorCode NMDGetDataType(NMD_metadata obj,const char *cat,const char *cmp,NMDDataType *t)
{
  NMD_metadata_category ctg; NMD_metadata_item cpt;
  NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectGetCategory(obj,cat,&ctg); NMD_ERR_RETURN(ierr);
  ierr = NMDCategoryGetComponent(ctg,cmp,&cpt); NMD_ERR_RETURN(ierr);
  CHECKHASNMDCOOKIE(cmp);
  *t = cpt->t;
  
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDIsArrayType"
/*! Test whether a data type is an array type */
NMDErrorCode NMDIsArrayType(NMDDataType type,NMDTruth *flg)
{
  *flg = ( (type==NMDIntarray) || (type==NMDRealarray) ? NMDTrue : NMDFalse ) ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDGetTypeMySQLName"
PetscErrorCode NMDGetTypeMySQLName(NMDDataType type,const char **name)
{
  int itype = (int)(type);
  if (itype<0 || itype>=nnmdtypenames)
    NMD_ERR_REPORTi("Invalid type",itype);
  *name = (char*)mysqltypenames[itype];
  return 0;
}

