#include <stdlib.h>
#include "string.h"
#include "nmd.h"
#include "nmd_impl.h"

#define CMPCHUNK 30

/*! \page categoryfunctions Metadata category manipulation

  A metadata object contains a number of categories, each containing
  multiple component which store the actual metadata. Here are the
  routines for manipulation the categories.

  See NMDObjectTryGetCategory(), NMDObjectGetCategory(),
  NMDAllocateCategory(), NMDObjectAllocateNewCategory(),
  NMDObjectGetOrCreateCategory(),
  NMDRemoveCategory(), NMDGetCategories(), NMDCopyCategory().
 */

#undef __FUNCT__
#define __FUNCT__ "NMDObjectTryGetCategory"
/*! Test whether a metadata object has a certain category, if so
 yield up its pointer. 

 The category pointer parameter can be null, in which case only existence 
 is tested.
*/
NMDErrorCode NMDObjectTryGetCategory
(NMD_metadata obj,const char *cat,NMD_metadata_category *rctg,NMDTruth *f)
{
  int icat;
  CHECKHASNMDCOOKIE(obj);
  *f = NMDFalse;
  for (icat=0; icat<obj->ncat; icat++) {
    NMD_metadata_category ctg = obj->cats[icat];
    CHECKHASNMDCOOKIE(ctg);
    if (strcmp(cat,ctg->name)==0) {
      if (rctg) *rctg = ctg; *f = NMDTrue; break;
    }
  }
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDObjectGetCategory"
/*! Retrieve a category from a metadata object. The category has to exist.
*/
NMDErrorCode NMDObjectGetCategory
(NMD_metadata obj,const char *cat,NMD_metadata_category *ctg)
{
  NMDTruth flg; NMDErrorCode ierr;
  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectTryGetCategory(obj,cat,ctg,&flg); NMD_ERR_RETURN(ierr);
  if (!flg) NMD_ERR_REPORTs("Category not found",cat);
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDAllocateCategory"
/*! This is an internal routine that merely allocates
  the data structure for storing a category.
*/
static NMDErrorCode NMDAllocateCategory(NMD_metadata_category *rcat)
{
  NMD_metadata_category cat;
  NMD_MALLOC(cat,1,struct NMD_metadata_category_,"category");
  cat->cookie = NMDCOOKIE;
  cat->alloc = CMPCHUNK; cat->ncmp = 0;
  NMD_MALLOC(cat->cmps,cat->alloc,NMD_metadata_item,"items array");
  *rcat = cat;  
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDObjectAllocateNewCategory"
/*! Allocate a category in a metadata object.
  There is no testing whether the category name is already in use.

  If a category pointer is supplied, the category is returned,
  but this pointer is allowed to be null.
*/
NMDErrorCode NMDObjectAllocateNewCategory
(NMD_metadata obj,const char *cat,NMD_metadata_category *rctg)
{
  NMD_metadata_category ctg; int idx; NMDErrorCode ierr;
  /*
   * if we are about to overflow, reallocate 
   */
  if (obj->ncat >= obj->alloc) {
    NMD_metadata_category* newcats; int newalloc;
    newalloc = 2*obj->alloc;
    ierr = PetscMalloc
      (newalloc*sizeof(struct NMD_metadata_category_),&newcats); CHKERRQ(ierr);
    for (idx=0; idx<obj->ncat; idx++)
      newcats[idx] = obj->cats[idx];
    ierr = PetscFree(obj->cats); CHKERRQ(ierr);
    obj->cats = newcats; obj->alloc = newalloc;
  }
  /*
   * with space guaranteed, create the category
   */
  idx = obj->ncat++;
  ierr = NMDAllocateCategory(&ctg); NMD_ERR_RETURN(ierr);
  NMD_STRDUP(cat,ctg->name);
  obj->cats[idx] = ctg;
  if (rctg) *rctg = ctg;
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDObjectGetOrCreateCategory"
/*! Retrieve a category from a metadata object, or create it
  if it doesn't exist yet.
*/
NMDErrorCode NMDObjectGetOrCreateCategory
(NMD_metadata obj,const char *cat,NMD_metadata_category *ctg)
{
  NMDTruth flg; NMDErrorCode ierr;
  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectTryGetCategory(obj,cat,ctg,&flg); NMD_ERR_RETURN(ierr);
  if (!flg) {
    ierr = NMDObjectAllocateNewCategory(obj,cat,ctg); NMD_ERR_RETURN(ierr);
  }
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDGetCategories"
/*! Get the number of categories and their names. Both arguments can be NULL.
  The names array is allocated; the user needs to free it. The names themselves
  are pointers to the strings in the metadata object, so they do not need to
  be freed.
*/
NMDErrorCode NMDGetCategories(NMD_metadata obj,int *ncat,char ***cats)
{
  CHECKHASNMDCOOKIE(obj);
  if (ncat) *ncat = obj->ncat;
  if (obj->ncat && cats) {
    char **names; int icat;
    NMD_MALLOC(names,obj->ncat,char*,"name array");
    for (icat=0; icat<obj->ncat; icat++)
      names[icat] = obj->cats[icat]->name;
    *cats = names;
  }
  CHKMEMQ;
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDRemoveCategory"
/* Remove a category from an NMD object. This is done through setting its
   name to \c "invalid". Later we will do real deallocation.

   The removed category still counts towards the number reported
   in NMDGetCategories().
*/
NMDErrorCode NMDRemoveCategory(NMD_metadata obj,const char *cat)
{
  NMD_metadata_category ctg; NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(obj);
  ierr = NMDObjectGetCategory(obj,(char*)cat,&ctg); NMD_ERR_RETURN(ierr);
  NMD_FREE(ctg->name);
  NMD_STRDUP("invalid",ctg->name);
  CHKMEMQ;

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "NMDCopyCategory"
/*! Copy category data from one metadata structure into another.
  This assumes that the category already exists in the target; 
  see for instance NMDHasCategory(), NMDCloneObject(),
  NMDCloneObjectStructure().
*/
NMDErrorCode NMDCopyCategory(NMD_metadata_category incat,NMD_metadata_category outcat)
{
  int icmp; NMDErrorCode ierr;

  CHECKHASNMDCOOKIE(incat);
  CHECKHASNMDCOOKIE(outcat);

  for (icmp=0; icmp<incat->ncmp; icmp++) {
    NMD_metadata_item src = incat->cmps[icmp],tar;
    ierr = NMDCategoryGetOrCreateComponent
      (outcat,src->name,src->t,&tar); NMD_ERR_RETURN(ierr);
    ierr = NMDCopyItemValues(src,tar); NMD_ERR_RETURN(ierr);
  }
  CHKMEMQ;

  return 0;
}

