/*! \file distribution.c \ingroup linear */

/*! \page distribution Permutation / load balancing

  Linear system solving is sensitive in several ways to permutations
  and load balancing applied to the system. This dependency typically comes
  through the preconditioner: incomplete factorizations are sensitive
  to permutations, and block Jacobi and Schwarz preconditioners are sensitive 
  to load distributions, even without any permutation applied.
 */
#include <stdlib.h>
#include "petscmat.h"
#include "petscconf.h"
#include "syspro.h"
#include "sysprotransform.h"
#include "sysprolinear.h"
#include "linear_impl.h"
#include "anamod.h"

#define PREPROCESSOR "distribution"
extern int SpectrumComputeUnpreconditionedSpectrum();

#undef __FUNCT__
#define __FUNCT__ "setup_distribution_choices"
static PetscErrorCode setup_distribution_choices()
{
  SalsaTransform tf; SalsaTransformObject cur;
  PetscBool flg; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = TransformGetByName(PREPROCESSOR,&tf); CHKERRQ(ierr);

  ierr = SysProDefineIntAnnotation
    (PREPROCESSOR,"preserve-simple"); CHKERRQ(ierr);
  ierr = SysProDefineIntAnnotation
    (PREPROCESSOR,"strictly-parallel"); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"induced",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"default distribution"); CHKERRQ(ierr);
  ierr = TransformObjectIntAnnotate(cur,"preserve-simple",1); CHKERRQ(ierr);

  ierr = HasComputeModule("icmk","splits",&flg); CHKERRQ(ierr);
  if (flg) {
    ierr = NewTransformObject(PREPROCESSOR,"icmk",&cur); CHKERRQ(ierr);
    ierr = TransformObjectSetExplanation
      (cur,"inverse cuthill-mckee"); CHKERRQ(ierr);
    ierr = TransformObjectIntAnnotate(cur,"strictly-parallel",1); CHKERRQ(ierr);
  } else printf("no icmk available\n");

#if defined(PETSC_HAVE_PARMETIS)
  ierr = NewTransformObject
    (PREPROCESSOR,"parmetis",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"parmetis package"); CHKERRQ(ierr);
#endif

#if defined(PETSC_HAVE_CHACO)
  ierr = NewTransformObject(PREPROCESSOR,"chaco",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"chaco package"); CHKERRQ(ierr);
  ierr = TransformObjectIntAnnotate(cur,"strictly-parallel",1); CHKERRQ(ierr);
#endif

  PetscFunctionReturn(0);
}

static PetscErrorCode specific_distribution_choices
(NumericalProblem problem,SalsaTransform tf)
{
  MPI_Comm comm = problem->comm; int ntids;
  int i,n; SalsaTransformObject *objs; PetscErrorCode ierr;
  PetscFunctionBegin;
  MPI_Comm_size(comm,&ntids);
  ierr = TransformGetObjects(tf,&n,&objs); CHKERRQ(ierr);
  for (i=0; i<n; i++) {
    SalsaTransformObject cur = objs[i]; 
    int v; PetscBool f;
    ierr = TransformObjectGetIntAnnotation
      (cur,"strictly-parallel",&v,&f); CHKERRQ(ierr);
    if (f && v) {
      ierr = TransformObjectMark(cur); CHKERRQ(ierr);
    }
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "sans_partition"
static PetscErrorCode sans_partition
(const char *type,NumericalProblem inproblem,int nparts,
 IS *local_to_global,VecScatter *perm)
{
  LinearSystem problem = (LinearSystem)inproblem;
  Mat A; Vec X; MPI_Comm comm; PetscBool flg;
  MatPartitioning Part;
  IS local,local_to_proc,new_part;
  int *part_sizes,local_size,first,last,ntids,mytid, ierr;
  
  PetscFunctionBegin;
  ierr = LinearSystemGetParts
    (problem,&A,PETSC_NULL,PETSC_NULL,&X,PETSC_NULL); CHKERRQ(ierr);
  ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
  ierr = MPI_Comm_size(comm,&ntids); CHKERRQ(ierr);
  ierr = MPI_Comm_rank(comm,&mytid); CHKERRQ(ierr);
  ierr = MatGetOwnershipRange(A,&first,&last); CHKERRQ(ierr);
  local_size = last-first;
  ierr = VecCreateMPI(comm,local_size,PETSC_DECIDE,&X); CHKERRQ(ierr);
  ierr = PetscMalloc((nparts+1)*sizeof(int),&part_sizes); CHKERRQ(ierr);
  ierr = ISCreateStride(comm,local_size,first,1,&local); CHKERRQ(ierr);

  flg = PETSC_FALSE;
#if defined(PETSC_HAVE_PARMETIS)
  if (!flg) {
    ierr = PetscStrcmp(type,"parmetis",&flg); CHKERRQ(ierr);}
#endif
#if defined(PETSC_HAVE_CHACO)
  if (!flg) {
    ierr = PetscStrcmp(type,"chaco",&flg); CHKERRQ(ierr);}
#endif
  if (flg) {
    IS tis; PetscInt local_size;

    /*ierr = MatConvert(A,MATMPIADJ,MAT_INITIAL_MATRIX,&Adj); CHKERRQ(ierr);*/
    ierr = MatPartitioningCreate(comm,&Part); CHKERRQ(ierr);
    ierr = MatPartitioningRegisterAll(PETSC_NULL); CHKERRQ(ierr);
    ierr = MatPartitioningSetType(Part,type); CHKERRQ(ierr);
    ierr = MatPartitioningSetAdjacency(Part,A/* Adj? */); CHKERRQ(ierr);

    ierr = MatPartitioningSetNParts(Part,nparts); CHKERRQ(ierr);
    ierr = MatPartitioningApply(Part,&local_to_proc); CHKERRQ(ierr);
    ierr = MatPartitioningDestroy(&Part); CHKERRQ(ierr);
    /*ierr = MatDestroy(Adj); CHKERRQ(ierr);*/

    ierr = ISPartitioningToNumbering(local_to_proc,&tis); CHKERRQ(ierr);
    ierr = ISPartitioningCount(local_to_proc,nparts,part_sizes); CHKERRQ(ierr);
    ierr = ISDestroy(&local_to_proc); CHKERRQ(ierr);
    ierr = MatGetLocalSize(A,&local_size,PETSC_NULL); CHKERRQ(ierr);
    ierr = ISInvertPermutation(tis,local_size,local_to_global); CHKERRQ(ierr);
    /*printf("mapping\n"); ISView(local_to_global,0);*/
    
    goto size_to_part;
  }

  ierr = PetscStrcmp(type,"icmk",&flg); CHKERRQ(ierr);
  if (flg) {
    int *ss;
    *local_to_global = local;
    ierr = SysProComputeQuantity
      (inproblem,"icmk","splits",(void*)&ss,PETSC_NULL,
       &flg); CHKERRQ(ierr);
    if (flg) {
      ierr = ISCreateGeneral
	(MPI_COMM_SELF,ss[0],ss+1,PETSC_OWN_POINTER,&new_part); CHKERRQ(ierr);
    } else {
      SETERRQ(PETSC_COMM_SELF,1,"Hm.....");
    }
    goto scatter;
  }    

  SETERRQ1(PETSC_COMM_SELF,1,"Unknown distribution <%s>",type);

 size_to_part:
  {
    int i,t,s=0;
    for (i=0; i<nparts; i++) {
      t = part_sizes[i]; part_sizes[i] = s; s+=t;}
    part_sizes[nparts] = s;
  }

  ierr = ISCreateGeneral
    (MPI_COMM_SELF,nparts+1,part_sizes,PETSC_OWN_POINTER,&new_part); CHKERRQ(ierr);
  ierr = PetscFree(part_sizes); CHKERRQ(ierr);
 scatter:
  ierr = VecScatterCreate(X,local,X,*local_to_global,perm); CHKERRQ(ierr);
  if (local!=*local_to_global) {
    ierr = ISDestroy(&local); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "distribute_system"
static PetscErrorCode distribute_system
(const char *type,int nopt,PetscBool overwrite,
 NumericalProblem inproblem,NumericalProblem *outproblem,
 void *gctx,void **ctx, PetscBool *success)
{
  LinearSystem problem = (LinearSystem)inproblem,newproblem;
  PetscBool flg; PetscErrorCode ierr; 

  PetscFunctionBegin;
  *success = PETSC_TRUE;

  ierr = PetscStrcmp(type,"induced",&flg); CHKERRQ(ierr);
  if (flg) {
    ierr = LinearSystemDuplicatePointers(problem,&newproblem); CHKERRQ(ierr);
    *(VecScatter*)ctx = NULL;
    goto done;
  }

  {
    Mat A,Ause,B,Buse; Vec X,Xuse,I,Iuse,V,Vuse;
    VecScatter perm;

    ierr = LinearSystemDuplicatePointers(problem,&newproblem); CHKERRQ(ierr);
    ierr = LinearSystemGetParts(problem,&A,&B,&V,&X,&I); CHKERRQ(ierr);

    {
      IS local_to_global=0,global_to_global=0;
      ierr = sans_partition(type,inproblem,6,&local_to_global,&perm); CHKERRQ(ierr);
      ierr = ISAllGather(local_to_global,&global_to_global); CHKERRQ(ierr);
      ierr = ISSetPermutation(local_to_global); CHKERRQ(ierr);
      ierr = ISSetPermutation(global_to_global); CHKERRQ(ierr);
      ierr = MatPermute
	(A,local_to_global,global_to_global,&Ause); CHKERRQ(ierr);
      if (B && B!=A) {
	ierr = MatPermute
	  (B,local_to_global,local_to_global,&Buse); CHKERRQ(ierr);
      } else Buse = Ause;
      ierr = ISDestroy(&local_to_global); CHKERRQ(ierr);
      ierr = ISDestroy(&global_to_global); CHKERRQ(ierr);
    }
    *(VecScatter*)ctx = perm;

    {
      int local_size; MPI_Comm comm;
      ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
      ierr = MatGetLocalSize(Ause,&local_size,PETSC_NULL); CHKERRQ(ierr);
      ierr = VecCreateMPI(comm,local_size,PETSC_DECIDE,&Xuse); CHKERRQ(ierr);
      ierr = VecDuplicate(Xuse,&Iuse); CHKERRQ(ierr);
      ierr = VecDuplicate(Xuse,&Vuse); CHKERRQ(ierr);
    }

    /* permute the rhs */
    ierr = VecScatterBegin
      (perm,V,Vuse,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
    ierr = VecScatterEnd
      (perm,V,Vuse,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
    /* permute known solution */
      ierr = VecScatterBegin
	(perm,X,Xuse,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
    ierr = VecScatterEnd
      (perm,X,Xuse,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
    /* permute initial guess */
    ierr = VecScatterBegin
      (perm,I,Iuse,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
    ierr = VecScatterEnd
      (perm,I,Iuse,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
  
    ierr = LinearSystemSetParts
      (newproblem,Ause,Buse,Vuse,Xuse,Iuse); CHKERRQ(ierr);

    goto done;
  }

  SETERRQ1(PETSC_COMM_SELF,1,"Undefined distribution <%s>",type);

 done:
  *outproblem = (NumericalProblem)newproblem;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "undistribute_system"
static PetscErrorCode undistribute_system
(const char *scaling_type,PetscBool overwrite,void *gctx,void *ctx,
 NumericalProblem problem,NumericalProblem nextproblem,
 NumericalSolution before,NumericalSolution after)
{
  LinearSolution shuffled = (LinearSolution)before, 
    rectified = (LinearSolution)after;
  VecScatter perm = (VecScatter)ctx; Vec Xfore,Xback; 
  PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = LinearSolutionCopyStats(shuffled,rectified); CHKERRQ(ierr);
  ierr = LinearSolutionGetVector(shuffled,&Xfore); CHKERRQ(ierr);
  ierr = LinearSolutionGetVector(rectified,&Xback); CHKERRQ(ierr);
  if (perm) {
    ierr = VecScatterBegin
      (perm,Xfore,Xback,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    ierr = VecScatterEnd
      (perm,Xfore,Xback,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    ierr = VecScatterDestroy(&perm); CHKERRQ(ierr);
  } else {
    ierr = VecCopy(Xfore,Xback); CHKERRQ(ierr);
  }

  /*ierr = LinearSolutionDelete((LinearSolution)before); CHKERRQ(ierr);*/
  //ierr = DeleteLinearSystem((LinearSystem)problem); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeclareDistributionPreprocessor"
PetscErrorCode DeclareDistributionPreprocessor(void)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = DeclarePreprocessor
    ("distribution",
     &setup_distribution_choices,&specific_distribution_choices,0,0,
     0,0,
     &distribute_system,&undistribute_system); CHKERRQ(ierr);
  ierr = PreprocessorSetPreservedCategories
    (PREPROCESSOR,"laplace"); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
