/*! \file flipsign.c \ingroup linear
 */

/*! \page flipsign Flip the sign of a matrix

  Most code for iterative methods and preconditioners assumes somewhere that
  the sign of a matrix is predominantely positive. Hence, we flip the sign
  of matrices that have no positive diagonal elements.
 */
#include <stdlib.h>
#include <stdio.h>
#include "petsc.h"
#include "syspro.h"
#include "sysprotransform.h"
#include "sysprolinear.h"
#include "linear_impl.h"
#include "anamod.h"

#define PREPROCESSOR "flipsign"

#undef __FUNCT__
#define __FUNCT__ "flipsign"
static PetscErrorCode flipsign
(const char *type,int nopt,PetscBool overwrite,
 NumericalProblem inproblem,NumericalProblem *outproblem,
 void *gctx,void **ctx,PetscBool *success)
{
  LinearSystem problem = (LinearSystem)inproblem, newproblem;
  PetscBool flg; PetscErrorCode ierr; 

  PetscFunctionBegin;

  *success = PETSC_TRUE; *(int**)ctx = NULL;
  ierr = LinearSystemDuplicatePointers(problem,&newproblem); CHKERRQ(ierr);

  ierr = PetscStrcmp(type,"identity",&flg); CHKERRQ(ierr);
  if (flg) {
    goto done;
  }


  ierr = PetscStrcmp(type,"flip",&flg); CHKERRQ(ierr);
  if (flg) {
    Mat A,B,Ause,Buse; Vec X,V,I,Vuse; int ds;
    
    ierr = LinearSystemGetParts(newproblem,&A,&B,&V,&X,&I); CHKERRQ(ierr);
    ierr = SysProComputeQuantity
      (inproblem,"variance","diagonal-sign",(void*)&ds,PETSC_NULL,
       &flg); CHKERRQ(ierr);
    /*
     * Flip the sign if necessary
     */
    if (flg && ds<0) {
      ierr = MatDuplicate(A,MAT_COPY_VALUES,&Ause); CHKERRQ(ierr);
      ierr = MatScale(Ause,-1); CHKERRQ(ierr);
      if (B && B!=A) {
	ierr = MatDuplicate(B,MAT_COPY_VALUES,&Buse); CHKERRQ(ierr);
	ierr = MatScale(Buse,-1); CHKERRQ(ierr);
      } else Buse = Ause;
      ierr = VecDuplicate(V,&Vuse); CHKERRQ(ierr);
      ierr = VecCopy(V,Vuse); CHKERRQ(ierr);
      ierr = VecScale(Vuse,-1); CHKERRQ(ierr);
      ierr = LinearSystemSetParts
	(newproblem,Ause,Buse,Vuse,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
      
      {
	int* saveone;
	ierr = PetscMalloc(sizeof(int),&saveone); CHKERRQ(ierr);
	*saveone = 1;
	*(int**)ctx = saveone;
      }
    }
    goto done;
  }

  SETERRQ1(PETSC_COMM_SELF,1,"Unknown flipsign preprocessor <%s>",type);
 done:
  *outproblem = (NumericalProblem)newproblem;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "back_flipsign"
static PetscErrorCode back_flipsign
(const char *flipsign_type,PetscBool overwrite,void *gctx,void *ctx,
 NumericalProblem nextproblem,NumericalProblem problem,
 NumericalSolution flipped,NumericalSolution straight)
{
  /*LinearSystem sys = (LinearSystem)nextproblem;*/
  LinearSolution flippedsol = (LinearSolution)flipped,
    straightsol = (LinearSolution)straight;
  PetscErrorCode ierr;

  PetscFunctionBegin;

  if (ctx) {
    int *saveone = (int*)ctx;
    if (*saveone != 1)
      SETERRQ1(PETSC_COMM_SELF,1,"Context should contain value 1; has %d",*(int*)ctx);
    ierr = PetscFree(saveone); CHKERRQ(ierr);
  }

  //ierr = DeleteLinearSystem(sys); CHKERRQ(ierr);
  ierr = LinearSolutionCopy(flippedsol,straightsol); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "setup_flipsign_choices"
/*! This routine is only called when the flipsign preprocessor
  is created by DeclarePreprocessor() inside DeclareFlipsignPreprocessor()
*/
static PetscErrorCode setup_flipsign_choices()
{
  SalsaTransform tf; SalsaTransformObject cur; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = TransformGetByName("flipsign",&tf); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"identity",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"do not flip the sign"); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"flip",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"flip the sign of the system"); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "specific_flipsign_choices"
/*!
  This is the 'specific setup' phase of the flipsign preprocessor.
  See \ref modes for details.

  It disables either the identity or the flip routine, to
  leave only the one applicable to this particular system.

*/
static PetscErrorCode specific_flipsign_choices
(NumericalProblem theproblem,SalsaTransform flipsign)
{
  SalsaTransformObject to;
  PetscBool flg; int nsign; PetscErrorCode ierr;
  PetscFunctionBegin;

  /*
   * Mark one or the other; the other is then taken
   */
  ierr = SysProComputeQuantity
    (theproblem,"variance","diagonal-sign",(void*)&nsign,PETSC_NULL,
     &flg); CHKERRQ(ierr);
  if (flg && nsign<0) {
    ierr = TransformObjectGetByName(PREPROCESSOR,"identity",&to); CHKERRQ(ierr);
    ierr = TransformObjectMark(to); CHKERRQ(ierr);
  } else {
    ierr = TransformObjectGetByName(PREPROCESSOR,"flip",&to); CHKERRQ(ierr);
    ierr = TransformObjectMark(to); CHKERRQ(ierr);
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeclareFlipsignPreprocessor"
PetscErrorCode DeclareFlipsignPreprocessor(void)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = DeclarePreprocessor
    ("flipsign",
     &setup_flipsign_choices,&specific_flipsign_choices,0,0,
     0,0,
     &flipsign,&back_flipsign); CHKERRQ(ierr);
  ierr = PreprocessorSetPreservedCategories
    (PREPROCESSOR,"laplace,structure,simple"); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
