/*! \file linear.c \anchor linearfile 
  
  \page linear The Linear package for SysPro

  A linear system is a special case of a numerical problem. This file
  contains the routines for creating, deleting, and duplicating linear 
  systems.

  \ref lineardef

  \ref lineartransform

  \section lineartransform Transformations for linear systems

  \subpage flipsign

  \subpage singleton

  \subpage distribution

  \subpage scaling

  \subpage approximation

  \subpage ksp

  \subpage pc

  The linear package for SysPro can use the NMD and AnaMod libraries,
  but does not require them. Any dependencies on NMD and Anamod should
  all be restricted to \ref computec.
 */

#include <stdlib.h>
#include "syspro.h"
#include "syspro_impl.h"
#include "sysprolinear.h"
#include "linear_impl.h"
#include "petscmat.h"
#include "nmd.h"
#include "anamod.h"

/*! \page lineardef Definition of a linear system

  A linear system has the following components:
  - \c A : coefficient matrix
  - \c B : a matrix from which to build a preconditioner. Often this
  will just be \c A.
  - \c rhs : the right hand side
  - \c sol : a known solution, if any; there is a boolean to indicate
  - \c init : a starting guess for iterative methods
  whether a solution is known.
  - \c ctx : a void pointer for storing an arbitrary data item; this
  can be used by the user.


  See CreateLinearSystem(), DeleteLinearSystem(), LinearSystemSetParts(),
  LinearSystemGetParts(), LinearSystemInheritParts().

  See also \ref linearsolution.
 */

#undef __FUNCT__
#define __FUNCT__ "LinearPackageSetUp"
PetscErrorCode LinearPackageSetUp()
{
  PetscFunctionBegin;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CreateLinearSystem"
/*! Allocate the structure for a linear system */
PetscErrorCode CreateLinearSystem(MPI_Comm comm,LinearSystem *system)
{
  LinearSystem lnew; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscMalloc(sizeof(struct LinearSystem_),&lnew); CHKERRQ(ierr);
  ierr = PetscMemzero(lnew,sizeof(struct LinearSystem_)); CHKERRQ(ierr);
  ((NumericalProblem)lnew)->comm = comm;
  lnew->cookie = LINSYSCOOKIE; lnew->partsoriginal = 0;
  *system = lnew;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemDelete"
PetscErrorCode LinearSystemDelete(LinearSystem system)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(system);
  /*printf("deleting parts: %d\n",system->partsoriginal);*/
  if (system->A!=system->B && system->B) {
    if (system->partsoriginal & 8) {
      /*      ierr = PurgeAttachedArrays(system->B); CHKERRQ(ierr);*/
      ierr = MatDestroy(&(system->B)); CHKERRQ(ierr); system->B = NULL;}}
  if (system->partsoriginal & 16) {
    /*    ierr = PurgeAttachedArrays(system->A); CHKERRQ(ierr);*/
    ierr = MatDestroy(&(system->A)); CHKERRQ(ierr); system->A = NULL;}
  if (system->partsoriginal & 4) {
    ierr = VecDestroy(&(system->Rhs)); CHKERRQ(ierr); system->Rhs = NULL;}
  if (system->partsoriginal & 2) {
    ierr = VecDestroy(&(system->Sol)); CHKERRQ(ierr); system->Sol = NULL;}
  if (system->Init && system->partsoriginal & 1) {
    ierr = VecDestroy(&(system->Init)); CHKERRQ(ierr); system->Init = NULL;}
  if (system->Tmp) {
    ierr = VecDestroy(&(system->Tmp)); CHKERRQ(ierr); system->Tmp = NULL;}
  CHKMEMQ;
  ierr = PetscFree(system); CHKERRQ(ierr);
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearDeleteNumericalProblem"
/*! This is like LinearSystemDelete(), except that the argument
  has been cast so that this routine can be used as the \c problemdelete
  argument of SysProDeclareFunctions().
*/
PetscErrorCode LinearDeleteNumericalProblem(NumericalProblem sys)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = LinearSystemDelete((LinearSystem)sys); CHKERRQ(ierr);
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemSetParts"
/*! Declare the matrices and vectors for a linear system.

  Arguments:
  - \c system
  - \c A : the matrix
  - \c B : operator to construct the preconditioner from; if NULL, 
  (or identical to A), A will be used
  - rhs : right hand side
  - sol : storage for the computed solution
  - init : (optional) nontrivial starting vector for iterative solution
 */
PetscErrorCode LinearSystemSetParts
(LinearSystem system,Mat A,Mat B,Vec Rhs,Vec Sol,Vec Init)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(system);
  if (A && A!=system->A) {
    system->A = A; system->partsoriginal |= 16;}
  if (B && B!=system->B) {
    system->B = B;  system->partsoriginal |= 8;}
  if (Rhs && Rhs!=system->Rhs) {
#if PETSC_USE_DEBUG==1
    PetscReal nrm;
    PetscErrorCode ierr = VecNorm(Rhs,NORM_2,&nrm); CHKERRQ(ierr);
    if (nrm==0.) SETERRQ(PETSC_COMM_SELF,1,"Setting zero Rhs");
#endif
    system->Rhs = Rhs; system->partsoriginal |= 4;}
  if (Sol && Sol!=system->Sol) {
    system->Sol = Sol; system->partsoriginal |= 2;}
  if (Init && Init!=system->Init) {
    system->Init = Init; system->partsoriginal |= 1;}
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemInheritParts"
/*! Declare the matrices and vectors for a linear system. 
  Unlike in LinearSystemSetParts(), here the parts are marked as
  not original, so they will not be deleted in DeleteLinearSystem().
*/
PetscErrorCode LinearSystemInheritParts
(LinearSystem system,Mat A,Mat B,Vec Rhs,Vec Sol,Vec Init)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(system);
  if (A && A!=system->A) {
    system->A = A; system->partsoriginal &= ALLPARTSNEW-16;}
  if (B && B!=system->B) {
    system->B = B;  system->partsoriginal &= ALLPARTSNEW-8;}
  if (Rhs && Rhs!=system->Rhs) {
#if PETSC_USE_DEBUG==1
    PetscReal nrm;
    PetscErrorCode ierr = VecNorm(Rhs,NORM_2,&nrm); CHKERRQ(ierr);
    if (nrm==0.) SETERRQ(PETSC_COMM_SELF,1,"Inheriting zero Rhs");
#endif
    system->Rhs = Rhs; system->partsoriginal &= ALLPARTSNEW-4;}
  if (Sol && Sol!=system->Sol) {
    system->Sol = Sol; system->partsoriginal &= ALLPARTSNEW-2;}
  if (Init && Init!=system->Init) {
    system->Init = Init; system->partsoriginal &= ALLPARTSNEW-1;}
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemGetParts"
/*! Get the matrices and vectors of the system */
PetscErrorCode LinearSystemGetParts
(LinearSystem system,Mat *A,Mat *B,Vec *Rhs,Vec *Sol,Vec *Init)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(system);
  if (A) *A = system->A;
  if (B) *B = system->B;
  if (Rhs) *Rhs = system->Rhs;
  if (Sol) *Sol = system->Sol;
  if (Init) *Init = system->Init;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemSetContext"
PetscErrorCode LinearSystemSetContext(LinearSystem system,void *ctx)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(system);
  system->ctx = ctx;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemGetContext"
PetscErrorCode LinearSystemGetContext(LinearSystem system,void **ctx)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(system);
  *ctx = system->ctx;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemSetKnownSolution"
PetscErrorCode LinearSystemSetKnownSolution(LinearSystem sys,PetscBool sol)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(sys);
  sys->known_solution = sol;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemGetKnownSolution"
PetscErrorCode LinearSystemGetKnownSolution(LinearSystem sys,PetscBool *sol)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(sys);
  *sol = sys->known_solution;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemSetMetadata"
PetscErrorCode LinearSystemSetMetadata(LinearSystem system,NMD_metadata nmd)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(system);
  system->metadata = nmd;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemGetMetadata"
PetscErrorCode LinearSystemGetMetadata(LinearSystem system,NMD_metadata *nmd)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(system);
  *nmd = system->metadata;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemGetTmpVector"
PetscErrorCode LinearSystemGetTmpVector(LinearSystem sys,Vec *tmp)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(sys);
  if (!sys->Tmp) {ierr = VecDuplicate(sys->Rhs,&(sys->Tmp)); CHKERRQ(ierr);}
  *tmp = sys->Tmp;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemDuplicatePointers"
/*! Allocate a new linear system and give it the components of the old
  by pointer duplication.
*/
PetscErrorCode LinearSystemDuplicatePointers
(LinearSystem problem,LinearSystem *newproblem)
{
  LinearSystem lnew; PetscErrorCode ierr;
  SYSPROCHECKVALIDLINSYS(problem);
  PetscFunctionBegin;
  ierr = CreateLinearSystem
    (((NumericalProblem)problem)->comm,&lnew); CHKERRQ(ierr);
  lnew->partsoriginal = 0;
  lnew->A = problem->A; lnew->B = problem->B;
  lnew->Rhs = problem->Rhs;
  lnew->Sol = problem->Sol;
  lnew->Init = problem->Init;
  lnew->ctx = problem->ctx;
  lnew->known_solution = problem->known_solution;
  lnew->metadata = problem->metadata;
  *newproblem = lnew;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemDuplicate"
/*! Allocate a new linear system, and create copies in it
  of the data structure, but not the values,
  of the components of the old system.

  See also LinearSystemCopy().
*/
PetscErrorCode LinearSystemDuplicate
(LinearSystem problem,LinearSystem *newproblem)
{
  LinearSystem lnew; PetscErrorCode ierr;
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(problem);
  ierr = CreateLinearSystem
    (((NumericalProblem)problem)->comm,&lnew); CHKERRQ(ierr);
  lnew->partsoriginal = ALLPARTSNEW;
  ierr = MatDuplicate
    (problem->A,MAT_DO_NOT_COPY_VALUES,&(lnew->A)); CHKERRQ(ierr);
  if (problem->A==problem->B)
    lnew->B = lnew->A;
  else if (problem->B) {
    ierr = MatDuplicate
      (problem->B,MAT_DO_NOT_COPY_VALUES,&(lnew->B)); CHKERRQ(ierr);
  }
  CHKMEMQ;
  if (!problem->Rhs)
    SETERRQ(PETSC_COMM_SELF,1,"Linear system has no rhs");
  ierr = VecDuplicate(problem->Rhs,&(lnew->Rhs)); CHKERRQ(ierr);
  if (!problem->Sol)
    SETERRQ(PETSC_COMM_SELF,1,"Linear system has no sol");
  ierr = VecDuplicate(problem->Sol,&(lnew->Sol)); CHKERRQ(ierr);
  if (problem->Init) {
    ierr = VecDuplicate(problem->Init,&(lnew->Init)); CHKERRQ(ierr);
  }
  lnew->known_solution = problem->known_solution;
  lnew->ctx = problem->ctx;
#if 0
  {
    ierr = NMDCloneObject(problem->nmd,&lnew->nmd); CHKERRQ(ierr);
  }
#endif
  *newproblem = lnew;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemCopy"
/*! Copy the values of the components of an old linear system into a new.
  The new system has to have been created with LinearSystemDuplicate()
  because this routine assumes that the data structures are already
  in place.
*/
PetscErrorCode LinearSystemCopy(LinearSystem old,LinearSystem lnew)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYSa(old,1);
  SYSPROCHECKVALIDLINSYSa(lnew,2);
  if (lnew->partsoriginal!=ALLPARTSNEW)
    SETERRQ(PETSC_COMM_SELF,1,"Attempting in-place copy of linear system");
  ierr = MatCopy(old->A,lnew->A,SAME_NONZERO_PATTERN); CHKERRQ(ierr);
  if (old->A==old->B)
    lnew->B = lnew->A;
  else if (old->B) {
    ierr = MatCopy(old->B,lnew->B,SAME_NONZERO_PATTERN); CHKERRQ(ierr);
  }
  CHKMEMQ;
  if (old->metadata) {
    ierr = NMDCloneObjectStructure
      (old->metadata,&lnew->metadata); CHKERRQ(ierr);
    ierr = NMDCloneObject(old->metadata,lnew->metadata); CHKERRQ(ierr);
  }
  ierr = VecCopy(old->Rhs,lnew->Rhs); CHKERRQ(ierr);
  ierr = VecCopy(old->Sol,lnew->Sol); CHKERRQ(ierr);
  if (old->Init) {
    ierr = VecCopy(old->Init,lnew->Init); CHKERRQ(ierr);}
  lnew->known_solution = old->known_solution;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

/*! \page linearsolution

  There is an object to store the solution of a linear system.

  The solution of a linear system is stored in a data structure that contains
  - \c out : the computed output vector
  - \c statistics : an NMD object. See LinearSolutionCreateStatistics().

  See CreateLinearSolution(), LinearSolutionDelete(), 
  LinearSolutionCopy(), LinearCopyNumericalSolution(), 
 */

#undef __FUNCT__
#define __FUNCT__ "CreateLinearSolution"
PetscErrorCode CreateLinearSolution(LinearSolution *sol)
{
  LinearSolution lnew; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscMalloc(sizeof(struct LinearSolution_),&lnew); CHKERRQ(ierr);
  ierr = PetscMemzero(lnew,sizeof(struct LinearSolution_)); CHKERRQ(ierr);
  ierr = NMDCreateObject(&(lnew->statistics)); CHKERRQ(ierr);
  lnew->cookie = LINSOLCOOKIE; *sol = lnew;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearCreateNumericalSolution"
/*!
  Shell routine around CreateLinearSolution() to save you some type casting.

  If the first argument is not NULL, its matrix is extracted and used
  to create the vector of the solution object.
*/
PetscErrorCode LinearCreateNumericalSolution
(NumericalProblem prob,NumericalSolution *sol)
{
  LinearSolution linsol; PetscErrorCode ierr; 
  PetscFunctionBegin;
  ierr = CreateLinearSolution(&linsol); CHKERRQ(ierr);
  if (prob) {
    MPI_Comm comm; Mat A; Vec V; int localsize;
    ierr = LinearSystemGetParts
      ((LinearSystem)prob,&A,NULL,NULL,NULL,NULL); CHKERRQ(ierr);
    ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
    ierr = MatGetLocalSize(A,&localsize,NULL); CHKERRQ(ierr);
    ierr = VecCreateMPI(comm,localsize,PETSC_DECIDE,&V); CHKERRQ(ierr);
    ierr = LinearSolutionSetVector(linsol,V); CHKERRQ(ierr);
  }
  *sol = (NumericalSolution)linsol;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSolutionDelete"
/*! Delete a linear solution.

  This does not affect the context stored in the solution. That needs
  a special purpose routine.

  See also LinearDeleteNumericalSolution().
*/
PetscErrorCode LinearSolutionDelete(LinearSolution sol)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSOL(sol);
  ierr = VecDestroy(&(sol->Out)); CHKERRQ(ierr);
  ierr = NMDDestroyObject(sol->statistics); CHKERRQ(ierr);
  ierr = PetscFree(sol); CHKERRQ(ierr);
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearDeleteNumericalSolution"
/*! This is like LinearSolutionDelete(), except that the argument
  has been cast so that this routine can be used as the \c solutiondelete
  argument of SysProDeclareFunctions().
*/
PetscErrorCode LinearDeleteNumericalSolution(NumericalSolution sol)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = LinearSolutionDelete((LinearSolution)sol); CHKERRQ(ierr);
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSolutionCopy"
/*! Copy one linear solution object into another.
  This clearly only works if their vectors are similarly layed out.

  The context pointer is blindly copied. We may have to think about this 
  a bit more.

  See also LinearCopyNumericalSolution().
*/
PetscErrorCode LinearSolutionCopy(LinearSolution old,LinearSolution lnew)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSOLa(old,1);
  SYSPROCHECKVALIDLINSOLa(lnew,2);
  ierr = NMDCloneObject(old->statistics,lnew->statistics); CHKERRQ(ierr);
  if (old->Out!=lnew->Out) {ierr = VecCopy(old->Out,lnew->Out); CHKERRQ(ierr);}
  lnew->ctx = old->ctx;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearCopyNumericalSolution"
/*! This routine is essentially LinearSolutionCopy(), except that it does
  casts of the arguments so that it can be used as the \c solutioncopy
  member of SysProDeclareFunctions()
*/
PetscErrorCode LinearCopyNumericalSolution(NumericalSolution old,NumericalSolution nnew)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = LinearSolutionCopy
    ((LinearSolution)old,(LinearSolution)nnew); CHKERRQ(ierr);
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "CreateDefaultLinearSolution"
PetscErrorCode CreateDefaultLinearSolution
(NumericalProblem problem,NumericalSolution *rsol)
{
  LinearSystem sys = (LinearSystem)problem;
  LinearSolution solution; Vec Rhs,Out;
  PetscErrorCode ierr;
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(sys);
  ierr = CreateLinearSolution(&solution); CHKERRQ(ierr);
  ierr = LinearSystemGetParts
    (sys,PETSC_NULL,PETSC_NULL,&Rhs,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
  ierr = VecDuplicate(Rhs,&Out); CHKERRQ(ierr);
  ierr = LinearSolutionSetVector(solution,Out); CHKERRQ(ierr);
  *rsol = (NumericalSolution)solution;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSolutionSetVector"
PetscErrorCode LinearSolutionSetVector(LinearSolution sol,Vec out)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSOL(sol);
  sol->Out = out;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSolutionGetVector"
PetscErrorCode LinearSolutionGetVector(LinearSolution sol,Vec *out)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSOL(sol);
  *out = sol->Out;
  PetscFunctionReturn(0);
}

/*! \page solutionstats Solution Statistics
  
  The \c LinearSolution object carries an \c NMD_metadata object that 
  contains performance measurements. This object is initially empty,
  so we need to build its content.
 */
#undef __FUNCT__
#define __FUNCT__ "LinearSolutionCreateStatistics"
PetscErrorCode LinearSolutionCreateStatistics(LinearSolution sol)
{
  NMD_metadata stats; NMD_metadata_category data; PetscErrorCode ierr;
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSOL(sol);
  stats = sol->statistics;
  ierr = NMDObjectAllocateNewCategory(stats,"statistics",&data); CHKERRQ(ierr);
  CHKMEMQ;
  ierr = NMDCategoryAllocateNewComponent
    (data,"converged",NMDInt,NULL); CHKERRQ(ierr);
  ierr = NMDCategoryAllocateNewComponent
    (data,"abort",NMDInt,NULL); CHKERRQ(ierr);
  ierr = NMDCategoryAllocateNewComponent
    (data,"iterations",NMDInt,NULL); CHKERRQ(ierr);
  ierr = NMDCategoryAllocateNewComponent
    (data,"setup-time",NMDReal,NULL); CHKERRQ(ierr);
  ierr = NMDCategoryAllocateNewComponent
    (data,"solve-time",NMDReal,NULL); CHKERRQ(ierr);
  ierr = NMDCategoryAllocateNewComponent
    (data,"convergence-type",NMDInt,NULL); CHKERRQ(ierr);
  ierr = NMDCategoryAllocateNewComponent
    (data,"error",NMDReal,NULL); CHKERRQ(ierr);
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSolutionGetStatistics"
PetscErrorCode LinearSolutionGetStatistics(LinearSolution sol,NMD_metadata *s)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSOL(sol);
  *s = sol->statistics;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSolutionCopyStats"
PetscErrorCode LinearSolutionCopyStats(LinearSolution in,LinearSolution out)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSOLa(in,1);
  SYSPROCHECKVALIDLINSOLa(out,2);
  ierr = NMDCloneObject(in->statistics,out->statistics); CHKERRQ(ierr);
  CHKMEMQ;
  PetscFunctionReturn(0);
}

/*! \page linearsolutioncontext The linear solution context
  We use the context pointer in a LinearSolution object to store diagnostics.
  This pointer is blindly copied in LinearSolutionCopy() (unlike the solution 
  vector, which is replicated) so we have to be careful with deallocating.
 */

#undef __FUNCT__
#define __FUNCT__ "LinearSolutionSetContext"
PetscErrorCode LinearSolutionSetContext(LinearSolution sol,void *ctx)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSOL(sol);
  sol->ctx = ctx;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSolutionGetContext"
PetscErrorCode LinearSolutionGetContext(LinearSolution sol,void **ctx)
{
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSOL(sol);
  *ctx = sol->ctx;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearDeleteNumericalSolutionContext"
PetscErrorCode LinearDeleteNumericalSolutionContext(NumericalSolution sol)
{
  LinearSolution solution = (LinearSolution)sol;
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSOL(solution);
  if (solution->ctx) {
    /*ierr = delete_diagnostics(solution->ctx); CHKERRQ(ierr);*/
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemTrueDistance"
PetscErrorCode LinearSystemTrueDistance
(LinearSystem system,LinearSolution linsol,PetscReal *rnrm)
{
  Vec tmp,out,sol;
  PetscReal nrm; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = LinearSystemGetParts
    (system,PETSC_NULL,PETSC_NULL,PETSC_NULL,&sol,PETSC_NULL); CHKERRQ(ierr);
  ierr = LinearSystemGetTmpVector(system,&tmp); CHKERRQ(ierr);
  ierr = LinearSolutionGetVector(linsol,&out); CHKERRQ(ierr);
  ierr = VecCopy(sol,tmp); CHKERRQ(ierr);
  ierr = VecAXPY(tmp,-1,out); CHKERRQ(ierr);
  ierr = VecNorm(tmp,NORM_2,&nrm); CHKERRQ(ierr);
  *rnrm = nrm;
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "LinearSystemTrueDistancePrint"
PetscErrorCode LinearSystemTrueDistancePrint
(NumericalProblem problem,NumericalSolution solution,char *caption)
{
  LinearSystem system = (LinearSystem)problem;
  LinearSolution linsol = (LinearSolution)solution;
  Mat A; Vec rhs,out,sol,tmp; PetscReal nrm;
  PetscBool known; PetscErrorCode ierr;

  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(system);
  SYSPROCHECKVALIDLINSOL(linsol);
  ierr = LinearSystemGetParts
    (system,&A,PETSC_NULL,&rhs,&sol,PETSC_NULL); CHKERRQ(ierr);
  ierr = LinearSystemGetTmpVector(system,&tmp); CHKERRQ(ierr);
  ierr = LinearSolutionGetVector(linsol,&out); CHKERRQ(ierr);
  ierr = MatMult(A,out,tmp); CHKERRQ(ierr);
  ierr = VecAXPY(tmp,-1,rhs); CHKERRQ(ierr);
  ierr = VecNorm(tmp,NORM_2,&nrm); CHKERRQ(ierr);
  CHKMEMQ;
  printf("== %s:\n",caption);
  printf("Residual 2 norm: %e\n",nrm);
  ierr = LinearSystemGetKnownSolution(system,&known); CHKERRQ(ierr);
  if (known) {
    PetscReal nrm; 
    ierr = LinearSystemTrueDistance(system,linsol,&nrm); CHKERRQ(ierr);
    printf("Distance to true solution: %e\n",nrm);
  }
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PreprocessedLinearSystemSolution"
PetscErrorCode PreprocessedLinearSystemSolution
(LinearSystem sys,LinearSolution *sol)
{
  PetscLogDouble *tsetup; PetscErrorCode ierr;
  PetscFunctionBegin;
  SYSPROCHECKVALIDLINSYS(sys);
  ierr = PetscMalloc(sizeof(PetscLogDouble),&tsetup); CHKERRQ(ierr);
  ierr = RegisterPreprocessorContext("solution",(void*)tsetup); CHKERRQ(ierr);
  CHKMEMQ;
  ierr = PreprocessedProblemSolving
    ((NumericalProblem)sys,(NumericalSolution*)sol); CHKERRQ(ierr);
  CHKMEMQ;
  PetscFunctionReturn(0);
}

#if 0
/*
 * ancillary function
 */
#undef __FUNCT__
#define __FUNCT__ "PurgeAttachedArrays"
PetscErrorCode PurgeAttachedArrays(Mat A)
{
  int icat,icmp; PetscErrorCode ierr;
  PetscFunctionBegin;
  for (icat=0; icat<ncategories; icat++) {
    for (icmp=0; icmp<ncomponents[icat]; icmp++) {
      int id=ids[icat][icmp],*isv; PetscScalar *rsv; 
      PetscBool f=PETSC_FALSE;
      switch (types[icat][icmp]) {
      case ANALYSISINTARRAY :
	ierr = PetscObjectComposedDataGetIntstar
	  ((PetscObject)A,id,isv,f); CHKERRQ(ierr);
	if (f) {
	  ierr = PetscFree(isv); CHKERRQ(ierr);
	  ierr = PetscObjectComposedDataSetIntstar
	    ((PetscObject)A,id,0); CHKERRQ(ierr);
	}
	break;
      case ANALYSISDBLARRAY :
	ierr = PetscObjectComposedDataGetRealstar
	  ((PetscObject)A,id,rsv,f); CHKERRQ(ierr);
	if (f) {
	  ierr = PetscFree(rsv); CHKERRQ(ierr);
	  ierr = PetscObjectComposedDataSetRealstar
	    ((PetscObject)A,id,0); CHKERRQ(ierr);
	}
	break;
      default : ;
      }
    }
  }
  PetscFunctionReturn(0);
}
#endif
