#include <stdlib.h>
#include <stdio.h>
#include "syspro.h"
#include "sysprotransform.h"
#include "syspro_impl.h"

/*! \page options Command line options handling

  Being based on Petsc, SysPro can tailor its workings by commandline options.
  Options are handled by  PreprocessorsOptionsHandling().
  This routine needs to be called
  explicitly by the user, after all calls to DeclarePreprocessor().
  Commandline options can be set from the program source by the Petsc 
  call PetscOptionsSetValue().

  The following commandline options are understood.
  - \c "-syspro_exhaustive" : every preprocessor is cycled exhaustively,
  unless otherwise limited.
  - \c "-syspro_someprocessor exhaustive" : the specified preprocessor
  is tested exhaustively.
  - \c "-syspro_someprocessor choice1,choice2,..." : the specified
  preprocessor takes on the specified values. This induces cycling on only
  the specified preprocessor; if the \c "-syspro_exhaustive" option 
  for exhaustive cycling of all preprocessors is given, the limited cycling
  takes precedence.
  - \c "-syspro_someprocessor not,choice1,choice2,..." : limited cycling 
  is setup, except that the the specified choices will not be used.
  (See TransformObjectsUseOnly() for details.)
  - \c "-syspro_someprocessor_somechoice_values v1,v2,v3,... " : if a 
  preprocessor choice has option values, this sets the values. This is also
  induces exhaustive cycling over this preprocessor. Note: unless the
  cycling is explicitly limited `somechoice' (see the previous item), 
  the exhaustive mode will cycle over all choices of this preprocessor.

  Any preprocessor can declare its own option handler routine. The option
  names it handles can be anything, but should presumably not clash with
  the above formats. E.g., use \c "-syspro_pc_iterative" rather than
  \c "-syspro_pc iterative".
 */

#undef __FUNCT__
#define __FUNCT__ "SysProShowOptions"
/*! Display all available options.
*/
PetscErrorCode SysProShowOptions(MPI_Comm comm)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  PetscPrintf(comm,"Available SysPro options\n  Global:\n");
  PetscPrintf(comm,"  -syspro_exhaustive : fully exhaustive run\n");
  PetscPrintf(comm,"\n  Preprocessors and options:\n");
  {
    const char *aclass,**types; int ntypes; PetscBool flg;
    ierr = StartRetrievingAllPreprocessors
      (&aclass,&types,&ntypes,&flg); CHKERRQ(ierr);
    while (flg) {
      PetscPrintf(comm,"  -syspro_%s : options for category %s\n",aclass,aclass);
      ierr = PetscFree(types); CHKERRQ(ierr);
      ierr = ContinueRetrievingAllPreprocessors
	(&aclass,&types,&ntypes,&flg); CHKERRQ(ierr);
    }
  }
  PetscPrintf(comm,"\n");
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PreprocessorsOptionsHandling"
/*! Process commandline options that control the behaviour of SysPro.
   For more information see \ref options.
*/
PetscErrorCode PreprocessorsOptionsHandling()
{
  SystemPreprocessor pp; char *preprocess;
#define TYPELEN 200
  char type[TYPELEN],values[TYPELEN],option[100];
  PetscBool exhaustive,flg; PetscErrorCode ierr;

  PetscFunctionBegin;
  /* explicit exhaustive run */
  ierr = PetscOptionsHasName
    (PETSC_NULL,"-syspro_exhaustive",&exhaustive); CHKERRQ(ierr);

  /* loop over all preprocessors */
  ierr = GetFirstPreprocessor((const char**)&preprocess); CHKERRQ(ierr);
  while (preprocess) {
    SalsaTransform transform;
    ierr = SystemPreprocessorGetByName(preprocess,&pp); CHKERRQ(ierr);
    ierr = TransformGetByName(preprocess,&transform); CHKERRQ(ierr);
    sprintf(option,"-syspro_%s",preprocess);
    ierr = PetscOptionsGetString
      (PETSC_NULL,option,type,TYPELEN,&flg); CHKERRQ(ierr);
    if (flg) {
      int n;
      /*printf("syspro option: <%s>=<%s>\n",preprocess,type);*/
      ierr = PetscStrcmp(type,"exhaustive",&flg); CHKERRQ(ierr);
      if (flg) {
	pp->exhaustive = PETSC_TRUE;
	ierr = TransformSetUserChoices(transform,PETSC_FALSE); CHKERRQ(ierr);
      } else {
	ierr = TransformSetUserChoices(transform,PETSC_TRUE); CHKERRQ(ierr);
	ierr = TransformObjectsUseOnly(transform,type); CHKERRQ(ierr);
	ierr = TransformGetNUnmarked(transform,&n); CHKERRQ(ierr);
	pp->exhaustive = TRUTH(n>1);
      }
    } else {
      pp->exhaustive = exhaustive;
      ierr = TransformSetUserChoices(transform,PETSC_FALSE); CHKERRQ(ierr);
    }
    { /* handle options, for instance \c "-syspro_ksp_gmres_values 5,10,20" */
      SalsaTransformObject to;
      ierr = TransformGetNextUnmarkedItem
	(transform,NULL,&to,&flg); CHKERRQ(ierr);
      while (flg) {
	const char *name;
	ierr = TransformObjectGetName(to,&name); CHKERRQ(ierr);
	sprintf(option,"-syspro_%s_%s_values",preprocess,name); CHKERRQ(ierr);
	ierr = PetscOptionsGetString
	  (PETSC_NULL,option,values,TYPELEN,&flg); CHKERRQ(ierr);
	if (flg) {
	  printf("syspro option values: <%s:%s>=<%s>\n",preprocess,name,values);
	  ierr = TransformItemOptionsUseOnly(to,values); CHKERRQ(ierr);
	  pp->exhaustive = PETSC_TRUE;
	}
	ierr = TransformGetNextUnmarkedItem
	  (transform,name,&to,&flg); CHKERRQ(ierr);
      }
    }
    if (pp->optionshandling) {
      ierr = (pp->optionshandling)(); CHKERRQ(ierr);
    }
    ierr = PreprocessorSaveAprioriSelection(pp); CHKERRQ(ierr);
    ierr = ReportEnabledPreprocessors(preprocess); CHKERRQ(ierr);
    ierr = GetNextPreprocessor((const char**)&preprocess); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

