/*! \file scaling.c \ingroup linear */

/*! \page scaling Scale a linear system

  This preprocessor can perform pointwise left, right, and symmetric scalings
  of a linear system by the diagonal of its coefficient matrix.
*/
#include <stdlib.h>
#include <stdio.h>
#include "petsc.h"
#include "syspro.h"
#include "sysprotransform.h"
#include "sysprolinear.h"
#include "petscmat.h"

#define PREPROCESSOR "scaling"

#undef __FUNCT__
#define __FUNCT__ "set_intelligent_scaling"
static PetscErrorCode set_intelligent_scaling
(NumericalProblem theproblem,const char **type,const char **reason)
{
  PetscBool f1,f2,flg; PetscReal r1,r2; PetscErrorCode ierr;

  PetscFunctionBegin;

  ierr = SysProRetrieveQuantity
    (theproblem,"variance","row-variability",(void*)&r1,PETSC_NULL,
     &f1); CHKERRQ(ierr);
  ierr = SysProRetrieveQuantity
    (theproblem,"variance","col-variability",(void*)&r2,PETSC_NULL,
     &f2); CHKERRQ(ierr);
  flg = TRUTH(f1 && f2);
  if (flg) {
    if (r1>r2+10) {
      *type = "right";
      *reason = "trying to tame extreme variability within rows";
    } else if (r2>r1+10) {
      *type = "left";
      *reason = "trying to tame extreme variability within columns";
    } else {
      *type = "symmetric";
      *reason = "rows and columns are already pretty equilibrated";
    }
  } else {
    ierr = SysProRetrieveQuantity
      (theproblem,"variance","diagonal-average",(void*)&r1,PETSC_NULL,
       &f1); CHKERRQ(ierr);
    ierr = SysProRetrieveQuantity
      (theproblem,"variance","diagonal-variance",(void*)&r2,PETSC_NULL,
       &f2); CHKERRQ(ierr);
    if (f1 && f2 && r2<r1/20) {
      *type = "none";
      *reason = "not enough variability to justify scaling";
    } else {
      *type = "symmetric";
      *reason = "symmetric scaling is always better than nothing";
    }
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "scale_system"
static PetscErrorCode scale_system
(const char *type,int nopt,PetscBool overwrite,
 NumericalProblem inproblem,NumericalProblem *outproblem,
 void *gctx,void **ctx, PetscBool *success)
{
  LinearSystem newproblem,problem = (LinearSystem)inproblem;
  Vec D; PetscBool flg; PetscErrorCode ierr; 
  PetscFunctionBegin;

  ierr = LinearSystemDuplicate(problem,&newproblem); CHKERRQ(ierr);
  ierr = LinearSystemCopy(problem,newproblem); CHKERRQ(ierr);

  *success = PETSC_TRUE;
  
#if 0
  /*
   * Scaling leaves the structure intact
   */
  /*
  {
    NMD_metadata nmd_in,nmd_out;

    ierr = LinearSystemGetMetadata(problem,&nmd_in); CHKERRQ(ierr);
    ierr = LinearSystemGetMetadata(newproblem,&nmd_out); CHKERRQ(ierr);
    ierr = SysProTraceMessage("Copying structure categories\n");
    ierr = NMDHasCategory(nmd_in,"structure",(int*)&flg); CHKERRQ(ierr);
    if (flg) {
      ierr = NMDCopyCategory(nmd_in,nmd_out,"structure"); CHKERRQ(ierr);}
    ierr = NMDHasCategory(nmd_in,"iprs",(int*)&flg); CHKERRQ(ierr);
    if (flg) {
      ierr = NMDCopyCategory(nmd_in,nmd_out,"iprs"); CHKERRQ(ierr);}
    ierr = NMDHasCategory(nmd_in,"jpl",(int*)&flg); CHKERRQ(ierr);
    if (flg) {
      ierr = NMDCopyCategory(nmd_in,nmd_out,"jpl"); CHKERRQ(ierr);}
    ierr = NMDHasCategory(nmd_in,"icmk",(int*)&flg); CHKERRQ(ierr);
    if (flg) {
      ierr = NMDCopyCategory(nmd_in,nmd_out,"icmk"); CHKERRQ(ierr);}
  }
  */
#endif

  ierr = PetscStrcmp(type,"none",&flg); CHKERRQ(ierr);
  if (flg)
    goto done;

  /*
   * construct the scaling
   */
  {
    Mat A; int local_size; MPI_Comm comm;

    ierr = LinearSystemGetParts
      (problem,&A,PETSC_NULL,PETSC_NULL,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);
    ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
    ierr = MatGetLocalSize(A,&local_size,&ierr); CHKERRQ(ierr);
    ierr = VecCreateMPI(comm,local_size,PETSC_DECIDE,&D); CHKERRQ(ierr);
    ierr = MatGetDiagonal(A,D); CHKERRQ(ierr);
    { /* prevent problems with zero diagonal elements */
      PetscScalar *a; int local_size,i;
      ierr = VecGetArray(D,&a); CHKERRQ(ierr);
      ierr = MatGetLocalSize(A,&local_size,&i); CHKERRQ(ierr);
      for (i=0; i<local_size; i++)
	if (a[i]==0.) a[i]=1.;
      ierr = VecRestoreArray(D,&a); CHKERRQ(ierr);
    }
    ierr = VecReciprocal(D); CHKERRQ(ierr);
    *(Vec*)ctx = D;
  }

  /*
   * do the actual scaling
   */
  {
    Mat A,B; Vec rhs,sol,init;
    ierr = LinearSystemGetParts
      (newproblem,&A,&B,&rhs,&sol,&init); CHKERRQ(ierr);

    ierr = PetscStrcmp(type,"symmetric",&flg); CHKERRQ(ierr);
    if (flg) {
      ierr = VecSqrtAbs(D); CHKERRQ(ierr);
      ierr = MatDiagonalScale(A,D,D); CHKERRQ(ierr);
      if (B && B!=A) {
	ierr = MatDiagonalScale(B,D,D); CHKERRQ(ierr);
      }
      ierr = VecPointwiseMult(rhs,D,rhs); CHKERRQ(ierr);
      ierr = VecPointwiseDivide(sol,sol,D); CHKERRQ(ierr);
      if (init) {
	ierr = VecPointwiseDivide(init,init,D); CHKERRQ(ierr);}
      goto done;
    }
    
    ierr = PetscStrcmp(type,"left",&flg); CHKERRQ(ierr);
    if (flg) {
      ierr = MatDiagonalScale(A,D,PETSC_NULL); CHKERRQ(ierr);
      if (B && B!=A) {
	ierr = MatDiagonalScale(B,D,PETSC_NULL); CHKERRQ(ierr);
      }
      ierr = VecPointwiseMult(rhs,D,rhs); CHKERRQ(ierr);
      goto done;
    }
    
    ierr = PetscStrcmp(type,"right",&flg); CHKERRQ(ierr);
    if (flg) {
      ierr = MatDiagonalScale(A,PETSC_NULL,D); CHKERRQ(ierr);
      if (B && B!=A) {
	ierr = MatDiagonalScale(B,PETSC_NULL,D); CHKERRQ(ierr);
      }
      ierr = VecPointwiseDivide(sol,sol,D); CHKERRQ(ierr);
      if (init) {
	ierr = VecPointwiseDivide(init,init,D); CHKERRQ(ierr);}
      goto done;
    }
    
    SETERRQ1(PETSC_COMM_SELF,1,"scaling type not implemented: %s",type);
  }

 done:
  *outproblem = (NumericalProblem)newproblem;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "unscale_system"
static PetscErrorCode unscale_system
(const char *scaling_type,PetscBool overwrite,void *gctx,void *ctx,
 NumericalProblem problem,NumericalProblem nextproblem,
 NumericalSolution scaled,NumericalSolution unscaled)
{
  LinearSolution old = (LinearSolution)scaled,lnew = (LinearSolution)unscaled;
  Vec D = (Vec)ctx;
  Vec X,Xscaled; PetscBool fnone,f1,f2; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = LinearSolutionCopyStats(old,lnew); CHKERRQ(ierr);
  ierr = LinearSolutionGetVector(old,&Xscaled); CHKERRQ(ierr);
  ierr = LinearSolutionGetVector(lnew,&X); CHKERRQ(ierr);
  ierr = VecCopy(Xscaled,X); CHKERRQ(ierr);

  ierr = PetscStrcmp(scaling_type,"none",&f1); CHKERRQ(ierr);
  fnone = f1;
  ierr = PetscStrcmp(scaling_type,"left",&f2); CHKERRQ(ierr);
  if (f1 || f2) {
    ierr = VecCopy(Xscaled,X); CHKERRQ(ierr);
  } else {
    ierr = PetscStrcmp(scaling_type,"right",&f1); CHKERRQ(ierr);
    ierr = PetscStrcmp(scaling_type,"symmetric",&f2); CHKERRQ(ierr);
    if (f1 || f2) {
      ierr = VecPointwiseMult(X,Xscaled,D); CHKERRQ(ierr);
    } else {
      SETERRQ1(PETSC_COMM_SELF,1,"Unknown scaling <%s>",scaling_type); 
    }
  }
  if (!fnone) {
    ierr = VecDestroy(&D); CHKERRQ(ierr);}

  /*ierr = LinearSolutionDelete((LinearSolution)scaled); CHKERRQ(ierr);*/
  //ierr = DeleteLinearSystem((LinearSystem)problem); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "setup_scaling_choices"
/*! This routine is called by DeclarePreprocessor() */
static PetscErrorCode setup_scaling_choices()
{
  SalsaTransform tf; SalsaTransformObject cur; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = TransformGetByName(PREPROCESSOR,&tf); CHKERRQ(ierr);
  ierr = SysProDefineIntAnnotation(PREPROCESSOR,"symmetric"); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"none",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation(cur,"no scaling"); CHKERRQ(ierr);
  ierr = TransformObjectIntAnnotate(cur,"symmetric",1); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"symmetric",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation(cur,"symmetric scaling"); CHKERRQ(ierr);
  ierr = TransformObjectIntAnnotate(cur,"symmetric",1); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"left",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation(cur,"left scaling"); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"right",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation(cur,"right scaling"); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "specific_scaling_choices"
/*!
  This is the 'specific setup' phase of the scaling preprocessor.
  See \ref modes for details.

  This routine eliminates unsymmetric scalings if we are dealing with a 
  symmetric system.
*/
static PetscErrorCode specific_scaling_choices
(NumericalProblem theproblem,SalsaTransform scaling)
{
  int isymm; PetscBool flg; PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = SysProRetrieveQuantity
    (theproblem,"structure","symmetry",(void*)&isymm,PETSC_NULL,
     &flg); CHKERRQ(ierr);
  if (flg && isymm>0) {
    int i,n; SalsaTransformObject *objs;
    ierr = TransformGetObjects(scaling,&n,&objs); CHKERRQ(ierr);
    for (i=0; i<n; i++) {
      SalsaTransformObject cur = objs[i]; 
      int v; PetscBool f;
      ierr = TransformObjectGetIntAnnotation
	(cur,"symmetric",&v,&f); CHKERRQ(ierr);
      if (f && !v) {
	ierr = TransformObjectMark(cur); CHKERRQ(ierr);
      }
    }
  }

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeclareScalingPreprocessor"
PetscErrorCode DeclareScalingPreprocessor(void)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = DeclarePreprocessor
    (PREPROCESSOR,
     &setup_scaling_choices, /* declare all scalings */
     &specific_scaling_choices, /* metadata inspection to set intell choice */
     0,0, /* local and global unset */
     0,0, /* context create and destroy */
     &scale_system, /* start function */
     &unscale_system /* end function */); CHKERRQ(ierr);
  ierr = DeclarePreprocessorIntelligentChoice
    ("scaling",&set_intelligent_scaling); CHKERRQ(ierr);
  ierr = PreprocessorSetPreservedCategories
    (PREPROCESSOR,"laplace,structure,iprs,icmk"); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
