/*! \file singleton.c \ingroup linear */

/*! \page singleton Eliminate fully determined (singleton) rows from a matrix

  In a linear system, any variable whose matrix row has only a single
  element is independent of the rest of the system, in the sense that the 
  other variables can be solved independently of it. The singleton
  preprocessor eliminates such rows, and performs the backsubstitution.
 */
#include <stdlib.h>
#include <stdio.h>
#include "petsc.h"
#include "syspro.h"
#include "sysprotransform.h"
#include "sysprolinear.h"
#include "linear_impl.h"

typedef struct {int n,t; VecScatter extractor;} singleton_struct;
#define PREPROCESSOR "singleton"

#undef __FUNCT__
#define __FUNCT__ "eliminate_singletons"
static PetscErrorCode eliminate_singletons
(const char *type,int nopt,PetscBool overwrite,
 NumericalProblem inproblem,NumericalProblem *outproblem,
 void *gctx,void **ctx,PetscBool *success)
{
  LinearSystem problem = (LinearSystem)inproblem, newproblem;
  PetscBool flg; PetscErrorCode ierr; 

  PetscFunctionBegin;

  *success = PETSC_TRUE; *(Vec*)ctx = NULL;
  ierr = LinearSystemDuplicatePointers(problem,&newproblem); CHKERRQ(ierr);

  ierr = PetscStrcmp(type,"identity",&flg); CHKERRQ(ierr);
  if (flg) {
    goto done;
  }


  ierr = PetscStrcmp(type,"eliminate",&flg); CHKERRQ(ierr);
  if (flg) {
    Mat A,Ause, B,Buse; Vec X,Xuse, V,Vuse, I,Iuse;
    MPI_Comm comm;
    IS dummy_rows,all_rows,true_rows,gtrue_rows,compact_rows;
    VecScatter compact_extractor;
    int ndummy,dummy_kind,first,last,local_size,compact_size,nd,*dd;
    singleton_struct *savestruct;

    ierr = LinearSystemGetParts(problem,&A,&B,&V,&X,&I); CHKERRQ(ierr);
    ierr = PetscObjectGetComm((PetscObject)A,&comm); CHKERRQ(ierr);
    ierr = PetscMalloc(sizeof(singleton_struct),&savestruct); CHKERRQ(ierr);
    savestruct->n = 1;

    /*
     * Get dummy row information
     */
    ierr = SysProRetrieveQuantity
      (inproblem,"structure","dummy-rows",(void*)&dd,PETSC_NULL,
       &flg); CHKERRQ(ierr);
    if (!flg) goto no_compute;
    ierr = SysProRetrieveQuantity
      (inproblem,"structure","n-dummy-rows",(void*)&ndummy,PETSC_NULL,
       &flg); CHKERRQ(ierr);
    if (!flg) goto no_compute;
    ierr = SysProRetrieveQuantity
      (inproblem,"structure","dummy-rows-kind",(void*)&dummy_kind,PETSC_NULL,
       &flg); CHKERRQ(ierr);
    if (!flg) goto no_compute;
    savestruct->t = dummy_kind;
    goto eliminate;

  no_compute:
    /* 
     * No dummy row information; take default action, then exit
     * WRONG this will crash in the back substitution
     */
    goto done;
  eliminate:
    /*
     * Eliminate dummy rows
     */
#if 0
    if (ndummy!=dd[0])
      SETERRQ2(1,"Dummy rows size mismatch",ndummy,dd[0]);
    dd++; /* skip past zero element which is array size */
#endif
    ierr = MatGetOwnershipRange(A,&first,&last); CHKERRQ(ierr);
    local_size = last-first;
    ierr = ISCreateStride
      (comm,local_size,first,1,&all_rows); CHKERRQ(ierr);
    for (nd=0; nd<ndummy; nd++) if (dd[nd]>=first) break;
    dd = dd+nd; ndummy -= nd;
    for (nd=ndummy-1; nd>=0; nd--) if (dd[nd]>=last) ndummy--; else break;
    compact_size = local_size-ndummy;
    ierr = ISCreateGeneral
      (comm,ndummy,dd,PETSC_OWN_POINTER,&dummy_rows); CHKERRQ(ierr);
    printf("dummy rows\n"); ISView(dummy_rows,0);
    ierr = ISDifference(all_rows,dummy_rows,&true_rows); CHKERRQ(ierr);
    ierr = ISAllGather(true_rows,&gtrue_rows); CHKERRQ(ierr);
    
    /*
     * Extract the true submatrix
     */
    ierr = MatGetSubMatrix
      (A,true_rows,gtrue_rows,MAT_INITIAL_MATRIX,
       &Ause); CHKERRQ(ierr);
    if (B && B!=A) {
      ierr = MatGetSubMatrix
	(B,true_rows,gtrue_rows,MAT_INITIAL_MATRIX,
	 &Buse); CHKERRQ(ierr);
    } else Buse = Ause;
    ierr = ISDestroy(&gtrue_rows); CHKERRQ(ierr);
    /* and find out its distribution */
    ierr = MatGetOwnershipRange(Ause,&first,&last); CHKERRQ(ierr);
    
    /*
     * Extract the true vector
     */
    /* create compact vector */
    ierr = VecCreateMPI
      (comm,compact_size,PETSC_DECIDE,&Xuse); CHKERRQ(ierr);
    ierr = VecDuplicate(Xuse,&Vuse); CHKERRQ(ierr);
    ierr = VecDuplicate(Xuse,&Iuse); CHKERRQ(ierr);

    /* create scatter */
    ierr = ISCreateStride
      (MPI_COMM_SELF,compact_size,first,1,&compact_rows); CHKERRQ(ierr);
    ierr = VecScatterCreate
      (X,true_rows,Xuse,compact_rows,&compact_extractor); CHKERRQ(ierr);
    savestruct->extractor = compact_extractor;
    ierr = ISDestroy(&compact_rows); CHKERRQ(ierr);
    ierr = ISDestroy(&dummy_rows); CHKERRQ(ierr);
    ierr = ISDestroy(&all_rows); CHKERRQ(ierr);
    ierr = ISDestroy(&true_rows); CHKERRQ(ierr);

    /* compact rhs */
    ierr = VecScatterBegin
      (compact_extractor,V,Vuse,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    ierr = VecScatterEnd
      (compact_extractor,V,Vuse,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    /* compact true solution */
    ierr = VecScatterBegin
      (compact_extractor,X,Xuse,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    ierr = VecScatterEnd
      (compact_extractor,X,Xuse,INSERT_VALUES,SCATTER_FORWARD); CHKERRQ(ierr);
    /* compact starting guess */
    if (I) {
      ierr = VecScatterBegin
	(compact_extractor,I,Iuse,INSERT_VALUES,SCATTER_FORWARD);CHKERRQ(ierr);
      ierr = VecScatterEnd
	(compact_extractor,I,Iuse,INSERT_VALUES,SCATTER_FORWARD);CHKERRQ(ierr);
    }
    ierr = LinearSystemSetParts
      (newproblem,Ause,Buse,Vuse,Xuse,Iuse); CHKERRQ(ierr);
    {
      void *info;
      ierr = LinearSystemGetContext(problem,&info); CHKERRQ(ierr);
      ierr = LinearSystemSetContext(newproblem,info); CHKERRQ(ierr);
    }
    *(singleton_struct**)ctx = savestruct;
    goto done;
  }
  
  SETERRQ1(PETSC_COMM_SELF,1,"Unknown singleton preprocessor <%s>",type);
 done:
  *outproblem = (NumericalProblem)newproblem;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "back_singleton"
static PetscErrorCode back_singleton
(const char *singleton_type,PetscBool overwrite,void *gctx,void *ctx,
 NumericalProblem compactproblem,NumericalProblem fullproblem,
 NumericalSolution compactvector,NumericalSolution fullvector)
{
  LinearSolution compactsolution = (LinearSolution)compactvector,
    fullsolution = (LinearSolution)fullvector;
  Vec Xcompact,Xfull,Rhs;
  singleton_struct* savestruct; int dummy_kind;
  VecScatter compact_extractor;
  PetscBool flg; PetscErrorCode ierr;

  PetscFunctionBegin;

  ierr = LinearSolutionCopyStats(compactsolution,fullsolution); CHKERRQ(ierr);
  ierr = LinearSolutionGetVector(compactsolution,&Xcompact); CHKERRQ(ierr);
  ierr = LinearSolutionGetVector(fullsolution,&Xfull); CHKERRQ(ierr);

  ierr = PetscStrcmp(singleton_type,"identity",&flg); CHKERRQ(ierr);
  if (flg) {
    ierr = VecCopy(Xcompact,Xfull); CHKERRQ(ierr);
  } else {
    ierr = LinearSystemGetParts
      ((LinearSystem)fullproblem,PETSC_NULL,PETSC_NULL,&Rhs,PETSC_NULL,PETSC_NULL); CHKERRQ(ierr);

    savestruct = (singleton_struct*) ctx;
    dummy_kind = savestruct->t;
    if (dummy_kind>0)
      printf("\n\n>>>> WARNING Cannot properly handle dummy rows of kind %d <<<<\n\n",dummy_kind);
    compact_extractor = savestruct->extractor;

    /*
     * first copy dummy rows from rhs to solution
     */
    ierr = VecDuplicate(Rhs,&Xfull); CHKERRQ(ierr);
    ierr = VecCopy(Rhs,Xfull); CHKERRQ(ierr);
    /* then insert computed solution on non-dummy variables */
    ierr = VecScatterBegin(compact_extractor,Xcompact,Xfull,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
    ierr = VecScatterEnd  (compact_extractor,Xcompact,Xfull,INSERT_VALUES,SCATTER_REVERSE); CHKERRQ(ierr);
    ierr = LinearSolutionSetVector(fullsolution,Xfull); CHKERRQ(ierr);

    ierr = VecScatterDestroy(&compact_extractor); CHKERRQ(ierr);
    ierr = PetscFree(savestruct); CHKERRQ(ierr);
  }
  
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "setup_singleton_choices"
/*! This routine is only called when the singleton preprocessor
  is created by DeclarePreprocessor() inside DeclareSingletonPreprocessor()
*/
static PetscErrorCode setup_singleton_choices()
{
  SalsaTransform tf; SalsaTransformObject cur; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = TransformGetByName(PREPROCESSOR,&tf); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"identity",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"no singletons"); CHKERRQ(ierr);

  ierr = NewTransformObject(PREPROCESSOR,"eliminate",&cur); CHKERRQ(ierr);
  ierr = TransformObjectSetExplanation
    (cur,"eliminate singletons"); CHKERRQ(ierr);

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "specific_singleton_choices"
/*!
  This is the 'specific setup' phase of the singleton preprocessor.
  See \ref modes for details.

  It disables either the identity or the elimination routine, to
  leave only the one applicable to this particular system.

  Maybe if we'd ever want to prove how effective singleton elimination
  is, we could leave identity in place for systems with singletons.
*/
static PetscErrorCode specific_singleton_choices
(NumericalProblem theproblem,SalsaTransform singleton)
{
  SalsaTransformObject tf; PetscBool flg; int ndummy,nnz; 
  PetscErrorCode ierr;
  PetscFunctionBegin;

  /*
   * Mark (as unsuitable) the identity if there are dummy rows,
   * and the matrix is not diagonal.
   */
  ierr = SysProComputeQuantity
    (theproblem,"structure","n-dummy-rows",(void*)&ndummy,PETSC_NULL,
     &flg); CHKERRQ(ierr);
  if (flg && ndummy>0) {
    ierr = SysProComputeQuantity
      (theproblem,"structure","nnzeros",(void*)&nnz,PETSC_NULL,
       &flg); CHKERRQ(ierr);
    if (flg && nnz>ndummy) {
      ierr = TransformObjectGetByName
	(PREPROCESSOR,"identity",&tf); CHKERRQ(ierr);
      ierr = TransformObjectMark(tf); CHKERRQ(ierr);
      goto done;
    }
  }

  /* Otherwise, mark elimination for exclusion */
  ierr = TransformObjectGetByName(PREPROCESSOR,"eliminate",&tf); CHKERRQ(ierr);
  ierr = TransformObjectMark(tf); CHKERRQ(ierr);
 done:

  PetscFunctionReturn(0);
}

static PetscErrorCode singleton_specific_unset(NumericalProblem theproblem)
{
  PetscErrorCode ierr; PetscBool flg;
  PetscFunctionBegin;
  ierr = SysProRemoveQuantity
    (theproblem,"structure","n-dummy-rows",&flg); CHKERRQ(ierr);
  ierr = SysProRemoveQuantity
    (theproblem,"structure","nnzeros",&flg); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "DeclareSingletonPreprocessor"
PetscErrorCode DeclareSingletonPreprocessor(void)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = DeclarePreprocessor
    (PREPROCESSOR,
     &setup_singleton_choices,&specific_singleton_choices,
     &singleton_specific_unset,0,
     0,0,
     &eliminate_singletons,&back_singleton); CHKERRQ(ierr);
  ierr = PreprocessorSetPreservedCategories
    (PREPROCESSOR,"laplace"); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}
