#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "syspro.h"

/*! \page tracing Tracing the preprocessors
  The SysPro package does not by default print out anything, other than
  severe error messages (Petsc macro SETERRQ) that accompany an abort.

  However, you can specify a trace function, which
  can further be tuned by specifying a trace context.

  See SysProDeclareTraceFunction(), SysProDeclareTraceContext(),
  SysProTraceMessage().
*/

static PetscErrorCode(*sysprotrace)(void*,const char *fmt,va_list) = NULL;
static size_t sysprotracectx = (size_t)NULL;

#undef __FUNCT__
#define __FUNCT__ "SysProDefaultTrace"
PetscErrorCode SysProDefaultTrace(void *ctx,const char *fmt,va_list argp)
{
  /*char *prefix = (char*)ctx;*/
  PetscFunctionBegin;
  printf("SysPro message: ");
  vprintf(fmt, argp);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProDeclareTraceFunction"
/*! Specify a trace function.

  The trace function has a prototype 
\verbatim
  PetscErrorCode tracefunction(void*,char*,va_list)
\endverbatim
  which means that it has an arbitrary number of arguments, much like
  \c printf. The first argument is a context, which can be set by 
  SysProDeclareTraceContext().

  Here is an example of how you would write a trace function:
\verbatim
#include <stdarg.h>
PetscErrorCode tracefunction(void *ctx,char *fmt,va_list argp)
{
  char *prefix = (char*)ctx;
  PetscFunctionBegin;
  printf("%s ",prefix);
  vprintf(fmt, argp);
  PetscFunctionReturn(0);
}
\endverbatim
  Consult \c string.h (probably in \c /usr/include) to see which "v"
  versions of \c printf are available.

  There is a default trace function SysProDefaultTrace().

  You can undeclare a trace function by passing NULL.

  See also SysProTraceMessage().
*/
PetscErrorCode SysProDeclareTraceFunction(PetscErrorCode(*fn)(void*,const char*,va_list))
{
  PetscFunctionBegin;
  sysprotrace = fn;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProDeclareTraceContext"
/* Specify a trace context.
  
  See SysProDeclareTraceFunction().
*/
PetscErrorCode SysProDeclareTraceContext(void *ctx)
{
  PetscFunctionBegin;
  sysprotracectx = (size_t)ctx;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProTraceMessage"
/*! This function prints a trace message if a trace function has been
  declared; see SysProDeclareTraceFunction().
*/
PetscErrorCode SysProTraceMessage(const char *fmt,...)
{
  va_list argp; PetscErrorCode ierr;
  PetscFunctionBegin;
  if (sysprotrace) {
    va_start(argp, fmt);
    ierr = (*sysprotrace)((void*)sysprotracectx,fmt,argp); CHKERRQ(ierr);
    va_end(argp);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProHasTrace"
/*! Test whether a trace function has been declared; 
  see SysProDeclareTraceFunction(). Normally you would use SysProTraceMessage()
  which performs this test internally,
  but this function can be useful if a large amount of processing
  has to be performed to construct the trace message to begin with.
*/
PetscErrorCode SysProHasTrace(PetscBool *flg)
{
  PetscFunctionBegin;
  if (sysprotrace)
    *flg = PETSC_TRUE;
  else
    *flg = PETSC_FALSE;
  PetscFunctionReturn(0);
}

