#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "petsc.h"
#include "sysprotransform.h"
#include "syspro_impl.h"

struct SalsaTransformObject_ {
  int cookie;
  const char *name,*explanation;
  SalsaTransform transform;
  int n_options,alloc_options,*options; const char *option,**optionexplanation;
  int *options_marked; int active_option;
  int alloc_annotate_c; const char **annotate_c;
  int alloc_annotate_i,*annotate_i;
  int marked;
  PetscErrorCode(*suitabilityfunction)
    (NumericalProblem,void*,SuitabilityValue*);
  void *suitabilityctx;
};

#define TFINC 20

/*! \page transform Preprocessor definition

  \section preprocessorclass Class definition

  A class of preprocessors (such as scaling, preconditioning) is defined
  using the function NewTransform().

  \section preprocessorindividual Individual preprocessor definition

  The individual preprocessors (left scaling, preprocessing by ILU)
  are defined in a function that is passed as the \c specific_setup
  argument to DeclarePreprocessor(). This function makes calls to 
  NewTransformObject(), TransformObjectIntAnnotate() et cetera.
  See for instance file \ref pc.c .
 */

#undef __FUNCT__
#define __FUNCT__ "NewTransform"
/*! Define a new class of preprocessors, for instance scaling 
  or permutation.
*/
PetscErrorCode NewTransform(const char *name,SalsaTransform *tf)
{
  SalsaTransform snew; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = PetscNew(struct SalsaTransform_,&snew); CHKERRQ(ierr);
  ierr = PetscMemzero(snew,sizeof(struct SalsaTransform_)); CHKERRQ(ierr);
  snew->name = strdup(name); snew->cookie = TRANSFORMCOOKIE;

  snew->alloc_objects = TFINC; snew->n_objects = 0;
  ierr = PetscMalloc(snew->alloc_objects*sizeof(SalsaTransformObject),&snew->transformobjects); CHKERRQ(ierr);
  ierr = PetscMemzero(snew->transformobjects,snew->alloc_objects*sizeof(SalsaTransformObject)); CHKERRQ(ierr);
  ierr = PetscMalloc(snew->alloc_objects*sizeof(int),&snew->aprioriselection); CHKERRQ(ierr);
  ierr = PetscMemzero(snew->aprioriselection,snew->alloc_objects*sizeof(int)); CHKERRQ(ierr);

  *tf = snew;
  PetscFunctionReturn(0);
}

PetscErrorCode DeregisterTransform(SalsaTransform tf)
{
  int i; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  for (i=0; i<tf->n_objects; i++) {
    ierr = FreeTransformObject(tf->transformobjects[i]); CHKERRQ(ierr);
  }
  ierr = PetscFree(tf->transformobjects); CHKERRQ(ierr);
  ierr = PetscFree(tf->aprioriselection); CHKERRQ(ierr);
  if (tf->annotations_i) {ierr = PetscFree(tf->annotations_i); CHKERRQ(ierr);}
  if (tf->annotations_c) {ierr = PetscFree(tf->annotations_c); CHKERRQ(ierr);}
  ierr = PetscFree(tf); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformGetName"
PetscErrorCode TransformGetName(SalsaTransform tf,const char **name)
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  *name = tf->name;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformGetObjects"
PetscErrorCode TransformGetObjects
(SalsaTransform tf,int *n,SalsaTransformObject **objs)
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  if (n) *n = tf->n_objects;
  if (objs) *objs = tf->transformobjects;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "NewTransformObject"
/*! Create a transform object specified by \c name 
   for the preprocessor class \c transform.
   The \c to parameter can be NULL if no further specifications
   of the object are needed, in which case this only registers the name.
*/
PetscErrorCode NewTransformObject
(const char *transform,const char *name,SalsaTransformObject *to)
{
  SalsaTransform tf; SalsaTransformObject snew; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = TransformGetByName(transform,&tf); CHKERRQ(ierr);
  ierr = PetscNew(struct SalsaTransformObject_,&snew); CHKERRQ(ierr);
  ierr = PetscMemzero(snew,sizeof(struct SalsaTransformObject_)); CHKERRQ(ierr);
  ierr = PetscStrallocpy(name,(char**)&(snew->name)); snew->transform = tf;
  snew->cookie = TRANSFORMOBJECTCOOKIE;
  if (tf->n_objects>=tf->alloc_objects)
    SETERRQ(PETSC_COMM_SELF,1,"Can not realloc transform objects");
  tf->transformobjects[tf->n_objects++] = snew;
  if (to) *to = snew;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "FreeTransformObject"
PetscErrorCode FreeTransformObject(SalsaTransformObject tf)
{
  int iopt; PetscErrorCode ierr;
  CHECKVALIDTRANSFORMOBJECT(tf);
  PetscFunctionBegin;
  if (tf->options) {
    ierr = PetscFree(tf->options); CHKERRQ(ierr);
    ierr = PetscFree(tf->options_marked); CHKERRQ(ierr);
    for (iopt=0; iopt<tf->n_options; iopt++)
      free((char*)tf->optionexplanation[iopt]);
    ierr = PetscFree(tf->optionexplanation); CHKERRQ(ierr);
  }
  if (tf->annotate_i) {ierr = PetscFree(tf->annotate_i); CHKERRQ(ierr);}
  if (tf->annotate_c) {ierr = PetscFree(tf->annotate_c); CHKERRQ(ierr);}
  ierr = PetscFree(tf->name); CHKERRQ(ierr); //free(tf->name);
  ierr = PetscFree(tf); CHKERRQ(ierr);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectGetName"
PetscErrorCode TransformObjectGetName(SalsaTransformObject tf,const char **name)
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  *name = tf->name;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectSetExplanation"
PetscErrorCode TransformObjectSetExplanation
(SalsaTransformObject tf,const char *x)
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  tf->explanation = x;
  PetscFunctionReturn(0);
}

/*! \page suit Suitability functions

  The general mechanism form choosing between algorithms is that of
  `suitability functions'. We associate with each specific preprocessor
  (for instance scaling/left) a function that returns either a fuzzy truth
  value (0--1) or `unknown' (-1). See TransformObjectSetSuitabilityFunction(),
  TransformObjectGetSuitabilityFunction(). See also \ref transform about 
  specific preprocessor construction.

  At the start of a preprocessor invocation,
  in PreprocessorSpecificSetup(), the suitability functions of all choices are 
  evaluated, and the choices are marked as unsuitable (if the evaluate
  is zero), or ranked otherwise.

  The only implemented suitability functions are for the linear problem;
  see \ref suitc.
 */

#undef __FUNCT__
#define __FUNCT__ "TransformObjectSetSuitabilityFunction"
/*! Set the suitability function; see \ref suit.

  See also TransformObjectGetSuitabilityFunction().
*/
PetscErrorCode TransformObjectSetSuitabilityFunction
(SalsaTransformObject tf,void *sctx,
 PetscErrorCode(*f)(NumericalProblem,void*,SuitabilityValue*))
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  tf->suitabilityfunction = f;
  tf->suitabilityctx = sctx;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectGetSuitabilityFunction"
/*! Retrieve the suitability function and context; see \ref suit. Both
  arguments can be null.

  See also TransformObjectSetSuitabilityFunction().
 */
PetscErrorCode TransformObjectGetSuitabilityFunction
(SalsaTransformObject tf,void **sctx,
 PetscErrorCode(**f)(NumericalProblem,void*,SuitabilityValue*))
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  if (sctx) *sctx = tf->suitabilityctx;
  if (f) *f = tf->suitabilityfunction;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectGetTransformName"
PetscErrorCode TransformObjectGetTransformName(SalsaTransformObject tf,char **name)
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  *name = tf->transform->name;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectGetByName"
PetscErrorCode TransformObjectGetByName
(const char *trans,const char *name,SalsaTransformObject *tf)
{
  SalsaTransform tt; int i; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = TransformGetByName(trans,&tt); CHKERRQ(ierr);
  for (i=0; i<tt->n_objects; i++) {
    if (!strcmp(name,tt->transformobjects[i]->name)) {
      *tf = tt->transformobjects[i];
      CHECKVALIDTRANSFORMOBJECT(*tf);
      PetscFunctionReturn(0);
    }
  }
  SETERRQ2(PETSC_COMM_SELF,1,"Could not find transform <%s:%s>",trans,name);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectsGetNames"
/*!
  Get the names of all declared transformobjects. An array is allocated
  for the names, which needs to be PetscFree()'d.
*/
PetscErrorCode TransformObjectsGetNames(SalsaTransform tf,const char ***names)
{
  int i; const char **thenames; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  ierr = PetscMalloc(tf->n_objects*sizeof(char*),&thenames); CHKERRQ(ierr);
  for (i=0; i<tf->n_objects; i++)
    thenames[i] = tf->transformobjects[i]->name;
  *names = thenames;
  PetscFunctionReturn(0);
}

/*
 * Annotations
 */
#undef __FUNCT__
#define __FUNCT__ "SysProDefineCharAnnotation"
/*! Define a character string annotation for a transform. The index of this
  annotation can be retrieved with TransformCharAnnotationGetIndex().
  The actual annotation can be found with TransformItemGetCharAnnotation().
*/
PetscErrorCode SysProDefineCharAnnotation(const char *transform,const char *ann)
{
  SalsaTransform tf; int loc; PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = TransformGetByName(transform,&tf); CHKERRQ(ierr);

  /*
   * Initial allocation of various arrays
   */
  if (tf->n_annotate_c==0) {
    /* names */
    ierr = PetscMalloc
      (TFINC*sizeof(char*),&(tf->annotations_c)); CHKERRQ(ierr);
    ierr = PetscMemzero
      (tf->annotations_c,TFINC*sizeof(char*)); CHKERRQ(ierr);
    loc = 0;
  } else if (tf->n_annotate_c==TFINC) {
    SETERRQ(PETSC_COMM_SELF,1,"Can not realloc annotations yet");
  } else {
    loc = tf->n_annotate_c;
  }

  /*
   * now setup this annotation
   */
  tf->annotations_c[loc] = strdup(ann);

  tf->n_annotate_c++;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformCharAnnotationGetIndex"
PetscErrorCode TransformCharAnnotationGetIndex
(SalsaTransform tf,const char *ann,int *idx,PetscBool *flg)
{
  int i; PetscBool f; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  *idx = -1;
  for (i=0; i<tf->n_annotate_c; i++) {
    ierr = PetscStrcmp(ann,tf->annotations_c[i],&f); CHKERRQ(ierr);
    if (f) {
      *idx = i; if (flg) *flg = f; break;
    }
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SysProDefineIntAnnotation"
/*! Define a integer string annotation for a transform. The index of this
  annotation can be retrieved with TransformIntAnnotationGetIndex().
  The actual annotation can be found with TransformItemGetIntAnnotation().
*/
PetscErrorCode SysProDefineIntAnnotation(const char *transform,const char *ann)
{
  SalsaTransform tf; int loc; PetscErrorCode ierr;
  PetscFunctionBegin;
  ierr = TransformGetByName(transform,&tf); CHKERRQ(ierr);
  /*
   * Initial allocation of various arrays
   */
  if (tf->n_annotate_i==0) {
    /* names */
    ierr = PetscMalloc
      (TFINC*sizeof(char*),&(tf->annotations_i)); CHKERRQ(ierr);
    ierr = PetscMemzero
      (tf->annotations_i,TFINC*sizeof(char*)); CHKERRQ(ierr);
    loc = 0;
  } else if (tf->n_annotate_i==TFINC) {
    SETERRQ(PETSC_COMM_SELF,1,"Can not realloc annotations yet");
  } else {
    loc = tf->n_annotate_i;
  }

  /*
   * now setup this annotation
   */
  tf->annotations_i[loc] = strdup(ann);

  tf->n_annotate_i++;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformIntAnnotationGetIndex"
PetscErrorCode TransformIntAnnotationGetIndex
(SalsaTransform tf,const char *ann,int *idx,PetscBool *flg)
{
  int i; PetscBool f; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  *idx = -1;
  for (i=0; i<tf->n_annotate_i; i++) {
    ierr = PetscStrcmp(ann,tf->annotations_i[i],&f); CHKERRQ(ierr);
    if (f) {
      *idx = i; if (flg) *flg = f; break;
    }
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectMark"
PetscErrorCode TransformObjectMark(SalsaTransformObject tf)
{
  int j;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  for (j=0; j<tf->n_options; j++)
    tf->options_marked[j] = 1;
  tf->marked = 1;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectUnmark"
PetscErrorCode TransformObjectUnmark(SalsaTransformObject tf)
{
  int j;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  for (j=0; j<tf->n_options; j++)
    tf->options_marked[j] = 0;
  tf->marked = 0;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectsMarkAll"
PetscErrorCode TransformObjectsMarkAll(SalsaTransform tf)
{
  int i; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  for (i=0; i<tf->n_objects; i++) {
    SalsaTransformObject to = tf->transformobjects[i];
    CHECKVALIDTRANSFORMOBJECT(to);
    ierr = TransformObjectMark(to); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectsUnmarkAll"
PetscErrorCode TransformObjectsUnmarkAll(SalsaTransform tf)
{
  int i; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  for (i=0; i<tf->n_objects; i++) {
    SalsaTransformObject to = tf->transformobjects[i];
    CHECKVALIDTRANSFORMOBJECT(to);
    ierr = TransformObjectUnmark(to); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectGetMark"
PetscErrorCode TransformObjectGetMark(SalsaTransformObject tf,int *m)
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  *m = tf->marked;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformGetNUnmarked"
PetscErrorCode TransformGetNUnmarked(SalsaTransform tf,int *n)
{
  int i;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  *n = 0;
  for (i=0; i<tf->n_objects; i++)
    if (!tf->transformobjects[i]->marked) (*n)++;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PreprocessorSaveAprioriSelection"
PetscErrorCode PreprocessorSaveAprioriSelection(SystemPreprocessor pp)
{
  SalsaTransform tf = pp->transform; int ipp;
  PetscFunctionBegin;
  for (ipp=0; ipp<tf->n_objects; ipp++) {
    SalsaTransformObject to = tf->transformobjects[ipp];
    CHECKVALIDTRANSFORMOBJECT(to);
    tf->aprioriselection[ipp] = to->marked;
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "PreprocessorApplyAprioriSelection"
PetscErrorCode PreprocessorApplyAprioriSelection(SystemPreprocessor pp)
{
  SalsaTransform tf = pp->transform; int ipp;
  PetscFunctionBegin;
  for (ipp=0; ipp<tf->n_objects; ipp++) {
    tf->transformobjects[ipp]->marked = tf->aprioriselection[ipp];
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectsUseOnly"
/*! Mark a list of names as to be used.

  Cases:
  - \c "name1,name2,name3" : all other names are marked as not to be used
  - \c "not,name1,name2" : all names will be used, except for the
  ones listed
*/
PetscErrorCode TransformObjectsUseOnly(SalsaTransform tf,const char *list)
{
  char *c; PetscErrorCode ierr; PetscBool eliminate;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  c = strtok((char*)list,",");
  eliminate = (PetscBool)(c && (PetscBool)!strcmp(c,"not"));
  if (eliminate) {
    ierr = TransformObjectsUnmarkAll(tf); CHKERRQ(ierr);
    c = strtok(NULL,",");
  } else {
    ierr = TransformObjectsMarkAll(tf); CHKERRQ(ierr);
  }
  while (c) {
    SalsaTransformObject to;
    ierr = TransformObjectGetByName(tf->name,c,&to); CHKERRQ(ierr);
    if (eliminate) {
      ierr = TransformObjectMark(to); CHKERRQ(ierr);
    } else {
      ierr = TransformObjectUnmark(to); CHKERRQ(ierr);
    }
    c = strtok(NULL,",");
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformGetNextUnmarkedItem"
/*! Find the next unmarked value; if \c old is NULL, the first
  first unmarked value is given, otherwise the first one after
  a match with \c old.
*/
PetscErrorCode TransformGetNextUnmarkedItem
(SalsaTransform tf,const char *old,SalsaTransformObject *snew,PetscBool *f)
{
  int idx; PetscBool foundtheoldone,found;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  found = PETSC_FALSE;
  foundtheoldone = TRUTH(old==NULL);
  for (idx=0; idx<tf->n_objects; idx++) {
    SalsaTransformObject to = tf->transformobjects[idx];
    CHECKVALIDTRANSFORMOBJECT(to);
    if (!foundtheoldone) {
      /* skip iterations until we have a match with the old name;
	 then we take the next iteration
      */
      foundtheoldone = (PetscBool)!strcmp(old,to->name);
      continue;
    }
    found = (PetscBool)(!to->marked);
    if (found) {
      *snew = to; 
      if (f) *f = PETSC_TRUE;
      PetscFunctionReturn(0);
    }
  }
  if (f) *f = found;
  else SETERRQ(PETSC_COMM_SELF,1,"Could not find, but no flag given to report");
  PetscFunctionReturn(0);
}

#define STRDUP(a) ( (a) ? strdup(a) : NULL)

#undef __FUNCT__
#define __FUNCT__ "TransformObjectDefineOption"
PetscErrorCode TransformObjectDefineOption
(SalsaTransformObject tf,const char *opt)
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  tf->option = strdup(opt);
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectAddOption"
PetscErrorCode TransformObjectAddOption(SalsaTransformObject tf,int v)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  if (!tf->options) {
    ierr = PetscMalloc(TFINC*sizeof(int),&tf->options); CHKERRQ(ierr);
    ierr = PetscMemzero(tf->options,TFINC*sizeof(int)); CHKERRQ(ierr);
    ierr = PetscMalloc(TFINC*sizeof(int),&(tf->options_marked)); CHKERRQ(ierr);
    ierr = PetscMemzero(tf->options_marked,TFINC*sizeof(int)); CHKERRQ(ierr);
    ierr = PetscMalloc(TFINC*sizeof(char*),&(tf->optionexplanation)); CHKERRQ(ierr);
    ierr = PetscMemzero(tf->optionexplanation,TFINC*sizeof(char*)); CHKERRQ(ierr);
    tf->n_options = 0; tf->alloc_options = TFINC;
  }
  if (tf->n_options>=tf->alloc_options)
    SETERRQ1(PETSC_COMM_SELF,1,"No space for further options in tranform %s",tf->name);
  tf->options[tf->n_options++] = v;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectAddOptionExplanation"
PetscErrorCode TransformObjectAddOptionExplanation
(SalsaTransformObject tf,int opt,const char *ex)
{
  int i;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  for (i=0; i<tf->n_options; i++) {
    if (tf->options[i]==opt) {
      tf->optionexplanation[i] = strdup(ex);
      goto out;
    }
  }
 out:
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformItemOptionsUseOnly"
PetscErrorCode TransformItemOptionsUseOnly(SalsaTransformObject tf,const char *opt)
{
  char *c; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  c = strtok((char*)opt,",");
  tf->n_options = 0;
  while (c) {
    int cv;
    sscanf(c,"%d",&cv);
    ierr = TransformObjectAddOption(tf,cv); CHKERRQ(ierr);
    c = strtok(NULL,",");
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformItemGetFirstOption"
PetscErrorCode TransformItemGetFirstOption
(const char* tf,const char *it,int *v,PetscBool *f)
{
  SalsaTransformObject to; PetscBool found,flg;
  int i; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = TransformObjectGetByName(tf,it,&to); CHKERRQ(ierr);
  found = PETSC_FALSE; flg = PETSC_FALSE;
  for (i=0; i<to->n_options; i++) {
    if (!to->options_marked[i]) {
      flg = PETSC_TRUE;
      *v = to->options[i];
      goto out;
    }
  }

 out:
  if (f) *f = flg;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformItemGetNextOption"
PetscErrorCode TransformItemGetNextOption
(const char* tf,const char *it,int *v,PetscBool *f)
{
  SalsaTransformObject to; PetscBool found,flg;
  int i; PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = TransformObjectGetByName(tf,it,&to); CHKERRQ(ierr);
  found = PETSC_FALSE; flg = PETSC_FALSE;
  for (i=0; i<to->n_options; i++) {
    if (found && !to->options_marked[i]) {
      flg = PETSC_TRUE;
      *v = to->options[i];
      goto out;
    }
    if (to->options[i]==*v) found = PETSC_TRUE;
 }

 out:
  if (f) *f = flg;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformItemOptionMark"
PetscErrorCode TransformItemOptionMark(SalsaTransform tf,const char *it,int o)
{
  SalsaTransformObject to; int i; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  ierr = TransformObjectGetByName(tf->name,it,&to); CHKERRQ(ierr);
  to->marked = 0;
  for (i=0; i<to->n_options; i++) 
    if (to->options[i]==o) to->options_marked[i] = 1;

  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectCharAnnotate"
PetscErrorCode TransformObjectCharAnnotate
(SalsaTransformObject tf,const char *an,const char *v)
{
  int idx; PetscBool f; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  if (!tf->annotate_c) {
    ierr = PetscMalloc(TFINC*sizeof(char*),&tf->annotate_c); CHKERRQ(ierr);
    ierr = PetscMemzero(tf->annotate_c,TFINC*sizeof(char*)); CHKERRQ(ierr);
    tf->alloc_annotate_c = TFINC;
  }
  ierr = TransformCharAnnotationGetIndex
    (tf->transform,an,&idx,&f); CHKERRQ(ierr);
  if (!f) SETERRQ1(PETSC_COMM_SELF,1,"Could not find annotation %s",an);
  if (idx>=tf->alloc_annotate_c)
    SETERRQ1(PETSC_COMM_SELF,1,"Char annotate index too large: %d.",idx);
  tf->annotate_c[idx] = v;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectIntAnnotate"
PetscErrorCode TransformObjectIntAnnotate
(SalsaTransformObject tf,const char *an,int v)
{
  int idx; PetscBool f; PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  if (!tf->annotate_i) {
    ierr = PetscMalloc(TFINC*sizeof(int),&tf->annotate_i); CHKERRQ(ierr);
    ierr = PetscMemzero(tf->annotate_i,TFINC*sizeof(int)); CHKERRQ(ierr);
    tf->alloc_annotate_i = TFINC;
  }
  ierr = TransformIntAnnotationGetIndex
    (tf->transform,an,&idx,&f); CHKERRQ(ierr);
  if (!f) SETERRQ1(PETSC_COMM_SELF,1,"Could not find annotation %s",an);
  if (idx>=tf->alloc_annotate_i)
    SETERRQ1(PETSC_COMM_SELF,1,"Int annotate index too large: %d.",idx);
  tf->annotate_i[idx] = v;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformObjectGetIntAnnotation"
PetscErrorCode TransformObjectGetIntAnnotation
(SalsaTransformObject tf,const char *an,int *v,PetscBool *f)
{
  SalsaTransform tt = tf->transform; int ian;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORMOBJECT(tf);
  *f = PETSC_FALSE;
  for (ian=0; ian<tt->n_annotate_i; ian++) {
    if (!strcmp(tt->annotations_i[ian],an)) {
      *f = TRUTH(tf->annotate_i!=NULL);
      if (*f) *v = tf->annotate_i[ian];
      PetscFunctionReturn(0);
    }
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformReportTeXTable"
PetscErrorCode TransformReportTeXTable(SalsaTransform tf,FILE *f)
{
  int i,j;
  PetscFunctionBegin;

  for (i=0; i<tf->n_objects; i++) {
    SalsaTransformObject to = tf->transformobjects[i];
    fprintf(f,"%s&%s",to->name,to->explanation);
    if (to->n_options>0) {
      /*
      fprintf(f,"&%s&",to->optionexplanation);
      */
      for (j=0; j<to->n_options; j++) {
	fprintf(f,"%d,",to->options[j]);	
      }
    } else fprintf(f,"&&");
    fprintf(f,"\\\\\n");
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformReportEnabled"
PetscErrorCode TransformReportEnabled(SalsaTransform tf,const char **rs)
{
  const char *s,**names; int i; size_t l; PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = PetscMalloc(1000*sizeof(char),&s); CHKERRQ(ierr);
  ierr = PetscMemzero((char*)s,1000*sizeof(char)); CHKERRQ(ierr);
  ierr = TransformObjectsGetNames(tf,&names); CHKERRQ(ierr);
  for (i=0; i<tf->n_objects; i++) {
    int mark;
    ierr = TransformObjectGetMark(tf->transformobjects[i],&mark); CHKERRQ(ierr);
    if (!mark) {
      ierr = PetscStrlen(s,&l); CHKERRQ(ierr);
      if (l) {sprintf((char*)s+l," "); l++;} sprintf((char*)s+l,"%s",names[i]);
    }
  }
  ierr = PetscFree(names); CHKERRQ(ierr);
  *rs = s;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformItemDescribeShort"
PetscErrorCode TransformItemDescribeShort(SalsaTransform tf,char *it,int opt,char **s)
{
  SalsaTransformObject to; char *ss; size_t l; PetscErrorCode ierr;
  PetscFunctionBegin;

  ierr = TransformObjectGetByName(tf->name,it,&to); CHKERRQ(ierr);
  ierr = PetscMalloc(200*sizeof(char),&ss); CHKERRQ(ierr);
  sprintf(ss,"%s",to->name); CHKERRQ(ierr);
  ierr = PetscStrlen(ss,&l); CHKERRQ(ierr);
  if (to->n_options) {
    sprintf(ss+l,"(%d)",opt);
    ierr = PetscStrlen(ss,&l); CHKERRQ(ierr);
  }
  *s = ss;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformItemDescribeLong"
PetscErrorCode TransformItemDescribeLong(SalsaTransform tf,char *it,int opt,char **s)
{
  SalsaTransformObject to;
  char *ss; size_t l; PetscErrorCode ierr;

  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  ierr = TransformObjectGetByName(tf->name,it,&to); CHKERRQ(ierr);
  ierr = PetscMalloc(200*sizeof(char),&ss); CHKERRQ(ierr);
  sprintf(ss,"%s",to->explanation); CHKERRQ(ierr);
  ierr = PetscStrlen(ss,&l); CHKERRQ(ierr);
#if 0
  if (to->options[idx]) {
    sprintf(ss+l,"; %s",tf->options[idx]);
    ierr = PetscStrlen(ss,&l); CHKERRQ(ierr);
  }
  if (tf->noptions[idx]) {
    ierr = TransformItemGetOptionIndex(tf,idx,opt,&odx); CHKERRQ(ierr);
    if (odx>=0)
      sprintf(ss+l," %s",tf->options_texts[idx][odx]);
    else
      sprintf(ss+l,"(%d)",opt);
    ierr = PetscStrlen(ss,&l); CHKERRQ(ierr);
  }
#endif
  *s = ss;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformSetUserChoices"
/*! Declare that the user has ordained which choice(s) to take for a 
  specific transform. The values of these choices are implicitly given
  by disabling some transforms.

  See also TransformGetUserChoices(), TransformObjectsUseOnly().
*/
PetscErrorCode TransformSetUserChoices(SalsaTransform tf,PetscBool ch)
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  tf->userchoices = ch;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TransformGetUserChoices"
/*! Query whether the user has ordained which choice(s) to take for a 
  specific transform. The values of these choices are implicitly given
  by disabling some transforms.

  See also TransformSetUserChoices(), TransformObjectsUseOnly(). 
*/
PetscErrorCode TransformGetUserChoices(SalsaTransform tf,PetscBool *ch)
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  *ch = tf->userchoices;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SetSuitabilityContextTester"
PetscErrorCode SetSuitabilityContextTester(SalsaTransform tf,PetscErrorCode(*f)(void*))
{
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM(tf);
  tf->suitabilityctxtester = f;
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "TestSuitabilityContext"
PetscErrorCode TestSuitabilityContext(SalsaTransform tf,void *ctx)
{
  PetscErrorCode ierr;
  PetscFunctionBegin;
  CHECKVALIDTRANSFORM0(tf);
  if (tf->suitabilityctxtester) {
    ierr = (tf->suitabilityctxtester)(ctx); CHKERRQ(ierr);
  }
  PetscFunctionReturn(0);
}

#undef __FUNCT__
#define __FUNCT__ "SalsaTransformIntegrityTest"
PetscErrorCode SalsaTransformIntegrityTest(SalsaTransform tf)
{
  int i; PetscErrorCode ierr;
  PetscFunctionBegin;
  if (!tf || !tf->suitabilityctxtester) PetscFunctionReturn(0);
  for (i=0; i<tf->n_objects; i++) {
    SalsaTransformObject to = tf->transformobjects[i]; void *sctx;
    ierr = TransformObjectGetSuitabilityFunction
      (to,&sctx,PETSC_NULL); CHKERRQ(ierr);
    if (sctx) {
      ierr = TestSuitabilityContext(tf,sctx); CHKERRQ(ierr);
    }
  }
  PetscFunctionReturn(0);
}
